import math

#This module includes functions for evaluating the performance of a tracker. The main functions are:
#evaluateTrajectoriesMode1
#loadMode2Results

def readTrajectories(filePath):
	#Return the trajectories in the pointed-to file as a list of trajectory lists. A trajectory list
	#is a list of cells. A cell is a list of 3 items [f, x, y] where f is a frame number and (x,y) is
	#the position of a features in frame f.
	trajectories = list()
	dataFile = open(filePath,'r')
	for line in dataFile:
		line = line.lstrip(' ')
		line = line.rstrip('\n\r')
		if (len(line) > 0) and (line[0] != '#'):
			cells = line.split(':')
			newTrajectory = list()
			for cell in cells:
				cell = cell.lstrip('(')
				cell = cell.rstrip(')')
				numbers = cell.split(',')
				if len(numbers) != 3:
					print "Error parsing input file. Invalid cell. Aborting."
					return
				frameNumber = int(numbers[0])
				x = float(numbers[1])
				y = float(numbers[2])
				newTrajectory.append([frameNumber,x,y])
			trajectories.append(newTrajectory)
	dataFile.close()
	return(trajectories)

def printTrajectories(trajectories):
	for traj in trajectories:
		for cell in traj:
			print "[%2d, %8.2f, % 8.2f]"%(cell[0],cell[1],cell[2]),
		print ""
	print "\n"

def ThrowOutTrajectoriesThatStartAfterFrame0(trajectories):
	#Alter trajectories in place to throw out all trajectories that are born after frame 0
	n = 0
	while n < len(trajectories):
		if (trajectories[n][0][0] != 0):
			trajectories.pop(n)
		else:
			n += 1

def throwOutTrajectoriesThatDieEarly(endFrame, trajectories):
	#Alters trajectories in place. Element n is removed if trajectories[n] dies before endFrame
	n = 0
	while n < len(trajectories):
		if (trajectories[n][len(trajectories[n])-1][0] < endFrame):
			trajectories.pop(n)
		else:
			n += 1

def extendTrajectoriesToFrame(trajectories, endFrame, xmin, xmax, ymin, ymax):
	#Any trajectory in trajectories that dies before endFrame has its last position projected into the viewable area
	#and copied through endFrame. This is done in place.
	for traj in trajectories:
		endCell = traj[len(traj)-1]
		if endCell[0] < endFrame:
			#We need to extend this trajectory
			finalx = endCell[1]
			finaly = endCell[2]
			if finalx < xmin:
				finalx = xmin
			if finalx > xmax:
				finalx = xmax
			if finaly < ymin:
				finaly = ymin
			if finaly > ymax:
				finaly = ymax
			while traj[len(traj)-1][0] < endFrame:
				traj.append([traj[len(traj)-1][0]+1, finalx, finaly])

def throwOutTrajectoriesWhereReferenceLeavesBox(xmin, xmax, ymin, ymax, endFrame, referenceTrajectories):
	#Alters referenceTrajectories in place. item n is removed if referenceTrajectories[n] contains
	#any coordinates prior to or on endFrame that are outside of the specified box.
	n = 0
	while n < len(referenceTrajectories):
		violationFound = False
		for cell in referenceTrajectories[n]:
			f = cell[0]
			x = cell[1]
			y = cell[2]
			if (f <= endFrame) and ((x < xmin) or (x > xmax) or (y < ymin) or (y > ymax)):
				violationFound = True
				break
		if violationFound:
			referenceTrajectories.pop(n)
		else:
			n += 1

def trimTrajectories(trajectories, endFrame):
	#Trim each trajectory in trajectories by tossing out positions that occur after endFrame
	for traj in trajectories:
		if traj[len(traj)-1][0] > endFrame:
			#We need to trim this trajectory
			n = 0
			while n < len(traj):
				if traj[n][0] > endFrame:
					traj.pop(n)
				else:
					n += 1

def getMatchingTrajectories(inputTrajectories, referenceTrajectories):
	#Return a list where item n is the trajectory from inputTrajectories that started at the same location as the trajectory
	#referenceTrajectories[n]. If there are trajectories in referenceTrajectories with no corresponding trajectories in
	#inputTrajectories, the function will raise an exception.
	outputTrajectories = list()
	for refTraj in referenceTrajectories:
		matchFound = False
		startFrame = refTraj[0][0]
		startx     = refTraj[0][1]
		starty     = refTraj[0][2]
		for traj in inputTrajectories:
			f = traj[0][0]
			x = traj[0][1]
			y = traj[0][2]
			if (f == startFrame) and (x == startx) and (y == starty):
				outputTrajectories.append(traj)
				matchFound = True
				break
		if (matchFound == False):
			print("Could not find matching trajectory for trajectory starting at (%f, %f) in frame %d"%(startx, starty, startFrame))
			raise error("No trajectory found in inputTrajectories to match the given reference trajectory")
	return(outputTrajectories)

def getAverageL1Error(trajectories1, trajectories2):
	#Return the mean L1 distance between two sets of trajectories. trajectories1[n] is compared with trajectories2[n].
	totalError = 0
	for n in range(len(trajectories1)):
		traj1 = trajectories1[n]
		traj2 = trajectories2[n]
		TotalErrorForTraj = 0
		for i in range(len(traj1)):
			try:
				f1 = traj1[i][0]
				x1 = traj1[i][1]
				y1 = traj1[i][2]
				f2 = traj2[i][0]
				x2 = traj2[i][1]
				y2 = traj2[i][2]
			except:
				print traj1
				print traj2
			if (f1 != f2):
				raise error("Comparing trajectories with different time domains!")
			TotalErrorForTraj += math.sqrt(pow(x2-x1,2) + pow(y2-y1,2))
		totalError += TotalErrorForTraj
	MeanL1Error = totalError/len(trajectories1)
	return(MeanL1Error)

def getAverageDrift(trajectories1, trajectories2):
	#Return the mean distance between the endpoints of two sets of trajectories. trajectories1[n] is compared with trajectories2[n].
	totalDrift = 0
	for n in range(len(trajectories1)):
		traj1 = trajectories1[n]
		traj2 = trajectories2[n]
		finalInd = len(traj1)-1
		f1 = traj1[finalInd][0]
		x1 = traj1[finalInd][1]
		y1 = traj1[finalInd][2]
		f2 = traj2[finalInd][0]
		x2 = traj2[finalInd][1]
		y2 = traj2[finalInd][2]
		if (f1 != f2):
			raise error("Comparing drift of trajectories with different end times!")
		drift = math.sqrt(pow(x2-x1,2) + pow(y2-y1,2))
		totalDrift += drift
	meanDrift = totalDrift/len(trajectories1)
	return(meanDrift)

def evaluateTrajectoriesMode1(trajectoriesFile, groundTruthTrajectoriesFile, numFrames, xmin, xmax, ymin, ymax, metric="Mean L1 Error"):
	#Returns mean L1 error or mean drift for trackable features that start at frame 0 and go through numFrames. A
	#feature is "trackable" if the ground truth for the feature stays in the bounds [xmin, xmax, ymin, ymax]
	#for the entire time interval between frame 0 and frame numFrames. Tracker trajectories that are terminated
	#pre-maturely for leaving the viewable area are extened by projecting their last known position into the
	#viewable area and repeating that position through to frame numFrames.
	groundTruthTrajectories = readTrajectories(groundTruthTrajectoriesFile)
	trackerTrajectories     = readTrajectories(trajectoriesFile)
	ThrowOutTrajectoriesThatStartAfterFrame0(groundTruthTrajectories)
	throwOutTrajectoriesThatDieEarly(numFrames, groundTruthTrajectories)
	throwOutTrajectoriesWhereReferenceLeavesBox(xmin, xmax, ymin, ymax, numFrames, groundTruthTrajectories)
	trackerTrajectories = getMatchingTrajectories(trackerTrajectories, groundTruthTrajectories)
	trimTrajectories(trackerTrajectories,     numFrames)
	trimTrajectories(groundTruthTrajectories, numFrames)
	extendTrajectoriesToFrame(trackerTrajectories, numFrames, xmin, xmax, ymin, ymax)
	if metric == "Mean L1 Error":
		return(getAverageL1Error(groundTruthTrajectories, trackerTrajectories))
	elif metric == "Mean Drift":
		return(getAverageDrift(groundTruthTrajectories, trackerTrajectories))
	else:
		raise error("Invalid metric argument!")

def loadMode2Results(ResultsFilePath):
	FID = open(ResultsFilePath, 'r')
	for line in FID:
		parts = line.split(':')
		if parts[0] == "Total Re-initializations":
			totalReInits = int(parts[1])
		elif parts[0] == "Total feature frame count":
			totalFeatureFrames = int(parts[1])
		elif parts[0] == "Average frames per re-initialization":
			meanFramesBetweenReInits = round(10*float(parts[1]))/10 #Only use 1 digit after decimal point
		elif parts[0] == "Re-initialization tolerance (pixels)":
			reInitTolerance = float(parts[1])
		elif parts[0] == "Mean processing frame-rate (fps)":
			processingFrameRate = float(parts[1])
	FID.close()
	return({'totalReInits':totalReInits,'totalFeatureFrames':totalFeatureFrames,'meanFramesBetweenReInits':meanFramesBetweenReInits,\
		   'reInitTolerance':reInitTolerance,'processingFrameRate':processingFrameRate})
