#include <cstdint>
#include <string>
#include "Feature.h"
#include "parameters.h"
#include "Utilities.h"
#include "EnergyEvaluation.h"

using namespace cv;
using namespace std;

void Feature::initializeFeature(int frameNumber, const ImageStack& frameStack, const Vec2f& location, int FeatureID) {
	featureID = FeatureID;
	BirthFrameNumber = frameNumber;
	CurrentFrameNumber = frameNumber;
	currentScore = 2.0*parameters::minFeatureScore;
	CurrentLocation = location;
	History.push_front(location);
	Template.BuildFromStack(frameStack, location, parameters::n);
	RecentTemplateCandidates.push_front(Template);
	
	groundTruthTrajectoryLoaded = false;
	GroundTruth_BirthFrameNumber = 0;
	GroundTruth_LastFrameNumber = 0;
	
	//This BatchNumber field is initialized to 0, which is an acceptable value.
	//However, to prevent all features from having their templates updated on the same frames,
	//Feature batch numbers should be evenly distributed between 0 and 9. This is managed by the
	//FeatureSet class - it sets the BatchNumber field when a feature is added to it.
	BatchNumber = 0;
}

void Feature::Update(int frameNumber, const ImageStack& frameStack, const cv::Vec2f& location) {
	CurrentFrameNumber = frameNumber;
	CurrentLocation = location;
	
	//Only save location to history if the current frame is after the birth frame. This is to ensure that
	//we do not have 2 entries corresponding to the birth frame, if one tries to track this feature on the birth frame.
	if (frameNumber > BirthFrameNumber)
		History.push_back(location);
	
	ImageStack candidateTemplate;
	candidateTemplate.BuildFromStack(frameStack, location, parameters::n);
	RecentTemplateCandidates.push_front(candidateTemplate);
	if (RecentTemplateCandidates.size() > 10)
		RecentTemplateCandidates.pop_back();
	
	//If necessary, update active template
	if (BatchNumber % 5 == frameNumber % 5)
		forceTemplateUpdate();
}

void Feature::forceTemplateUpdate(void) {
	//Find the best template candidate and replace the active template with that candidate
	Mat Reference = Template.Images[0];
	list< ImageStack >::iterator bestCandidateIterator;
	double lowestError = 0;
	double error = 0.0;
	for (list< ImageStack >::iterator it = RecentTemplateCandidates.begin(); it != RecentTemplateCandidates.end(); it++) {
		Mat Candidate = it->Images[0];
		
		error = Utilities::computeL1Diff_SIMD(Candidate, Reference);
		if (it == RecentTemplateCandidates.begin()) {
			//Initialize bestCandidate data
			bestCandidateIterator = it;
			lowestError = error;
		}
		else {
			//If the current error is lower than the current best, update bestCandidate data
			if (error < lowestError) {
				bestCandidateIterator = it;
				lowestError = error;
			}
		}
	}
	Template = *(bestCandidateIterator);
}

void Feature::renderFeature(ImageStack& frameStack) {
	//Render Template
	imshow("Template", Template.Images[0]);
	
	//Render Energy Surface
	int X = 45;
	int Y = 45;
	Vec2f location;
	Mat energySurface(X,Y,CV_32FC1);
	for (int x=0; x<X; x++) {
		for (int y=0; y<Y; y++) {
			location = CurrentLocation - Vec2f(((float)X - 1)/2, ((float)Y - 1)/2) + Vec2f(x, y);
			energySurface.at<float>(x,y) = (float)compute_fitTerm(Template, frameStack, 0, location, parameters::n);
		}
	}
	double alpha = 1.0 / (pow(parameters::n, 2) * (double) parameters::maxIntensity);
	energySurface = alpha * energySurface;
	double minVal;
	double maxVal;
	minMaxLoc(energySurface, &minVal, &maxVal, NULL, NULL);
	imshow("Energy Surface Level 0", (1.0/maxVal)*energySurface);
	//fprintf(stderr,"\nMin: % 7f    Max: % 7f", minVal, maxVal);
	
	//Render Energy Surface
	Vec2f CurrentLocationLevel1;
	ImageStack::convertImageCoordinates(CurrentLocation, CurrentLocationLevel1, 0, 1);
	for (int x=0; x<X; x++) {
		for (int y=0; y<Y; y++) {
			location = CurrentLocationLevel1 - Vec2f(((float)X - 1)/2, ((float)Y - 1)/2) + Vec2f(x, y);
			energySurface.at<float>(x,y) = (float)compute_fitTerm(Template, frameStack, 1, location, parameters::n);
		}
	}
	energySurface = alpha * energySurface;
	minMaxLoc(energySurface, &minVal, &maxVal, NULL, NULL);
	imshow("Energy Surface Level 1", (1.0/maxVal)*energySurface);
	//fprintf(stderr,"\nMin: % 7f    Max: % 7f", minVal, maxVal);
}

void Feature::loadGroundTruth(std::list<cv::Vec2f> &Trajectory, int birthFrameNumber) {
	//This function loads in a ground truth trajectory for the given feature.
	GroundTruth_Trajectory = vector<Vec2f>(Trajectory.begin(), Trajectory.end());
	GroundTruth_BirthFrameNumber = birthFrameNumber;
	GroundTruth_LastFrameNumber = ((int) GroundTruth_Trajectory.size()) + birthFrameNumber - 1;	
	groundTruthTrajectoryLoaded = true;
}

int Feature::groundTruth_BirthFrame(void) {
	return(GroundTruth_BirthFrameNumber);
}

int Feature::groundTruth_LastFrame(void) {
	return(GroundTruth_LastFrameNumber);
}

bool Feature::groundTruth_Alive(void) {
	if ((CurrentFrameNumber >= GroundTruth_BirthFrameNumber) && (CurrentFrameNumber <= GroundTruth_LastFrameNumber))
		return(true);
	else
		return(false);
}

Vec2f Feature::groundTruth_Location(int frameNumber) {
	Vec2f location(NAN, NAN);
	if ((frameNumber >= GroundTruth_BirthFrameNumber) && (frameNumber <= GroundTruth_LastFrameNumber)) {
		int index = frameNumber - GroundTruth_BirthFrameNumber;
		location = GroundTruth_Trajectory[index];
	}
	return(location);
}

double Feature::getDeviationFromGroundTruth(void) {
	Vec2f groundTruthlocation = groundTruth_Location(CurrentFrameNumber);
	return((double) norm(groundTruthlocation - CurrentLocation));
}


FeatureSet::FeatureSet() {
	//Initialize batch histogram
	for (int counter = 0; counter < 10; counter++)
		batchHistogram[counter] = 0;
}

int FeatureSet::size(void) {
	return((int)(Features.size()));
}

void FeatureSet::addFeature(Feature& feature) {
	//First detirmine the best batch number for the new feature
	int batchNumber = 0;
	int tally = batchHistogram[0];
	for (int counter = 1; counter < 10; counter++) {
		if (batchHistogram[counter] < tally) {
			batchNumber = counter;
			tally = batchHistogram[counter];
		}
	}
	feature.BatchNumber = batchNumber;
	Features.push_back(feature);
	batchHistogram[batchNumber]++;
}

void FeatureSet::removeFeature(list< Feature >::iterator& featureIter) {
	if ((featureIter->BatchNumber < 0) || (featureIter->BatchNumber > 9))
		fprintf(stderr,"\nError! Invalid Batch Number! Not adjusting batch histogram.");
	else
		batchHistogram[featureIter->BatchNumber]--;
	featureIter = Features.erase(featureIter);
}

void FeatureSet::clearFeatures(void) {
	Features.clear();
	for (int counter = 0; counter < 10; counter++)
		batchHistogram[counter] = 0;
}

void FeatureSet::build_xbar(Mat& xbar) {
	//Build xbar vector from current locations of each feature in set
	xbar = Mat(2*Features.size(),1,CV_32FC1);
	int row = 0;
	for (list< Feature >::iterator it = Features.begin(); it != Features.end(); it++) {
		xbar.at<float>(row++) = it->CurrentLocation[0];
		xbar.at<float>(row++) = it->CurrentLocation[1];
	}
}

void FeatureSet::build_HistoryMatrix(Mat& History, int L) {
	Mat mask;
	History = Mat::zeros(2*L, Features.size(), CV_32FC1);
	mask    = Mat::ones (2*L, Features.size(), CV_8UC1);
	
	bool missingDataFlag = false;
	int f=0;
	for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
		//Insert Features Trajectory into matrix column f
		int n = 0;
		for (list< cv::Vec2f >::iterator historyIter = featureIter->History.begin(); historyIter != featureIter->History.end(); historyIter++) {
			History.at<float>(2*L-2*n-2,f) = (*historyIter)[0];
			History.at<float>(2*L-2*n-1,f) = (*historyIter)[1];
			n++;
			if (n >= L) break;
		}
		for (; n<L; n++) {
			mask.at<uchar>(2*L-2*n-2,f) = 0U; //Mark missing data
			mask.at<uchar>(2*L-2*n-1,f) = 0U; //Mark missing data
		}
		if (featureIter->History.size() < (unsigned int) L) missingDataFlag = true;
		f++;
	}
	
	//Complete the matrix if there is any missing data
	if (missingDataFlag)
		completeMatrix(History, mask, 3, true);
}

void FeatureSet::completeMatrix(Mat& A, Mat& mask, int dim, bool affineFlag) {
	//If there are missing entries in the matrix, perform low-rank matrix completion.
	//A must be a single channel matrix of floats, and mask must be a single channel
	//matrix of unsigned chars.
	Mat centroid;	//Only used in affine mode
	Mat B;
	
	//Zero out un-observed entries in A
	for (int x=0; x<A.rows; x++) {
		for (int y=0; y<A.cols; y++) {
			if (mask.at<uchar>(x,y) == 0U)
				A.at<float>(x,y) = 0.0;
		}
	}
	
	if (affineFlag) {
		Mat ObservableRowTotals, ARowSums;
		reduce(A, ARowSums, 1, CV_REDUCE_SUM, CV_32FC1);
		reduce(mask, ObservableRowTotals, 1, CV_REDUCE_SUM, CV_32SC1);
		centroid = Mat(ObservableRowTotals.rows, 1, CV_32FC1);
		for (int x=0; x<ObservableRowTotals.rows; x++) {
			if (ObservableRowTotals.at<int32_t>(x) > 0)
				centroid.at<float>(x) = ARowSums.at<float>(x)/((float)(ObservableRowTotals.at<int32_t>(x)));
			else
				centroid.at<float>(x) = 0.0;
		}
		
		//Remove Centroid from each column of A
		for (int x=0; x<A.rows; x++) {
			for (int y=0; y<A.cols; y++) {
				A.at<float>(x,y) -= centroid.at<float>(x);
			}
		}
	}
	
	//Perform linear (as opposed to affine) matrix completion
	//This is carried out in double precision to circumvent a possible bug in the openCV SVD algorithm
	A.convertTo(A, CV_64FC1);
	A.copyTo(B);
	Mat sigma, u, vt, Sigma;
	SVD SVDObject;
	for (int counter=0; counter<30; counter++) {
		//Project Matrix to known rank
		SVDObject.compute(B, sigma, u, vt);
		for (int n=dim; n<sigma.rows; n++)
			sigma.at<double>(n) = 0.0;
		Sigma = Mat::diag(sigma);
		B = u*Sigma*vt;
		
		//Restore observed entries
		for (int x=0; x<A.rows; x++)
			for (int y=0; y<A.cols; y++)
				if (mask.at<uchar>(x,y) == 1U)
					B.at<double>(x,y) = A.at<double>(x,y);
	}
	B.convertTo(A, CV_32FC1);
	
	//If in affine mode, de-center matrix
	if (affineFlag) {
		//Remove Centroid from each column of A
		for (int x=0; x<A.rows; x++) {
			for (int y=0; y<A.cols; y++) {
				A.at<float>(x,y) += centroid.at<float>(x);
			}
		}
	}
}

float FeatureSet::getDistanceFromTrackset(Vec2f location) {
	float distance = 0.0;
	float minDistance = 0.0;
	for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
		if (featureIter == Features.begin())
			minDistance = norm(featureIter->CurrentLocation - location);
		else {
			distance = norm(featureIter->CurrentLocation - location);
			if (distance < minDistance)
				minDistance = distance;
		}
	}
	return(minDistance);
}


//Write the current locations of all features to the provided string - the strings old value is lost.
void FeatureSet::renderToString(std::string& outputString) {
	char buffer[50];
	outputString = string();
	for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
		sprintf(buffer, "% 10f % 10f ", featureIter->CurrentLocation[1], featureIter->CurrentLocation[0]);
		outputString.append(buffer);
	}
}

void FeatureSet::renderToImage(Mat& Image, Scalar color) {
	int radius = (parameters::n + 1)/2;
	if (radius < 3)
		radius = 3;
	Point location;
	if (parameters::renderFeatureNumbers) {
		int f = 0;
		string numberString;
		for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
			location = (Point)(featureIter->CurrentLocation);
			numberString = to_string(f++);
			Utilities::drawCenteredText(Image, numberString, Point(location.y, location.x), 2, 2, FONT_HERSHEY_PLAIN, color);
		}
	}
	else {
		for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
			location = (Point)(featureIter->CurrentLocation);
			circle(Image, Point(location.y, location.x), radius, color, 2);
		}
	}
}

void FeatureSet::renderToImage(Mat& Image) {
	Scalar color;
	color[0] = 0;
	color[1] = 0;
	color[2] = 0;
	color[3] = 0;
	int radius = (parameters::n + 1)/2;
	if (radius < 3)
		radius = 3;
	Point location;
	if (parameters::renderFeatureNumbers) {
		int f = 0;
		string numberString;
		for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
			location = (Point)(featureIter->CurrentLocation);
			numberString = to_string(f++);
			color[1] = 255.0*((double)(featureIter->CurrentFrameNumber-featureIter->BirthFrameNumber)/30.0);
			color[2] = 255.0 - 255.0*((double)(featureIter->CurrentFrameNumber-featureIter->BirthFrameNumber)/30.0);
			if (color[1] < 0.0)   color[1] = 0.0;
			if (color[1] > 255.0) color[1] = 255.0;
			if (color[1] < 0.0)   color[2] = 0.0;
			if (color[1] > 255.0) color[2] = 255.0;
			Utilities::drawCenteredText(Image, numberString, Point(location.y, location.x), 2, 2, FONT_HERSHEY_PLAIN, color);
		}
	}
	else {
		for (list< Feature >::iterator featureIter = Features.begin(); featureIter != Features.end(); featureIter++) {
			location = (Point)(featureIter->CurrentLocation);
			color[1] = 255.0*((double)(featureIter->CurrentFrameNumber-featureIter->BirthFrameNumber)/30.0);
			color[2] = 255.0 - 255.0*((double)(featureIter->CurrentFrameNumber-featureIter->BirthFrameNumber)/30.0);
			if (color[1] < 0.0)   color[1] = 0.0;
			if (color[1] > 255.0) color[1] = 255.0;
			if (color[1] < 0.0)   color[2] = 0.0;
			if (color[1] > 255.0) color[2] = 255.0;
			circle(Image, Point(location.y, location.x), radius, color, 2);
		}
	}
}

