#pragma once
#include "opencv2/opencv.hpp"
#include <list>
#include "ImageStack.h"

//The Feature class is for encapsulating a single feature. This class maintains a brief history of candidate tamplates
//and autonomously updates the active template when the current frame number agrees with the features batch number (mod 10).
//A feature can optionally contain a ground truth trajectory for use in tracker comparisons. This field will hold an empty
//container when tracking in a live or autonomous mode.
class Feature {
	public:
		int featureID; //A unique ID for the feature (not re-used on any new features)
		int BirthFrameNumber;
		int CurrentFrameNumber;
		float currentScore;
		
		cv::Vec2f CurrentLocation;
		std::list< cv::Vec2f > History;
		
		ImageStack Template; 
		std::list< ImageStack> RecentTemplateCandidates;
		
		//The batch number detirmines on which frames the features active template gets updated.
		int BatchNumber;
		
		//Initialize a new feature and extract its template from the provided frame stack
		void initializeFeature(int frameNumber, const ImageStack& frameStack, const cv::Vec2f& location, int FeatureID);
		
		void Update(int frameNumber, const ImageStack& frameStack, const cv::Vec2f& location);
		void forceTemplateUpdate(void); //Update the active template with the best matching candidate
		
		//The following methods are for loading and accessing a features ground truth. If any of these methods are
		//used, then the first must be 'loadGroundTruth' to initialize the feature ground truth trajectory.
		void      loadGroundTruth(std::list<cv::Vec2f> &Trajectory, int birthFrameNumber);
		int       groundTruth_BirthFrame(void);          //Get ground truth birth frame number for feature
		int       groundTruth_LastFrame(void);           //Get ground truth final frame number for feature
		bool      groundTruth_Alive(void);               //Return true if feature should currently exist according to ground truth and false otherwise.
		cv::Vec2f groundTruth_Location(int frameNumber); //Get ground truth location of feature at given frame number
		double    getDeviationFromGroundTruth(void);     //Get current distance of feature from its ground-truth location (NaN if not supposed to exist)
		
		void renderFeature(ImageStack& frameStack); //Show feature with energy surface
	private:
		bool groundTruthTrajectoryLoaded;
		std::vector<cv::Vec2f> GroundTruth_Trajectory;
		int GroundTruth_BirthFrameNumber;
		int GroundTruth_LastFrameNumber;
};


//The FeatureSet class holds a set of Feature objects. It also maintains an up-to-date histogram of used batch numbers
//for the set to facilitate rapid identification of good batch numbers to use for new features. This class provides
//methods for generating useful information about the set of features, including an xbar state vector and the current
//trackpoint (history) matrix.
class FeatureSet {
	public:
		std::list<Feature> Features;
		
		//Basic Constructor
		FeatureSet();
		
		//Accessors and Modifiers
		int size(void);
		void addFeature(Feature& feature);  //Add feature and set batch number
		void removeFeature(std::list< Feature >::iterator& featureIter);  //Remove feature from set
		void clearFeatures(void); //Clear all features
		
		//Methods for generating useful information from feature set
		//Populate xbar from current feature set
		void build_xbar(cv::Mat& xbar);
		
		//Populate history matrix from current feature set
		void build_HistoryMatrix(cv::Mat& History, int L);
		
		//Find distance from Trackset
		float getDistanceFromTrackset(cv::Vec2f location);
		
		//Utility function for display purposes
		void renderToImage(cv::Mat& Image);
		void renderToImage(cv::Mat& Image, cv::Scalar color);
		void renderToString(std::string& outputString);
		
	private:
		int batchHistogram[10];
		void completeMatrix(cv::Mat& A, cv::Mat& mask, int dim, bool affineFlag);
};
