#include "Tracker.h"
#include "parameters.h"
#include "EnergyEvaluation.h"
#include "Utilities.h"
//#include "prettyPrint.h"
#include "singleFeatureEnergy.h"

using namespace std;
using namespace cv;
using namespace Eigen;
//using namespace prettyPrint;

void Tracker::initialize(int firstFrameNumber, ImageStack firstFrameStack, int FeatureManagementMode, string inputFilePath) {
	frameNumber = firstFrameNumber;
	lastFrameStack = firstFrameStack;
	thisFrameStack = firstFrameStack;
	lastIDAssigned = 0;
	featureManagementMode = FeatureManagementMode;
	
	if (featureManagementMode == 0)
		reviewFeatureSet();
	else if (featureManagementMode == 1) {
		//Initialize features using initial locations in file
		list<Vec2f> initPositions;
		Utilities::loadInitialPositions(inputFilePath, initPositions);
		
		for (list<Vec2f>::iterator iter = initPositions.begin(); iter != initPositions.end(); iter++) {
			Feature newFeature;
			newFeature.initializeFeature(firstFrameNumber, firstFrameStack, *(iter), lastIDAssigned++);
			features.addFeature(newFeature);
		}
		
		//We don't need to do an initial review of features in mode 1 (we never add new features)
	}
	else if (featureManagementMode == 2) {
		reinitializationCounter = 0; //Clear re-initialization counter
		
		//Initialize features and load groundtruth trajectories for each feature
		list<int> birthFrameNumbers;
		list<list<Vec2f>> trajectories;
		if (Utilities::loadTrajectories(inputFilePath, birthFrameNumbers, trajectories) != 0) {
			fprintf(stderr,"Error initializing tracker. Could not load ground-truth file.\n\r");
			return;
		}
		
		int highestBirthFrame = 0;
		for (list<int>::iterator it=birthFrameNumbers.begin(); it != birthFrameNumbers.end(); it++) {
			if (*it > highestBirthFrame)
				highestBirthFrame = *it;
		}
		
		groundTruth_Trajectories.resize(highestBirthFrame + 1);
		list<int>::iterator frameNumber_it = birthFrameNumbers.begin();
		list<list<Vec2f>>::iterator trajectory_it = trajectories.begin();
		while (frameNumber_it != birthFrameNumbers.end()) {
			groundTruth_Trajectories[*frameNumber_it].push_back(*trajectory_it);
			frameNumber_it++;
			trajectory_it++;
		}
		
		//Now review feature set to generate the initial set of features
		reviewFeatureSet();
	}
	else
		fprintf(stderr,"Unrecognized featureManagementMode. Skipping feature initialization.\n\r");
}

//Review the current set of features. Depending on the feature management mode, this function does very different
//things. See the mode definitions for details.
void Tracker::reviewFeatureSet(void) {
	switch(featureManagementMode) {
		case 0:  reviewFeatureSet_mode0(); break;
		case 1:  reviewFeatureSet_mode1(); break;
		case 2:  reviewFeatureSet_mode2(); break;
		default: fprintf(stderr,"Unrecognized featureManagementMode. Skipping review of feature set.\n\r");
	}
}

void Tracker::reviewFeatureSet_mode0(void) {
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	
	Mat cornerScoreMap;
	double harrisK = 0.04; //OpenCV goodFeaturesToTrack Default
	int blockSize = 3;     //OpenCV goodFeaturesToTrack Default

	Mat inputImage;
	if (parameters::M == 0) {
		if (thisFrameStack.Images[0].channels() == 1)
			inputImage = thisFrameStack.Images[0];
		else if (thisFrameStack.Images[0].channels() == 3)
			cvtColor(thisFrameStack.Images[0], inputImage, CV_RGB2GRAY);
		else
			fprintf(stderr,"\nInput image to findNewFeatures must have type CV_8UC1 or CV_8UC3.");
		cornerHarris(inputImage, cornerScoreMap, blockSize, 3, harrisK);
	}
	else {
		if (thisFrameStack.Images[1].channels() == 1)
			inputImage = thisFrameStack.Images[1];
		else if (thisFrameStack.Images[1].channels() == 3)
			cvtColor(thisFrameStack.Images[1], inputImage, CV_RGB2GRAY);
		else
			fprintf(stderr,"\nInput image to findNewFeatures must have type CV_8UC1 or CV_8UC3.");
		Mat lowResScoreMap;
		cornerHarris(inputImage, lowResScoreMap, blockSize, 3, harrisK);
		pyrUp(lowResScoreMap, cornerScoreMap, thisFrameStack.Images[0].size());
	}
	cornerScoreMap = abs(cornerScoreMap);
	
	//Update each features current score
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		int x = (int)round(featureIter->CurrentLocation[0]);
		int y = (int)round(featureIter->CurrentLocation[1]);
		if (x < 0) x=0;
		if (x > X-1) x=X-1;
		if (y < 0) y=0;
		if (y > Y-1) y=Y-1;
		
		//The constant alpha dictates how rapidly a features score adjusts to new imagery.
		const float alpha = 0.05;
		featureIter->currentScore = alpha*cornerScoreMap.at<float>(x,y) + (1.0 - alpha)*featureIter->currentScore;
		
		//For debug purposes, print out scores
		//if (featureIter == features.Features.begin())
		//	fprintf(stderr,"\nFeature Score for feature 0: %f", featureIter->currentScore);
	}
	
	//Review current features
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		//Check to see if the feature is too close to an image edge or if the features score is insufficient for tracking
		double x = featureIter->CurrentLocation[0];
		double y = featureIter->CurrentLocation[1];
		float featureScore = featureIter->currentScore;
		if ((x < parameters::oldFeatureEdgeRadius) || (x >= (double)X - parameters::oldFeatureEdgeRadius) ||
		    (y < parameters::oldFeatureEdgeRadius) || (y >= (double)Y - parameters::oldFeatureEdgeRadius) ||
		    (featureScore < parameters::minFeatureScore)) {
			//Feature needs to be removed
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
	
	if (parameters::extensiveFeatureReviewPeriod > 0)
		extensiveReviewOfFeatures(cornerScoreMap);
	
	//Add new features, if necessary.
	int numberOfNewFeatures = parameters::numberOfFeatures - features.size();
	if (numberOfNewFeatures > 0) {
		list<Vec2f> newFeatureLocations;
		findNewFeatures(numberOfNewFeatures, cornerScoreMap, newFeatureLocations);
		for (list< Vec2f >::iterator newFeatureIter = newFeatureLocations.begin(); newFeatureIter != newFeatureLocations.end(); newFeatureIter++) {
			Feature newFeature;
			newFeature.initializeFeature(frameNumber, thisFrameStack, *(newFeatureIter), lastIDAssigned++);
			features.addFeature(newFeature);
		}
	}
}

void Tracker::extensiveReviewOfFeatures(Mat& cornerScoreMap) {
	//Exhaustive reviews are being conducted periodically - we make sure we are tracking the best features in the scene.
	if ((parameters::extensiveFeatureReviewPeriod > 0) && (frameNumber % parameters::extensiveFeatureReviewPeriod == 0)) {
		//It is time to perform an exhaustive review of our feature set
		list<Vec2f> TopFeatureLocations;
	
		int X = thisFrameStack.Images[0].rows;
		int Y = thisFrameStack.Images[0].cols;
		int cellSize = (X + Y)/20;

		int XNumCells, YNumCells;
		if (X % cellSize == 0)
			XNumCells = X/cellSize;
		else
			XNumCells = X/cellSize + 1;
		if (Y % cellSize == 0)
			YNumCells = Y/cellSize;
		else
			YNumCells = Y/cellSize + 1;
		Mat gridScoreMap = Mat::zeros(XNumCells, YNumCells, CV_32FC1);
		Mat grid_x       = Mat(XNumCells, YNumCells, CV_32SC1);
		Mat grid_y       = Mat(XNumCells, YNumCells, CV_32SC1);

		//For each cell in the coarse grid, save the best scoring feature location and score in that cell.
		int gridx, gridy;
		for (int x=(parameters::newFeatureEdgeRadius); x<(X-parameters::newFeatureEdgeRadius); x++) {
			for (int y=(parameters::newFeatureEdgeRadius); y<(Y-parameters::newFeatureEdgeRadius); y++) {
				gridx = x/cellSize;
				gridy = y/cellSize;
				if (cornerScoreMap.at<float>(x,y) > gridScoreMap.at<float>(gridx, gridy)) {
					gridScoreMap.at<float>(gridx, gridy) = cornerScoreMap.at<float>(x,y);
					grid_x.at<int32_t>(gridx, gridy) = x;
					grid_y.at<int32_t>(gridx, gridy) = y;
				}
			}
		}

		//Now reshape the matrices into linear buffers and sort the scores
		gridScoreMap = gridScoreMap.reshape(0, 1);
		grid_x       = grid_x.reshape(0, 1);
		grid_y       = grid_y.reshape(0, 1);
		Mat Indices;
		sortIdx(gridScoreMap, Indices, CV_SORT_EVERY_ROW | CV_SORT_DESCENDING);

		//Populate new feature list with best new candidates
		TopFeatureLocations.clear();
		for (int n=0; n<Utilities::min_int(gridScoreMap.cols, parameters::numberOfTopFeaturesThatMustBeTracked); n++) {
			int index = Indices.at<int32_t>(n);
			TopFeatureLocations.push_back(Vec2f(grid_x.at<int32_t>(index),grid_y.at<int32_t>(index)));
		}
	
		for (list< Vec2f >::iterator LocIter = TopFeatureLocations.begin(); LocIter != TopFeatureLocations.end(); LocIter++) {
			float distance = features.getDistanceFromTrackset(*LocIter);
			if (distance > parameters::identificationTolerance) {
				//We are not currently tracking this feature
				if (features.size() >= parameters::numberOfFeatures) {
					//Before adding new feature, throw out existing worst feature
				
					//First score all features and identify the worst current feature
					float featureScore, worstScore = 0.0;
					list< Feature >::iterator worstFeatureIter;
					
					for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
						int x = (int)round(featureIter->CurrentLocation[0]);
						int y = (int)round(featureIter->CurrentLocation[1]);
						if (x < 0) x=0;
						if (x > X-1) x=X-1;
						if (y < 0) y=0;
						if (y > Y-1) y=Y-1;
						
						if (featureIter == features.Features.begin()) {
							worstScore = cornerScoreMap.at<float>(x,y);
							worstFeatureIter = featureIter;
						}
						else {
							featureScore = cornerScoreMap.at<float>(x,y);
							if (featureScore < worstScore) {
								worstScore = featureScore;
								worstFeatureIter = featureIter;
							}
						}
					}
				
					//Remove the feature with the lowest score
					deadFeatures.addFeature(*worstFeatureIter);
					features.removeFeature(worstFeatureIter);
				}
			
				//Add new feature to trackset
				Feature newFeature;
				newFeature.initializeFeature(frameNumber, thisFrameStack, *(LocIter), lastIDAssigned++);
				features.addFeature(newFeature);
			}
		}
	}
}

void Tracker::reviewFeatureSet_mode1(void) {
	//In this mode, we simply throw out features that get too close to leaving the field of view
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		//Check to see if the feature is too close to an image edge
		double x = featureIter->CurrentLocation[0];
		double y = featureIter->CurrentLocation[1];
		if ((x < parameters::oldFeatureEdgeRadius) || (x >= (double)X - parameters::oldFeatureEdgeRadius) ||
		    (y < parameters::oldFeatureEdgeRadius) || (y >= (double)Y - parameters::oldFeatureEdgeRadius)) {
			//Feature needs to be removed
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
}

void Tracker::reviewFeatureSet_mode2(void) {
	//In this mode, we see if there are eny new features that need to be initialized and we initialize them. Then, we
	//go through our feature set and re-initialize those features that have drifted too far from ground truth and we
	//kill those features that are supposed to have terminated (according to ground truth).
	
	//First, add features that are supposed to be born this frame
	if (frameNumber < ((int) groundTruth_Trajectories.size())) {
		list<list<Vec2f>> newTrajectories = groundTruth_Trajectories[frameNumber];
		for (list<list<Vec2f>>::iterator it = newTrajectories.begin(); it != newTrajectories.end(); it++) {
			list<Vec2f> newTrajectory = *it;
			
			Feature newFeature;
			newFeature.initializeFeature(frameNumber, thisFrameStack, newTrajectory.front(), lastIDAssigned++);
			newFeature.loadGroundTruth(newTrajectory, frameNumber);
			features.addFeature(newFeature);
		}
	}
	
	//Next, kill any features than are not supposed to exist anymore
	list< Feature >::iterator featureIter = features.Features.begin();
	while (featureIter != features.Features.end()) {
		if (!featureIter->groundTruth_Alive()) {
			deadFeatures.addFeature(*featureIter);
			features.removeFeature(featureIter);
		}
		else
			featureIter++;
	}
	
	//Re-initialize featureas that have wandered too far from ground truth
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		if (featureIter->getDeviationFromGroundTruth() > featureReinitTolerance) {
			//Re-initialize feature - including updating the template
			featureIter->CurrentLocation = featureIter->groundTruth_Location(featureIter->CurrentFrameNumber);
			
			featureIter->Template.BuildFromStack(thisFrameStack, featureIter->CurrentLocation, parameters::n);
			featureIter->RecentTemplateCandidates.clear();
			featureIter->RecentTemplateCandidates.push_front(featureIter->Template);
			
			//Record this re-initialization event
			reinitializationCounter++;
		}
	}
}


void Tracker::trackToNextFrame_KLT(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		Mat prev_grey, current_grey;
		if (thisFrameStack.Images[0].type() == CV_8UC3) {
			cvtColor(lastFrameStack.Images[0], prev_grey,    CV_RGB2GRAY);
			cvtColor(thisFrameStack.Images[0], current_grey, CV_RGB2GRAY);
		}
		else {
			prev_grey    = lastFrameStack.Images[0];
			current_grey = thisFrameStack.Images[0];
		}
		vector<Point2f> inputFeatureCoords;
		inputFeatureCoords.resize(features.size());
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			inputFeatureCoords[f].y = featureIter->CurrentLocation[0];
			inputFeatureCoords[f].x = featureIter->CurrentLocation[1];
			f++;
		}
		vector<Point2f> outputFeatureCoords;
		vector<uchar> status;
		vector<float> error;
		calcOpticalFlowPyrLK(prev_grey, current_grey, inputFeatureCoords, outputFeatureCoords, status, error, Size(parameters::n, parameters::n), parameters::M);
		f = 0;
		Vec2f location;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(outputFeatureCoords[f].y, outputFeatureCoords[f].x);
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_KLT_HomInit(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		Mat prev_grey, current_grey;
		if (thisFrameStack.Images[0].type() == CV_8UC3) {
			cvtColor(lastFrameStack.Images[0], prev_grey,    CV_RGB2GRAY);
			cvtColor(thisFrameStack.Images[0], current_grey, CV_RGB2GRAY);
		}
		else {
			prev_grey    = lastFrameStack.Images[0];
			current_grey = thisFrameStack.Images[0];
		}
		vector<Point2f> inputFeatureCoords, outputFeatureCoords;
		inputFeatureCoords.resize(features.size());
		outputFeatureCoords.resize(features.size());
		Vector3d point1Hom, point2Hom;
		point1Hom(2) = 1.0;
		Point2f location_pt;
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location_pt.x = featureIter->CurrentLocation[1];
			location_pt.y = featureIter->CurrentLocation[0];
			Utilities::normalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
			point1Hom(0) = location_pt.x;
			point1Hom(1) = location_pt.y;
	
			point2Hom = H*point1Hom;
			location_pt.x = point2Hom(0)/point2Hom(2);
			location_pt.y = point2Hom(1)/point2Hom(2);
			Utilities::denormalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
			//inputFeatureCoords holds last positions of points. outputFeatureCoords will hold predicted target locations
			inputFeatureCoords[f].y = featureIter->CurrentLocation[0];
			inputFeatureCoords[f].x = featureIter->CurrentLocation[1];
		
			outputFeatureCoords[f].x = location_pt.x;
			outputFeatureCoords[f].y = location_pt.y;
			f++;
		}
		vector<uchar> status;
		vector<float> error;
		calcOpticalFlowPyrLK(prev_grey, current_grey, inputFeatureCoords, outputFeatureCoords, status, error, Size(parameters::n, parameters::n), parameters::M, TermCriteria(TermCriteria::COUNT+TermCriteria::EPS, 30, 0.01), OPTFLOW_USE_INITIAL_FLOW);
		f = 0;
		Vec2f location;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(outputFeatureCoords[f].y, outputFeatureCoords[f].x);
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location = trackSingleFeature_1stOrderDescent(*featureIter, nextFrameStack, featureIter->CurrentLocation);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent_HomInit(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vector3d point1Hom, point2Hom;
	point1Hom(2) = 1.0;
	Point2f location_pt;
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location_pt.x = featureIter->CurrentLocation[1];
		location_pt.y = featureIter->CurrentLocation[0];
		Utilities::normalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		point1Hom(0) = location_pt.x;
		point1Hom(1) = location_pt.y;
	
		point2Hom = H*point1Hom;
		location_pt.x = point2Hom(0)/point2Hom(2);
		location_pt.y = point2Hom(1)/point2Hom(2);
		Utilities::denormalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		location = trackSingleFeature_1stOrderDescent(*featureIter, nextFrameStack, Vec2f(location_pt.y, location_pt.x));
		//location = Vec2f(location_pt.y, location_pt.x);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_KLT_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		//Estimate Global Motion
		float globalMotion_x = 0.0;
		float globalMotion_y = 0.0;
		estimateGlobalMotion(globalMotion_x, globalMotion_y);
		Vec2f globalMotion(globalMotion_x, globalMotion_y);
	
		Mat prev_grey, current_grey;
		if (thisFrameStack.Images[0].type() == CV_8UC3) {
			cvtColor(lastFrameStack.Images[0], prev_grey,    CV_RGB2GRAY);
			cvtColor(thisFrameStack.Images[0], current_grey, CV_RGB2GRAY);
		}
		else {
			prev_grey    = lastFrameStack.Images[0];
			current_grey = thisFrameStack.Images[0];
		}
		vector<Point2f> inputFeatureCoords, outputFeatureCoords;
		inputFeatureCoords.resize(features.size());
		outputFeatureCoords.resize(features.size());
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			inputFeatureCoords[f].y = featureIter->CurrentLocation[0];
			inputFeatureCoords[f].x = featureIter->CurrentLocation[1];
			outputFeatureCoords[f].y = featureIter->CurrentLocation[0] + globalMotion_x;
			outputFeatureCoords[f].x = featureIter->CurrentLocation[1] + globalMotion_y;
			f++;
		}
		vector<uchar> status;
		vector<float> error;
		calcOpticalFlowPyrLK(prev_grey, current_grey, inputFeatureCoords, outputFeatureCoords, status, error, Size(parameters::n, parameters::n), parameters::M, TermCriteria(TermCriteria::COUNT+TermCriteria::EPS, 30, 0.01), OPTFLOW_USE_INITIAL_FLOW);
		f = 0;
		Vec2f location;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(outputFeatureCoords[f].y, outputFeatureCoords[f].x);
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Estimate Global Motion
	float globalMotion_x = 0.0;
	float globalMotion_y = 0.0;
	estimateGlobalMotion(globalMotion_x, globalMotion_y);
	Vec2f globalMotion(globalMotion_x, globalMotion_y);
	
	//Track each feature and update as you go
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location = trackSingleFeature_1stOrderDescent(*featureIter, nextFrameStack, featureIter->CurrentLocation + globalMotion);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_GyroOptical_Experimental_1(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vector3d point1Hom, point2Hom;
	point1Hom(2) = 1.0;
	double SearchRadius = 5.0;
	Point2f location_pt;
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location_pt.x = featureIter->CurrentLocation[1];
		location_pt.y = featureIter->CurrentLocation[0];
		Utilities::normalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		point1Hom(0) = location_pt.x;
		point1Hom(1) = location_pt.y;
	
		point2Hom = H*point1Hom;
		location_pt.x = point2Hom(0)/point2Hom(2);
		location_pt.y = point2Hom(1)/point2Hom(2);
		Utilities::denormalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		location = trackSingleFeatureGyroOptical_Experimental_1(*featureIter, nextFrameStack, Vec2f(location_pt.y, location_pt.x), SearchRadius);
		//location = Vec2f(location_pt.y, location_pt.x);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}	
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent_Blended(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location = trackSingleFeature_1stOrderDescent_BlendedEnergy(*featureIter, nextFrameStack, featureIter->CurrentLocation);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_1stOrderDescent_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Matrix3d H) {
	double lambda = 0.0001;
	trackToNextFrame_1stOrderDescent_GyroPrior(nextFrameNumber, nextFrameStack, H, lambda);
}

void Tracker::trackToNextFrame_1stOrderDescent_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Matrix3d H, double lambda) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	//Track each feature and update as you go
	Vector3d point1Hom, point2Hom;
	point1Hom(2) = 1.0;
	Point2f location_pt;
	Vec2f location;
	int f = 0;
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		location_pt.x = featureIter->CurrentLocation[1];
		location_pt.y = featureIter->CurrentLocation[0];
		Utilities::normalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		point1Hom(0) = location_pt.x;
		point1Hom(1) = location_pt.y;
	
		point2Hom = H*point1Hom;
		location_pt.x = point2Hom(0)/point2Hom(2);
		location_pt.y = point2Hom(1)/point2Hom(2);
		Utilities::denormalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
		location = trackSingleFeature_1stOrderDescent_regularized(*featureIter, nextFrameStack, Vec2f(location_pt.y, location_pt.x), lambda);
		//location = Vec2f(location_pt.y, location_pt.x);
		featureIter->Update(frameNumber, thisFrameStack, location);
		f++;
	}
	reviewFeatureSet();
}

void Tracker::trackToNextFrame_multiTrack_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H) {
	double lambda = 0.0001;
	trackToNextFrame_multiTrack_GyroPrior(nextFrameNumber, nextFrameStack, H, lambda);
}

void Tracker::trackToNextFrame_multiTrack_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H, double lambda) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		//Generate history matrix
		Mat History;
		features.build_HistoryMatrix(History, multiTrackerParametersPtr->L);

		//Generate xbar prior using gyro-predicted homography
		Mat xbar(2*features.size(),1,CV_32FC1);
		int f=0;
		Vector3d point1Hom, point2Hom;
		point1Hom(2) = 1.0;
		Point2f location_pt;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location_pt.x = featureIter->CurrentLocation[1];
			location_pt.y = featureIter->CurrentLocation[0];
			Utilities::normalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
		
			point1Hom(0) = location_pt.x;
			point1Hom(1) = location_pt.y;
	
			point2Hom = H*point1Hom;
			location_pt.x = point2Hom(0)/point2Hom(2);
			location_pt.y = point2Hom(1)/point2Hom(2);
			Utilities::denormalizeImageCoords(nextFrameStack.Images[0].cols, nextFrameStack.Images[0].rows, location_pt);
			
			xbar.at<float>(2*f)   = location_pt.y;
			xbar.at<float>(2*f+1) = location_pt.x;
			f++;
		}
		Mat xbar_prior;
		xbar.copyTo(xbar_prior);

		//Track on coarsest resolution and refine on higher-res images
		for (int m=(parameters::M-1); m>=0; m--)
			ConstrainedMultiTrackerWithPrior_TrackOnSingleLevel(xbar, xbar_prior, thisFrameStack, m, History, lambda);

		//xbar now holds best tracker state at resolution level 0. Update features with new locations
		Vec2f location;
		f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(xbar.at<float>(2*f) , xbar.at<float>(2*f+1));
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}





void Tracker::trackToNextFrame_multiTrack_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack) {
	//First, update frame stacks and frame number
	updateFrameNumberAndStacks(nextFrameNumber, nextFrameStack);
	
	if (features.size() > 0) {
		//Estimate Global Motion
		float globalMotion_x = 0.0;
		float globalMotion_y = 0.0;
		estimateGlobalMotion(globalMotion_x, globalMotion_y);
		//fprintf(stderr,"\nglobal motion: (%f, %f)\n", globalMotion_x, globalMotion_y);

		//Generate xbar and history matrix
		Mat xbar, History;
		features.build_xbar(xbar);
		features.build_HistoryMatrix(History, multiTrackerParametersPtr->L);

		//Add estimated global motion to xbar
		for (int f=0; f<features.size(); f++) {
			xbar.at<float>(2*f)   += globalMotion_x;
			xbar.at<float>(2*f+1) += globalMotion_y;
		}

		//Track on coarsest resolution and refine on higher-res images
		for (int m=(parameters::M-1); m>=0; m--)
			ConstrainedMultiTracker_TrackOnSingleLevel(xbar, thisFrameStack, m, History);

		//xbar now holds best tracker state at resolution level 0. Update features with new locations
		Vec2f location;
		int f = 0;
		for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			location = Vec2f(xbar.at<float>(2*f) , xbar.at<float>(2*f+1));
			featureIter->Update(frameNumber, thisFrameStack, location);
			f++;
		}
	}
	reviewFeatureSet();
}


void Tracker::updateFrameNumberAndStacks(int nextFrameNumber, ImageStack &nextFrameStack) {
	//Update the frame number and frame stacks. This is the first step of tracking to a new frame
	frameNumber = nextFrameNumber;
	lastFrameStack = thisFrameStack;
	thisFrameStack = nextFrameStack;
}


Vec2f Tracker::trackSingleFeature_1stOrderDescent(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate) {
	Vec2f state = initialLocationEstimate;
	Vec2f step, testLocation;
	double c = 0.0;
	double Dx = 0.0;
	double Dy = 0.0;
	double testa, testb, testc;
	double stepNorm = 0.0;
	double lastStepNorm = 0.0;
	int minSteps, maxSteps;
	
	for (int m=(parameters::M-1); m>=0; m--) {
		if (m == parameters::M-1) {
			minSteps = parameters::minStepsTopLevel;
			maxSteps = parameters::maxStepsTopLevel;
		}
		else {
			minSteps = parameters::minStepsLowerLevels;
			maxSteps = parameters::maxStepsLowerLevels;
		}
		
		
		for (int stepNumber = 0; stepNumber < maxSteps; stepNumber++) {
			computeSingleFeatureEnergy(f.Template, nextFrameStack, m, state, c, Dx, Dy);
			step = -1.0*Vec2f((float)Dx ,(float)Dy);
			stepNorm = norm(step);
			if ((stepNumber > 0) && (stepNumber >= minSteps) && (stepNorm > 0.9999*lastStepNorm))
				break;
			lastStepNorm = stepNorm;
			if (stepNorm < 0.00001)
				break;
			step = step/stepNorm;
			
			double scale = 2.0;
			
			int maxNumberOfTests = 100;
			int maxScaleRefinements = 10;
		
			int scaleRefinements = 0;
		
			double testaDist = 0.0;
			testa = c;
		
			double testbDist = 1.0*scale;
			testLocation = state + testbDist*step;
			testb = computeSingleFeatureEnergy(f.Template, nextFrameStack, m, testLocation);
		
			double testcDist = 2.0*scale;
			testLocation = state + testcDist*step;
			testc = computeSingleFeatureEnergy(f.Template, nextFrameStack, m, testLocation);
		
			for (int testNumber=0; testNumber<maxNumberOfTests; testNumber++) {
				if ((testa > testb) && (testb > testc)) {
					//Keep moving out on this scale
					testaDist = testbDist;
					testa = testb;
					testbDist = testcDist;
					testb = testc;
					testcDist += scale;
					testLocation = state + testcDist*step;
					testc = computeSingleFeatureEnergy(f.Template, nextFrameStack, m, testLocation);
				}
				else {
					//Shorten scale
					testcDist = testbDist;
					testc = testb;
					scale = (0.5)*scale;
					scaleRefinements ++;
					testbDist = testaDist + scale;
					testLocation = state + testbDist*step;
					testb = computeSingleFeatureEnergy(f.Template, nextFrameStack, m, testLocation);
				}
			
				//Test break conditions
				if (scaleRefinements >= maxScaleRefinements)
					break;
			}
		
			//Update state
			state = state + testbDist*step;
		}
	}
	return(state);
}


Vec2f Tracker::trackSingleFeature_1stOrderDescent_regularized(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate, double lambda) {
	Vec2f state = initialLocationEstimate;
	Vec2f step, testLocation;
	double c = 0.0;
	double Dx = 0.0;
	double Dy = 0.0;
	double testa, testb, testc;
	double stepNorm = 0.0;
	double lastStepNorm = 0.0;
	int minSteps, maxSteps;
	
	for (int m=(parameters::M-1); m>=0; m--) {
		if (m == parameters::M-1) {
			minSteps = parameters::minStepsTopLevel;
			maxSteps = parameters::maxStepsTopLevel;
		}
		else {
			minSteps = parameters::minStepsLowerLevels;
			maxSteps = parameters::maxStepsLowerLevels;
		}
		
		for (int stepNumber = 0; stepNumber < maxSteps; stepNumber++) {
			computeSingleFeatureEnergy_regularized(f.Template, nextFrameStack, m, state, initialLocationEstimate, lambda, c, Dx, Dy);
			step = -1.0*Vec2f((float)Dx ,(float)Dy);
			stepNorm = norm(step);
			if ((stepNumber > 0) && (stepNumber >= minSteps) && (stepNorm > 0.9999*lastStepNorm))
				break;
			lastStepNorm = stepNorm;
			if (stepNorm < 0.00001)
				break;
			step = step/stepNorm;
			
			double scale = 2.0;
			
			int maxNumberOfTests = 100;
			int maxScaleRefinements = 10;
		
			int scaleRefinements = 0;
		
			double testaDist = 0.0;
			testa = c;
		
			double testbDist = 1.0*scale;
			testLocation = state + testbDist*step;
			testb = computeSingleFeatureEnergy_regularized(f.Template, nextFrameStack, m, testLocation, initialLocationEstimate, lambda);
		
			double testcDist = 2.0*scale;
			testLocation = state + testcDist*step;
			testc = computeSingleFeatureEnergy_regularized(f.Template, nextFrameStack, m, testLocation, initialLocationEstimate, lambda);
		
			for (int testNumber=0; testNumber<maxNumberOfTests; testNumber++) {
				if ((testa > testb) && (testb > testc)) {
					//Keep moving out on this scale
					testaDist = testbDist;
					testa = testb;
					testbDist = testcDist;
					testb = testc;
					testcDist += scale;
					testLocation = state + testcDist*step;
					testc = computeSingleFeatureEnergy_regularized(f.Template, nextFrameStack, m, testLocation, initialLocationEstimate, lambda);
				}
				else {
					//Shorten scale
					testcDist = testbDist;
					testc = testb;
					scale = (0.5)*scale;
					scaleRefinements ++;
					testbDist = testaDist + scale;
					testLocation = state + testbDist*step;
					testb = computeSingleFeatureEnergy_regularized(f.Template, nextFrameStack, m, testLocation, initialLocationEstimate, lambda);
				}
			
				//Test break conditions
				if (scaleRefinements >= maxScaleRefinements)
					break;
			}
		
			//Update state
			state = state + testbDist*step;
		}
	}
	return(state);
}


Vec2f Tracker::trackSingleFeature_1stOrderDescent_BlendedEnergy(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate) {
	Vec2f location = initialLocationEstimate;
	const double blendingCoefficients1[] = {0.0, 0.0, 0.0, 1.0};
	//const double blendingCoefficients2[] = {0.0, 1.0, 1.0, 1.0};
	const double blendingCoefficients3[] = {1.0, 0.0, 0.0, 2.0};
	int minSteps, maxSteps;
	
	minSteps = 10;
	maxSteps = 30;
	location = trackSingleFeature_1stOrderDescent_SingleBlendedEnergy(f, nextFrameStack, location, blendingCoefficients1, minSteps, maxSteps);
	
	minSteps = 5;
	maxSteps = 15;
	location = trackSingleFeature_1stOrderDescent_SingleBlendedEnergy(f, nextFrameStack, location, blendingCoefficients3, minSteps, maxSteps);
	
	return(location);
}

Vec2f Tracker::trackSingleFeature_1stOrderDescent_SingleBlendedEnergy(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate, const double blendingCoefficients[], int minSteps, int maxSteps) {
	Vec2f state = initialLocationEstimate;
	Vec2f step, testLocation;
	double c = 0.0;
	double Dx = 0.0;
	double Dy = 0.0;
	double testa, testb, testc;
	double stepNorm = 0.0;
	double lastStepNorm = 0.0;
	
	for (int stepNumber = 0; stepNumber < maxSteps; stepNumber++) {
		computeBlendedEnergy(f.Template, nextFrameStack, state, blendingCoefficients, c, Dx, Dy);
		step = -1.0*Vec2f((float)Dx ,(float)Dy);
		stepNorm = norm(step);
		if ((stepNumber > 0) && (stepNumber >= minSteps) && (stepNorm > 0.9999*lastStepNorm))
			break;
		lastStepNorm = stepNorm;
		if (stepNorm < 0.00001)
			break;
		step = step/stepNorm;
		
		double scale = 2.0;
		
		int maxNumberOfTests = 100;
		int maxScaleRefinements = 10;
	
		int scaleRefinements = 0;
	
		double testaDist = 0.0;
		testa = c;
	
		double testbDist = 1.0*scale;
		testLocation = state + testbDist*step;
		
		testb = computeBlendedEnergy(f.Template, nextFrameStack, testLocation, blendingCoefficients);
	
		double testcDist = 2.0*scale;
		testLocation = state + testcDist*step;
		testc = computeBlendedEnergy(f.Template, nextFrameStack, testLocation, blendingCoefficients);
	
		for (int testNumber=0; testNumber<maxNumberOfTests; testNumber++) {
			if ((testa > testb) && (testb > testc)) {
				//Keep moving out on this scale
				testaDist = testbDist;
				testa = testb;
				testbDist = testcDist;
				testb = testc;
				testcDist += scale;
				testLocation = state + testcDist*step;
				testc = computeBlendedEnergy(f.Template, nextFrameStack, testLocation, blendingCoefficients);
			}
			else {
				//Shorten scale
				testcDist = testbDist;
				testc = testb;
				scale = (0.5)*scale;
				scaleRefinements ++;
				testbDist = testaDist + scale;
				testLocation = state + testbDist*step;
				testb = computeBlendedEnergy(f.Template, nextFrameStack, testLocation, blendingCoefficients);
			}
		
			//Test break conditions
			if (scaleRefinements >= maxScaleRefinements)
				break;
		}
	
		//Update state
		state = state + testbDist*step;
	}
	
	return(state);
}


//Thoughts: Fitting a quadratic surface will only work when the search space is very small. Start with a search in the larger search space.
//then when we have the min nailed down to approximately 1 pixel, fit the quadratic and snap to the arg min.

//In directions that lack distinctiveness in the energy function, we can ride the gyro-predicted flow. This is not possible without a gyro.
Vec2f Tracker::trackSingleFeatureGyroOptical_Experimental_1(Feature& f, ImageStack& nextFrameStack, Vec2f initialLocationEstimate, double SearchRadius) {
	//This function uses a flow estimate from a gyro and a known search radius to find the given feature in a new frame
	Vec2f center = initialLocationEstimate;
	
	const double blendingCoefficients1[] = {0.0, 0.0, 1.0};
	const double blendingCoefficients2[] = {0.0, 1.0, 0.25};
	const double blendingCoefficients3[] = {1.0, 0.25, 0.0625};
	double lambda = 0.001;
	center = SearchForFeature_Blended(f, nextFrameStack, center, SearchRadius, 2.0,  lambda, blendingCoefficients1);
	center = SearchForFeature_Blended(f, nextFrameStack, center, 2.9,          1.0,  lambda, blendingCoefficients2);
	center = SearchForFeature_Blended(f, nextFrameStack, center, 1.6,          1.0,  lambda, blendingCoefficients2);
	center = SearchForFeature_Blended(f, nextFrameStack, center, 1.6,          0.5,  lambda, blendingCoefficients3);
	center = SearchForFeature_Blended(f, nextFrameStack, center, 0.85,         0.25, lambda, blendingCoefficients3);
	
	return(center);
}

Vec2f Tracker::SearchForFeature_Blended(Feature& f, ImageStack& nextFrameStack, Vec2f center, double SearchRadius, double meshSize, double lambda, const double blendingCoefficients[]) {
	//This function searches a circular window for a given feature (using the energy from a single resolution level). It samples at points on a
	//square grid, intersected with a circle centered at "center" with radius "SearchRadius". Sample points are spaced "meshSize" pixels apart.
	//The energy function is added to "lambda"*||x - center||. Thus, if lambda is high, there is a tendency to ride gyro measurements.
	int numSamples1D = 2*((int)(SearchRadius/meshSize)) + 1;
	
	Vec2f sampleLoc;
	double sampleEnergy;
	Vec2f corner = center - Vec2f(floor(SearchRadius/meshSize)*meshSize, floor(SearchRadius/meshSize)*meshSize);
	Vec2f bestSampleLoc = corner;
	double lowestEnergy = INFINITY;
	double distance;
	for (int i=0; i<numSamples1D; i++) {
		for (int j=0; j<numSamples1D; j++) {
			sampleLoc = corner + Vec2f(i*meshSize, j*meshSize);
			distance = norm(sampleLoc - center);
			if (distance <= SearchRadius) {
				//sampleEnergy = compute_energy1Level(f.Template, nextFrameStack, sampleLoc, resolutionLevel) + lambda*distance;
				sampleEnergy = computeBlendedEnergy(f.Template, nextFrameStack, sampleLoc, blendingCoefficients) + lambda*distance;
				//fprintf(stderr,"%f\n\r",sampleEnergy);
				if (sampleEnergy < lowestEnergy) {
					lowestEnergy = sampleEnergy;
					bestSampleLoc = sampleLoc;
				}
			}
		}
	}
	return(bestSampleLoc);
}


Vec2f Tracker::getArgMinFromCircularSamples(Feature& f, ImageStack& nextFrameStack, Vec2f center, double radius) {
	//This function creates a search pattern consisting of 7 points: 6 in a circle around the center of the search area and one in the center.
	//The blended energy is evaluated at each point. A quadratic is fit to he surface and the argument of the minimum is returned.
	Vec2f d1( 1.000000000000000,  0.000000000000000);
	Vec2f d2( 0.500000000000000,  0.866025403784439);
	Vec2f d3(-0.500000000000000,  0.866025403784439);
	Vec2f d4(-1.000000000000000,  0.000000000000000);
	Vec2f d5(-0.500000000000000, -0.866025403784439);
	Vec2f d6( 0.500000000000000, -0.866025403784439);
	
	const double blendingCoefficients[] = {1.0, 0.5, 0.25};
	
	vector<Vec2f> points;
	points.resize(7);
	points[0] = Vec2f(0.0, 0.0);
	points[1] = radius*d1;
	points[2] = radius*d2;
	points[3] = radius*d3;
	points[4] = radius*d4;
	points[5] = radius*d5;
	points[6] = radius*d6;
	
	MatrixXd A(7,6);
	VectorXd v(6);
	VectorXd b(7);
	
	double x, y;
	for (int i=0; i<7; i++) {
		x = points[i][0];
		y = points[i][1];
		
		A(i,0) = x*x;
		A(i,1) = y*y;
		A(i,2) = x*y;
		A(i,3) = x;
		A(i,4) = y;
		A(i,5) = 1.0;
		
		b(i) = computeBlendedEnergy(f.Template, nextFrameStack, center + points[i], blendingCoefficients);
	}
	MatrixXd At = A.transpose();
	v = (At*A).inverse()*At*b;
	
	Matrix2d B;
	Vector2d u, disp;
	
	B(0,0) = 2*v(0);
	B(0,1) = v(2);
	B(1,0) = v(2);
	B(1,1) = 2*v(1);
	
	u(0) = v(3);
	u(1) = v(4);
	
	//fprintf(stderr,"\n\rB: %f %f %f %f\n\r", B(0,0), B(0,1), B(1,0), B(1,1));
	
	double dispLength;
	if (abs(B.determinant()) > 0.0001) {
		disp = -1.0*B.inverse()*u;
		
		//If displacement exceeds search radius, project into search zone.
		dispLength = disp.norm();
		if (dispLength > radius)
			disp = (radius/dispLength)*disp;
	
		Vec2f argmin;
		argmin[0] = center[0] + disp(0);
		argmin[1] = center[1] + disp(1);
		
		//fprintf(stderr,"\n\rb: %f %f %f %f %f %f %f\n\r", b(0), b(1), b(2), b(3), b(4), b(5), b(6));
		//fprintf(stderr,"displacement: %f %f\n\r", disp(0), disp(1));
		
		return(argmin);
	}
	else {
		fprintf(stderr,"Skipping\n\r");
		return(center);
	}
}


void Tracker::ConstrainedMultiTracker_TrackOnSingleLevel(Mat& xbar, ImageStack& frameStack, int resolutionLevel, Mat& History) {
	//The inputs xbar and History must be in image coordinates at resolution level 0. xbar will be updated on completion of this function.
	//The updated xbar will also be in image coordinates at resolution level 0.
	int F = features.size();
	
	//Iteratively Descend Energy Surface
	double lastDcNorm = 10000.0;
	double DcNorm = 0.0;
	double c, gradDirNorm;
	Mat Dc, step, stepDouble, gradDir, equalizedGradDir, testLocation;
	Mat lastStepDir = Mat::zeros(2*F,1,CV_64FC1);
	int maxStepsThisLevel, minStepsThisLevel;
	if (resolutionLevel == parameters::M-1) {
		maxStepsThisLevel = parameters::maxStepsTopLevel;
		minStepsThisLevel = parameters::minStepsTopLevel;
	}
	else {
		maxStepsThisLevel = parameters::maxStepsLowerLevels;
		minStepsThisLevel = parameters::minStepsLowerLevels;
	}
	for (int stepNumber = 0; stepNumber < maxStepsThisLevel; stepNumber++) {
		compute_c(xbar, features, frameStack, resolutionLevel, History, c, Dc, multiTrackerParametersPtr);
		
		//Compute grad descent step direction
		gradDir = -1.0*Dc.t();
		gradDirNorm = norm(gradDir);
		if (gradDirNorm > 0.001)
			gradDir /= gradDirNorm;
		
		//Copute equalized gradient descent step direction
		gradDir.copyTo(equalizedGradDir);
		for (int f=0; f<F; f++) {
			double subNorm = sqrt(pow(equalizedGradDir.at<double>(2*f), 2.0) + pow(equalizedGradDir.at<double>(2*f+1), 2.0));
			if (subNorm > 0.001) {
				equalizedGradDir.at<double>(2*f) /= sqrt(subNorm);
				equalizedGradDir.at<double>(2*f+1) /= sqrt(subNorm);
			}
		}
		
		//Compute step direction
		stepDouble = 0.5*gradDir + 0.5*equalizedGradDir;
		stepDouble.convertTo(step, CV_32F, 1.0);
		//fprintf(stderr,"\nDir size: %f", norm(step));
		
		
		//Test early exit conditions
		DcNorm = norm(Dc);
		if (stepNumber >= minStepsThisLevel) {
			if (DcNorm > 1.0*lastDcNorm) {
				//fprintf(stderr,"\nTerminating because |DC| is not dropping anymore (level %d). Ran %d steps.", resolutionLevel, stepNumber);
				//if (resolutionLevel == 0) fprintf(stderr,"\n");
				break;
			}
			else if (DcNorm < parameters::DerivativeTolerance) {
				//fprintf(stderr,"\nTerminating because |DC| < DerivativeTolerance (level %d). Ran %d steps.", resolutionLevel, stepNumber);
				//if (resolutionLevel == 0) fprintf(stderr,"\n");
				break;
			}
		}
		lastDcNorm = DcNorm;
		
		//Line search in step direction
		double scale = 1.0;
		double maxScaleRefinements = 10;
		double maxTests = 100;
		ConstrainedMultiTracker_LineSearch(xbar, c, step, scale*pow(2,resolutionLevel), maxTests, maxScaleRefinements, frameStack, resolutionLevel, History);
		
	}
}

void Tracker::ConstrainedMultiTrackerWithPrior_TrackOnSingleLevel(Mat& xbar, Mat& xbar_prior, ImageStack& frameStack, int resolutionLevel, Mat& History, double lambda) {
	//The inputs xbar and History must be in image coordinates at resolution level 0. xbar will be updated on completion of this function.
	//The updated xbar will also be in image coordinates at resolution level 0.
	int F = features.size();
	
	//Iteratively Descend Energy Surface
	double lastDcNorm = 10000.0;
	double DcNorm = 0.0;
	double c, gradDirNorm;
	Mat Dc, step, stepDouble, gradDir, equalizedGradDir, testLocation;
	Mat lastStepDir = Mat::zeros(2*F,1,CV_64FC1);
	int maxStepsThisLevel, minStepsThisLevel;
	if (resolutionLevel == parameters::M-1) {
		maxStepsThisLevel = parameters::maxStepsTopLevel;
		minStepsThisLevel = parameters::minStepsTopLevel;
	}
	else {
		maxStepsThisLevel = parameters::maxStepsLowerLevels;
		minStepsThisLevel = parameters::minStepsLowerLevels;
	}
	for (int stepNumber = 0; stepNumber < maxStepsThisLevel; stepNumber++) {
		compute_c_withPrior(xbar, xbar_prior, features, frameStack, resolutionLevel, History, c, Dc, multiTrackerParametersPtr, lambda);
		
		//Compute grad descent step direction
		gradDir = -1.0*Dc.t();
		gradDirNorm = norm(gradDir);
		if (gradDirNorm > 0.001)
			gradDir /= gradDirNorm;
		
		//Copute equalized gradient descent step direction
		gradDir.copyTo(equalizedGradDir);
		for (int f=0; f<F; f++) {
			double subNorm = sqrt(pow(equalizedGradDir.at<double>(2*f), 2.0) + pow(equalizedGradDir.at<double>(2*f+1), 2.0));
			if (subNorm > 0.001) {
				equalizedGradDir.at<double>(2*f) /= sqrt(subNorm);
				equalizedGradDir.at<double>(2*f+1) /= sqrt(subNorm);
			}
		}
		
		//Compute step direction
		stepDouble = 0.5*gradDir + 0.5*equalizedGradDir;
		stepDouble.convertTo(step, CV_32F, 1.0);
		//fprintf(stderr,"\nDir size: %f", norm(step));
		
		
		//Test early exit conditions
		DcNorm = norm(Dc);
		if (stepNumber >= minStepsThisLevel) {
			if (DcNorm > 1.0*lastDcNorm) {
				//fprintf(stderr,"\nTerminating because |DC| is not dropping anymore (level %d). Ran %d steps.", resolutionLevel, stepNumber);
				//if (resolutionLevel == 0) fprintf(stderr,"\n");
				break;
			}
			else if (DcNorm < parameters::DerivativeTolerance) {
				//fprintf(stderr,"\nTerminating because |DC| < DerivativeTolerance (level %d). Ran %d steps.", resolutionLevel, stepNumber);
				//if (resolutionLevel == 0) fprintf(stderr,"\n");
				break;
			}
		}
		lastDcNorm = DcNorm;
		
		//Line search in step direction
		double scale = 1.0;
		double maxScaleRefinements = 10;
		double maxTests = 100;
		ConstrainedMultiTracker_LineSearch(xbar, c, step, scale*pow(2,resolutionLevel), maxTests, maxScaleRefinements, frameStack, resolutionLevel, History);
	}
}

void Tracker::ConstrainedMultiTracker_LineSearch(Mat& xbar, double &c, Mat &step, double scale, int maxTests, int maxScaleRefinements, ImageStack& frameStack, int resolutionLevel, Mat& History) {
	//Perform precision line search
	Mat testLocation;
	int scaleRefinements = 0;

	double testaDist = 0.0;
	double testa = c;

	double testbDist = 1.0*scale;
	testLocation = xbar + testbDist*step;
	double testb = compute_c(testLocation, features, frameStack, resolutionLevel, History, multiTrackerParametersPtr);

	double testcDist = 2.0*scale;
	testLocation = xbar + testcDist*step;
	double testc = compute_c(testLocation, features, frameStack, resolutionLevel, History, multiTrackerParametersPtr);

	for (int testNumber=0; testNumber<maxTests; testNumber++) {
		if ((testa > testb) && (testb > testc)) {
			//Keep moving out on this scale
			testaDist = testbDist;
			testa = testb;
			testbDist = testcDist;
			testb = testc;
			testcDist += scale;
			testLocation = xbar + testcDist*step;
			testc = compute_c(testLocation, features, frameStack, resolutionLevel, History, multiTrackerParametersPtr);
		}
		else {
			//Shorten scale
			testcDist = testbDist;
			testc = testb;
			scale = (0.5)*scale;
			scaleRefinements ++;
			testbDist = testaDist + scale;
			testLocation = xbar + testbDist*step;
			testb = compute_c(testLocation, features, frameStack, resolutionLevel, History, multiTrackerParametersPtr);
		}
	
		//Test break conditions
		if (scaleRefinements >= maxScaleRefinements)
			break;
	}

	//Update state
	xbar += testbDist*step;
	c = testb;
}

void Tracker::estimateGlobalMotion(float& delta_x, float& delta_y) {
	//Track on coarsest resolution and refine on 1 higher-resolution version
	int finestLevel = parameters::M-2;
	if (finestLevel < 0)
		finestLevel = 0;
	
	Vec2f initialLocation, initialLocation_levelm, location;
	initialLocation[0] = ((float)(thisFrameStack.Images[0].rows - 1))/2.0;
	initialLocation[1] = ((float)(thisFrameStack.Images[0].cols - 1))/2.0;
	location[0] = initialLocation[0];
	location[1] = initialLocation[1];
	
	for (int m=(parameters::M-1); m >= finestLevel; m--) {
		float temp1 = (float)pow(2.0, (double)m);
		float temp2 = 1.0/temp1;
		initialLocation_levelm = temp2*initialLocation;
		
		double residual, Dx, Dy, stepNorm;
		Vec2f step;
		int n = min2i(thisFrameStack.Images[m].cols-10, thisFrameStack.Images[m].rows-10);
		ImageStack Template;
		Template.Images.resize(parameters::M);
		getRectSubPix(lastFrameStack.Images[m], Size(n,n), Point2f(initialLocation_levelm[1], initialLocation_levelm[0]), Template.Images[m]);
		
		int numsteps;
		if (m == parameters::M-1)
			numsteps = 50;
		else
			numsteps = 10;
		
		double minStep = 0.25*pow(2,m);
		for (int stepNumber = 0; stepNumber < numsteps; stepNumber++) {
			//Gradient Decent step
			compute_fitTerm(Template, thisFrameStack, m, location, n, residual, Dx, Dy);
			stepNorm = sqrt(Dx*Dx + Dy*Dy);
			if (stepNorm < minStep) {
				Dx *= minStep/stepNorm;
				Dy *= minStep/stepNorm;
			}
			location = location - Vec2f(Dx, Dy);
		}
		
		//fprintf(stderr,"\nGlobal Motion from level %d: (%f, %f)", m, location[1]-initialLocation[1], -1.0*(location[0]-initialLocation[0]));
	}
	
	delta_x = location[0] - initialLocation[0];
	delta_y = location[1] - initialLocation[1];
}

int Tracker::min2i(int a, int b) {
	if (a <= b)
		return(a);
	else
		return(b);
}

void Tracker::findNewFeatures(int numberOfNewFeatures, Mat& cornerScoreMap, list<Vec2f>& newFeatureLocations) {
	//fprintf(stderr,"\nEntering findNewFeatures function.");
	
	int X = thisFrameStack.Images[0].rows;
	int Y = thisFrameStack.Images[0].cols;
	int cellSize = (X + Y)/50;
	
	int XNumCells, YNumCells;
	if (X % cellSize == 0)
		XNumCells = X/cellSize;
	else
		XNumCells = X/cellSize + 1;
	if (Y % cellSize == 0)
		YNumCells = Y/cellSize;
	else
		YNumCells = Y/cellSize + 1;
	Mat gridScoreMap = Mat::zeros(XNumCells, YNumCells, CV_32FC1);
	Mat grid_x       = Mat(XNumCells, YNumCells, CV_32SC1);
	Mat grid_y       = Mat(XNumCells, YNumCells, CV_32SC1);
	Mat grid_Hit     = Mat::zeros(XNumCells, YNumCells, CV_8UC1);
	
	//For each cell in the coarse grid, save the best scoring feature location and score in that cell.
	int gridx, gridy;
	for (int x=(parameters::newFeatureEdgeRadius); x<(X-parameters::newFeatureEdgeRadius); x++) {
		for (int y=(parameters::newFeatureEdgeRadius); y<(Y-parameters::newFeatureEdgeRadius); y++) {
			gridx = x/cellSize;
			gridy = y/cellSize;
			if (cornerScoreMap.at<float>(x,y) > gridScoreMap.at<float>(gridx, gridy)) {
				gridScoreMap.at<float>(gridx, gridy) = cornerScoreMap.at<float>(x,y);
				grid_x.at<int32_t>(gridx, gridy) = x;
				grid_y.at<int32_t>(gridx, gridy) = y;
				grid_Hit.at<uchar>(gridx, gridy) = 1U;
			}
		}
	}
	
	//Invalidate cells that contain existing features.
	for (list< Feature >::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
		//Mark the cell holding the current feature as containing no new features
		gridx = featureIter->CurrentLocation[0]/cellSize;
		gridy = featureIter->CurrentLocation[1]/cellSize;
		
		grid_Hit.at<uchar>(Utilities::clampInt(gridx-1,0,XNumCells-1), Utilities::clampInt(gridy-1,0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx-1,0,XNumCells-1), Utilities::clampInt(gridy,  0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx-1,0,XNumCells-1), Utilities::clampInt(gridy+1,0,YNumCells-1)) = 0U;
		
		grid_Hit.at<uchar>(Utilities::clampInt(gridx,  0,XNumCells-1), Utilities::clampInt(gridy-1,0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx,  0,XNumCells-1), Utilities::clampInt(gridy,  0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx,  0,XNumCells-1), Utilities::clampInt(gridy+1,0,YNumCells-1)) = 0U;
		
		grid_Hit.at<uchar>(Utilities::clampInt(gridx+1,0,XNumCells-1), Utilities::clampInt(gridy-1,0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx+1,0,XNumCells-1), Utilities::clampInt(gridy,  0,YNumCells-1)) = 0U;
		grid_Hit.at<uchar>(Utilities::clampInt(gridx+1,0,XNumCells-1), Utilities::clampInt(gridy+1,0,YNumCells-1)) = 0U;
	}
	
	//Now reshape the matrices into linear buffers and sort the scores
	gridScoreMap = gridScoreMap.reshape(0, 1);
	grid_x       = grid_x.reshape(0, 1);
	grid_y       = grid_y.reshape(0, 1);
	grid_Hit     = grid_Hit.reshape(0, 1);
	Mat Indices;
	sortIdx(gridScoreMap, Indices, CV_SORT_EVERY_ROW | CV_SORT_DESCENDING);
	
	//Populate new feature list with best new candidates
	newFeatureLocations.clear();
	int newFeaturesAdded = 0;
	for (int n=0; n<gridScoreMap.cols; n++) {
		int index = Indices.at<int32_t>(n);
		if (grid_Hit.at<uchar>(index) == 1U) {
			//There is a valid new feature here
			//fprintf(stderr,"Feature at location (%4d,%4d) origionated from cell %d.\n\r",grid_x.at<int32_t>(index),grid_y.at<int32_t>(index),n);
			newFeatureLocations.push_back(Vec2f(grid_x.at<int32_t>(index),grid_y.at<int32_t>(index)));
			newFeaturesAdded++;
			if (newFeaturesAdded >= numberOfNewFeatures)
				break;
		}
	}
}


