#pragma once
#include <eigen3/Eigen/Eigen>
#include "opencv2/opencv.hpp"
#include <list>
#include "ImageStack.h"
#include "Feature.h"

//The standard usage of a tracker object is to initialize with the "initialize" function. Then, iteratively call the desired trackToNextFrame_* function.
//Each of these functions is responsible for updating the trackers feature sets according to the trackers parameters.

//The "initialize" function will set up new features for the trackers feature set according to the trackers feature management mode.

struct MultiTrackerParameters{
	typedef enum {
		None,
		EmpiricalDimension,
		NuclearNorm,
		ExplicitFactorization
	}penaltyType;
	
	//Penalty form for constraied tracker
	penaltyType penaltyForm = EmpiricalDimension;
	
	//History length to use in penalty
	int L = 5;
	
	//Energy Function Parameters
	double epsilon = 0.6;
	bool   fitTermEnabled = true;
	double penaltyCoefficient = 0.5;
	bool   centerHistoryMatrix = false;
};

class Tracker {
	public:
		//Core data fields
		int frameNumber;
		FeatureSet features;     //Set of currently active features
		FeatureSet deadFeatures; //Set of features that have been terminated
		
		//Tracker parameters
		int featureManagementMode;
		//0: Detect and manage features autonomously. Replace features that leave field of view
		//1: Load initial feature locations from a file. Do not replace features that leave field of view
		//2: Load ground truth from file. When a feature leaves the field of view pre-maturely, or wanders too
		//   far from ground truth, the feature is re-initialized to the ground truth location. The number of
		//   re-initializations is recorded for each feature and can be used as a performance metric.
		
		//The following holds a pointer to a MultiTrackerParameters object for a multi-feature tracker. If not
		//using a multi-feature tracker, this should be set to NULL
		MultiTrackerParameters *multiTrackerParametersPtr = NULL;
		
		//Ground Truth feature trajectories are only used if featureManagementMode = 2.
		//groundTruth_Trajectories[i] contains the trajectories of all features born on frame i.
		//groundTruth_Trajectories[i] is a list of trajectories. A trajectory is actually a list of Vec2f locations.
		//groundTruth_Trajectories will generally have fewer items than the total number of frames. It only is as large
		//as is needed to reach the oldest birth frame number in the ground truth set.
		std::vector< std::list< std::list< cv::Vec2f > > > groundTruth_Trajectories;
		double featureReinitTolerance; //In mode 2, this holds the amount a feature can drift from ground truth before being reset. (Must be set prior to tracking in mode 2)
		int reinitializationCounter;   //In mode 2, this holds a running tally of the number of feature re-initializations.
		
		//Tracker initialization function. In addition to calling this function, if running in mode 2, featureReinitTolerance must be set prior to tracking.
		//If you plan on using a multi-feature tracker, you must also set multiTrackerParametersPtr.
		void initialize(int firstFrameNumber, ImageStack firstFrameStack, int FeatureManagementMode, std::string inputFilePath = std::string(""));
		
		//Independent feature trackers
		void trackToNextFrame_KLT(int nextFrameNumber, ImageStack nextFrameStack);
		void trackToNextFrame_1stOrderDescent(int nextFrameNumber, ImageStack nextFrameStack);
		
		void trackToNextFrame_KLT_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack);
		void trackToNextFrame_1stOrderDescent_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack);
		
		void trackToNextFrame_KLT_HomInit(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H);
		void trackToNextFrame_1stOrderDescent_HomInit(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H);
		
		void trackToNextFrame_GyroOptical_Experimental_1(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H);
		void trackToNextFrame_1stOrderDescent_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H);
		void trackToNextFrame_1stOrderDescent_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H, double lambda);
		
		//Blended-Energy trackers
		void trackToNextFrame_1stOrderDescent_Blended(int nextFrameNumber, ImageStack nextFrameStack);
		
		//Multi-feature trackers
		void trackToNextFrame_multiTrack_GlobalMotionInit(int nextFrameNumber, ImageStack nextFrameStack);
		void trackToNextFrame_multiTrack_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H);
		void trackToNextFrame_multiTrack_GyroPrior(int nextFrameNumber, ImageStack nextFrameStack, Eigen::Matrix3d H, double lambda);
		
	private:
		ImageStack lastFrameStack;
		ImageStack thisFrameStack;
		int lastIDAssigned;
		
		void reviewFeatureSet(void);
		void reviewFeatureSet_mode0(void);
		void reviewFeatureSet_mode1(void);
		void reviewFeatureSet_mode2(void);
		
		void updateFrameNumberAndStacks(int nextFrameNumber, ImageStack &nextFrameStack);
		
		cv::Vec2f trackSingleFeature_1stOrderDescent(Feature& f, ImageStack& nextFrameStack, cv::Vec2f initialLocationEstimate);
		cv::Vec2f trackSingleFeatureGyroOptical_Experimental_1(Feature& f, ImageStack& nextFrameStack, cv::Vec2f initialLocationEstimate, double SearchRadius);
		cv::Vec2f trackSingleFeature_1stOrderDescent_BlendedEnergy(Feature& f, ImageStack& nextFrameStack, cv::Vec2f initialLocationEstimate);
		cv::Vec2f trackSingleFeature_1stOrderDescent_regularized(Feature& f, ImageStack& nextFrameStack, cv::Vec2f initialLocationEstimate, double lambda);
		
		void ConstrainedMultiTracker_LineSearch(cv::Mat& xbar, double &c, cv::Mat &step, double scale, int maxTests, int maxScaleRefinements, ImageStack& frameStack, int resolutionLevel, cv::Mat& History);
		void estimateGlobalMotion(float& delta_x, float& delta_y); //Estimate global motion from last frame to this frame
		void findNewFeatures(int numberOfNewFeatures, cv::Mat& cornerScoreMap, std::list<cv::Vec2f>& newFeatureLocations);
		void ConstrainedMultiTracker_TrackOnSingleLevel(cv::Mat& xbar, ImageStack& frameStack, int resolutionLevel, cv::Mat& History);
		void ConstrainedMultiTrackerWithPrior_TrackOnSingleLevel(cv::Mat& xbar, cv::Mat& xbar_prior, ImageStack& frameStack, int resolutionLevel, cv::Mat& History, double lambda);
		void extensiveReviewOfFeatures(cv::Mat& cornerScoreMap);
		
		int min2i(int a, int b);
		
		cv::Vec2f getArgMinFromCircularSamples(Feature& f, ImageStack& nextFrameStack, cv::Vec2f center, double radius);
		cv::Vec2f SearchForFeature_Blended(Feature& f, ImageStack& nextFrameStack, cv::Vec2f center, double SearchRadius, double meshSize, double lambda, const double blendingCoefficients[]);
		cv::Vec2f trackSingleFeature_1stOrderDescent_SingleBlendedEnergy(Feature& f, ImageStack& nextFrameStack, cv::Vec2f initialLocationEstimate, const double blendingCoefficients[], int minSteps, int maxSteps);
};
