#include "opencv2/opencv.hpp"
#include <unistd.h>
#include <sys/time.h>
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"
#include "plotting.h"

#include <iostream>
#include <fstream>
#include <iomanip> 

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_GYROPATH "/dev/ttyACM0"

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//Forward Declarations
void displayHelp(void);

int main(int argc, char** argv) {
     string GyroDevicePath;
     int retval;
     
     //Parse command-line options
	int gflag = 0;
	int hflag = 0;
	char *gval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "g:h")) != -1)
		switch (c) {
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}

	if (gflag == 1)
		GyroDevicePath = string(gval);
	else
		GyroDevicePath = string(DEFAULT_GYROPATH);
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
	
	fprintf(stderr,"Launching Gyro Subsystem\n\r");
	//Start Gyro Subsystem
	int returnCode;
	returnCode = gyro::openPort(GyroDevicePath.c_str());
	if (returnCode != 0) {
		fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Set Gyro divider
	usleep(100000U); //Wait 1/10 of a second.
	returnCode = gyro::setDivider(1U);
	if (returnCode != 0) {
		fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Check that sensor is giving us data
	usleep(100000U); //Wait 1/10 of a second.
	//sleep(1U); //Wait 1 second.
	if (!gyro::receivingGyroData()) {
		fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Step 1: Prepare for and execute gyro bias estimation
	int numberOfSecondsForBiasEstimation = 15;
	fprintf(stderr,"Gyro subsystem running. Ready to begin gyro bias estimation.\n\r");
	fprintf(stderr,"Bias estimation takes %d seconds. When running, do not disturb the gyro.\n\r",numberOfSecondsForBiasEstimation);
	fprintf(stderr,"Press the Enter key to begin (there will be a countdown.)\n\r");
	cin.get();
	fprintf(stderr,"Beginning gyro bias estimation in 3\r");
	sleep(1);
	fprintf(stderr,"Beginning gyro bias estimation in 2\r");
	sleep(1);
	fprintf(stderr,"Beginning gyro bias estimation in 1\r");
	sleep(1);
	fprintf(stderr,"Beginning gyro bias estimation in 0\n\r");
	fprintf(stderr,"Estimating Gyro Biases... ");
	
	double Xavg, Yavg, Zavg;
	int numberOfSamplesUsed = gyro::averageReadings(numberOfSecondsForBiasEstimation, Xavg, Yavg, Zavg);
	fprintf(stderr,"Done\n\r");
	fprintf(stderr,"Gyro Bias Estimation Complete - Biases based on %d measurements over %d seconds\n\r\n\r",numberOfSamplesUsed,numberOfSecondsForBiasEstimation);
	fprintf(stderr,"Gyro Bias: |% .10f |\n\r", Xavg);
	fprintf(stderr,"           |% .10f |\n\r", Yavg);
	fprintf(stderr,"           |% .10f |\n\r", Zavg);
	
	fprintf(stderr,"\n\rUpdating Gyro calibration constants... ");
	gyro::setGyroBiasCorrections(-1.0*Xavg, -1.0*Yavg, -1.0*Zavg);
	fprintf(stderr,"Done.\n\r");
	
	//Write calibration constants to file - Don't alter constants other than biases
	string outputFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Writing Calibration Data to file: %s ... ", outputFileName.c_str());
	
	Vector3d b;
	double Lgyro;
	Matrix3d Ktilde;
	double sigma;
	loadCalibrationData(outputFileName, b, Ktilde, sigma, Lgyro);
	
	double bx, by, bz;
	gyro::getGyroBiasCorrections(bx, by, bz);
	b(0) = bx;
	b(1) = by;
	b(2) = bz;
	
	retval = saveCalibrationData(outputFileName, b, Ktilde, sigma, Lgyro);
	if (retval == 0) fprintf(stderr,"Success\n\r");
	else fprintf(stderr,"Failure\n\r");
	
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"***********************************    Gyro Bias Estimation    *********************************\n\r");
	fprintf(stderr,"This program runs the gyro bias estimation sequence. Certain parameters can be specified through\n\r");
	fprintf(stderr,"command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file.\n\r");
	fprintf(stderr,"           Defalt value is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}
