#include "opencv2/opencv.hpp"
#include <unistd.h>
#include <sys/time.h>
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"
#include "plotting.h"

#include <iostream>
#include <fstream>
#include <iomanip> 

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_GYROPATH "/dev/ttyACM0"
#define DEFAULT_VIDEOCAM_NUM 1

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//If this program is run without arguments, it looks to a hard-coded default path for the gyro module: /dev/ttyACM0
//If this program is run with a single argument, that argument should be the device path of the serial gyro module.

//Forward Declarations
void displayHelp(void);
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);
void flatten2D_1D(vector<Vector2d> &samples, vector<double> &flatSignal, bool normalize);
void flatten3D_1D(vector<Vector3d> &samples, vector<double> &flatSignal, bool normalize);
void resample(vector<double>& ts, vector<double>& Xs, vector<double> &new_ts, vector<double> &new_Xs);
double interp1(double t1, double x1, double t2, double x2, double t);

//Constants
const double pi = 3.141592653589793;

namespace parameters {
	int  trackingScheme = 1;
	bool replaceLostFeatures = true;
}

int main(int argc, char** argv) {
     string GyroDevicePath;
     int cameraOpenCVIndex = 0;
     int retval, key;
     
     //Parse command-line options
	int gflag = 0;
	int vflag = 0;
	int hflag = 0;
	char *gval = NULL;
	char *vval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "g:v:h")) != -1)
		switch (c) {
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'v':
				vflag = 1;
				vval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}

	if (gflag == 1)
		GyroDevicePath = string(gval);
	else
		GyroDevicePath = string(DEFAULT_GYROPATH);
	
	if (vflag == 1)
		cameraOpenCVIndex = atoi(vval);
	else
		cameraOpenCVIndex = DEFAULT_VIDEOCAM_NUM;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
	
	fprintf(stderr,"Initializing Optical Subsystem\n\r");
	//Initialize Output Window
     namedWindow("Main Output", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	fprintf(stderr,"Launching Gyro Subsystem\n\r");
	//Start Gyro Subsystem
	int returnCode;
	returnCode = gyro::openPort(GyroDevicePath.c_str());
	if (returnCode != 0) {
		fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Set Gyro divider
	usleep(100000U); //Wait 1/10 of a second.
	returnCode = gyro::setDivider(1U);
	if (returnCode != 0) {
		fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Check that sensor is giving us data
	usleep(100000U); //Wait 1/10 of a second.
	//sleep(1U); //Wait 1 second.
	if (!gyro::receivingGyroData()) {
		fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Load Gyro calibration data
	string gyroCalibrationFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Loading Gyro Calibration Data from file: %s\n\r", gyroCalibrationFileName.c_str());
	Vector3d b;
	double Lgyro;
	Matrix3d Ktilde;
	double sigma;
	loadCalibrationData(gyroCalibrationFileName, b, Ktilde, sigma, Lgyro);
	
	//Set gyro bias corrections
	fprintf(stderr,"Setting gyro bias corrections... ");
	gyro::setGyroBiasCorrections(b(0), b(1), b(2));
	fprintf(stderr,"Done.\n\r");
	
	//Estimate Gyro subsystem latency
	fprintf(stderr,"Ready to measure gyro system latency (relative to camera system latency).\n\r");
	fprintf(stderr,"To do this, rotate the camera back and forth about an axis orthogonal to the cameras optical axis.\n\r");
	fprintf(stderr,"Go back and forth several times. Once recording begins, press space to stop recording.\n\r");
	
	//Disable autofocus or auto-exposure for camera (if desired) - Linux and V4L only
	Utilities::adjustCameraSettings(cap, "Main Output", cameraOpenCVIndex);
	
	key = waitForButtonPress(cap, "Main Output", 5, string("Ready to start latancy test"), string("(there will be a countdown)"));
	countDown(cap, "Main Output", 4, string("Starting in..."), string(""));
	gyro::startRecordingReadings();
	struct timeval oldTime, newTime;
	Mat oldFrame, newFrame;
	vector<Point2f> points1, points2;
	vector<uchar> trackedFlags;
	int numberOfGoodCorrespondences;
	Point2f meanFlow;
	struct timeval t0;
	gettimeofday(&t0, NULL);
	double centerTime;
	list<Point2f> meanFlows_list;
	list<double>  centerTimes_list;
	getFrameWithTimestamp(cap, newFrame, newTime);
	while (true) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		
		getPointCorrespondences(oldFrame, newFrame, points1, points2, trackedFlags, numberOfGoodCorrespondences);
		
		//Compute mean flow between frames
		meanFlow = Point2f(0.0, 0.0);
		for (int i=0; i<((int) points1.size()); i++) {
			if (trackedFlags[i] == 1U)
				meanFlow += points2[i] - points1[i];
		}
		meanFlow = (1.0/((double)numberOfGoodCorrespondences))*meanFlow;
		
		//Compute center time
		centerTime = 0.5*timeDiff(newTime, t0) + 0.5*timeDiff(oldTime, t0);
		
		//Save flow and time
		meanFlows_list.push_back(meanFlow);
		centerTimes_list.push_back(centerTime);
		
		//Render and check for stop condition
		imshow("Main Output", newFrame);
		key = waitKey(1);
		if (key == 32) break;
	}
	vector<struct timeval> gyro_sampleTimes_compTimebase;
	vector<Vector3d> gyroMeasurements;
	gyro::stopAndReturnMeasurements(gyro_sampleTimes_compTimebase, gyroMeasurements);
	
	//Convert gyro measurement times to timebase centered at t0.
	vector<double> gyro_sampleTimes;
	gyro_sampleTimes.resize((int)gyro_sampleTimes_compTimebase.size());
	for (int i=0; i<((int)gyro_sampleTimes_compTimebase.size()); i++) {
		gyro_sampleTimes[i] = timeDiff(gyro_sampleTimes_compTimebase[i], t0);
	}
	
	//Convert Optical flow measurement lists to vectors
	vector<Vector2d> meanFlows;
	vector<double>   centerTimes(centerTimes_list.begin(), centerTimes_list.end());
	meanFlows.resize((int) meanFlows_list.size());
	int index = 0;
	for (list<Point2f>::iterator it = meanFlows_list.begin(); it != meanFlows_list.end(); it++)
		meanFlows[index++] = Vector2d(it->x, it->y);
	
	//Flatten signals to 1D functions of time
	vector<double> flows1D;
	vector<double> gyro1D;
	flatten2D_1D(meanFlows, flows1D, true);
	flatten3D_1D(gyroMeasurements, gyro1D, true);
	
	//Plot the raw data (flattened) as a sanity check
	fprintf(stderr,"Rendering gyro and optical flow signals. Press Enter to continue.\n\r");
	plot::plot2Vec(centerTimes, flows1D, gyro_sampleTimes, gyro1D, string("Time (s)"), string("Flow/Rotation Rate"), string(""), string("Optical Flow"), string("Gyro Data"));
	
	//Find the time shift that minimizes the difference between the flow and gyro signals.
	fprintf(stderr,"Analyzing recorded data to estimate relative latency [0%%]");
	double lag, energy;
	vector<double> lags;
	vector<double> corr;
	vector<double> sample_ts_gyro, resampled_gyro_measurements;
	
	double maxLag = 0.1;
	double lagRes = 0.001;
	
	int numLags = 2*((int) (maxLag/lagRes)) + 1;
	double lag0 = -1.0*((numLags-1)/2)*lagRes;
	lags.resize(numLags);
	for (int i=0; i<numLags; i++) {
		lags[i] = lag0 + i*lagRes;
	}
	
	corr.resize((int) lags.size());
	for (int i=0; i<((int) lags.size()); i++) {
		lag = lags[i];
		sample_ts_gyro.resize((int) centerTimes.size());
		for (int j=0; j<((int) centerTimes.size()); j++) {
			sample_ts_gyro[j] = centerTimes[j] + lag;
		}
		
		//Resample gyro data at new times
		resample(gyro_sampleTimes, gyro1D, sample_ts_gyro, resampled_gyro_measurements);
		
		//Now integrate product of signals
		energy = 0.0;
		for (int j=0; j<((int) centerTimes.size()); j++) {
			energy += flows1D[j]*resampled_gyro_measurements[j];
		}
		corr[i] = abs(energy);
		fprintf(stderr,"\rAnalyzing recorded data to estimate relative latency [%d%%]",(int)((100*i)/((int) lags.size() - 1)));
	}
	fprintf(stderr,"\n\r");
	
	//Find best lag
	double bestLag = lags[0];
	double bestEnergy = corr[0];
	for (int i=0; i<((int) lags.size()); i++) {
		if (corr[i] > bestEnergy) {
			bestLag = lags[i];
			bestEnergy = corr[i];
		}
	}
	
	fprintf(stderr,"Gyro system lags behind optical system by %f seconds.\n\r", bestLag);
	
	//Plot signal cross correlation
	fprintf(stderr,"Rendering signal cross-correlation energy. Press Enter to continue.\n\r");
	plot::plot1Vec(lags, corr, string("lag (s)"), string("XCorr Energy"), string(""));
	
	//Write calibration constants to file - only overwrite latency
	fprintf(stderr,"Gyro System Calibration Finished\n\r");
	
	string outputFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Writing Calibration Data to file: %s ... ", outputFileName.c_str());
	loadCalibrationData(outputFileName, b, Ktilde, sigma, Lgyro);
	Lgyro = bestLag;
	
	retval = saveCalibrationData(outputFileName, b, Ktilde, sigma, Lgyro);
	if (retval == 0) fprintf(stderr,"Success\n\r");
	else fprintf(stderr,"Failure\n\r");
	
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"***********************************     Latency Calibration     *********************************\n\r");
	fprintf(stderr,"This program runs the latency calibration sequence. Certain parameters can be specified through\n\r");
	fprintf(stderr,"command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file.\n\r");
	fprintf(stderr,"           Defalt value is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking. Default is %d.\n\r\n\r", DEFAULT_VIDEOCAM_NUM);
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	cvtColor(frameColor, frame, CV_RGB2GRAY);
}

void flatten2D_1D(vector<Vector2d> &samples, vector<double> &flatSignal, bool normalize) {
	//The signal should be approximately centered.
	MatrixXd A(2, (int) samples.size());
	for (int i=0; i<(int) samples.size(); i++) {
		A(0,i) = samples[i](0);
		A(1,i) = samples[i](1);
	}
	
	JacobiSVD<MatrixXd> svd(A, ComputeThinV);
	MatrixXd V = svd.matrixV();
	double sigma1 = svd.singularValues()(0);
	
	flatSignal.resize((int) samples.size());
	for (int i=0; i<(int) samples.size(); i++) {
		flatSignal[i] = sigma1*V(i,0);
	}
	
	if (normalize) {
		double maxVal = 0.0;
		for (int i=0; i<(int) flatSignal.size(); i++) {
			if (abs(flatSignal[i]) > maxVal)
				maxVal = abs(flatSignal[i]);
		}
		if (maxVal > 0.00001) {
			for (int i=0; i<(int) flatSignal.size(); i++)
				flatSignal[i] = flatSignal[i]/maxVal;
		}
	}
}

void flatten3D_1D(vector<Vector3d> &samples, vector<double> &flatSignal, bool normalize) {
	//The signal should be approximately centered.
	MatrixXd A(3, (int) samples.size());
	for (int i=0; i<(int) samples.size(); i++) {
		A(0,i) = samples[i](0);
		A(1,i) = samples[i](1);
		A(2,i) = samples[i](2);
	}
	
	JacobiSVD<MatrixXd> svd(A, ComputeThinV);
	MatrixXd V = svd.matrixV();
	double sigma1 = svd.singularValues()(0);
	
	flatSignal.resize((int) samples.size());
	for (int i=0; i<(int) samples.size(); i++) {
		flatSignal[i] = sigma1*V(i,0);
	}
	
	if (normalize) {
		double maxVal = 0.0;
		for (int i=0; i<(int) flatSignal.size(); i++) {
			if (abs(flatSignal[i]) > maxVal)
				maxVal = abs(flatSignal[i]);
		}
		if (maxVal > 0.00001) {
			for (int i=0; i<(int) flatSignal.size(); i++)
				flatSignal[i] = flatSignal[i]/maxVal;
		}
	}
}

void resample(vector<double>& ts, vector<double>& Xs, vector<double> &new_ts, vector<double> &new_Xs) {
	//Perform linear interpolation to re-sample a time series (in R^1) at new time instants. Input arrays must not be empty or the function will fail and do nothing.
	//If resample times do not fall within between 2 values in "ts", then they are extrapolated with nearest neighbor extrapolation.
	//ts and new_ts must both be weakly monotonically increasing, or the output will be incorrect (although the function will execute and return normally).
	if (ts.size() == 0) {
		fprintf(stderr,"resample: Function called with empty input arrays. Aborting function.\n\r");
		return;
	}
	if (ts.size() != Xs.size()) {
		fprintf(stderr,"resample: Function called ts and Xs of different sizes. Aborting function.\n\r");
		return;
	}
	
	new_Xs.resize((int) Xs.size());
	int newDataIndex = 0;
	int leftIndex = 0;
	
	while (newDataIndex < ((int) (new_ts.size()))) {
		double t = new_ts[newDataIndex];
		if (t < ts[leftIndex]) {
			//We must extrapolate the value on the left - use nearest neighbor
			new_Xs[newDataIndex++] = Xs[0];
		}
		else if (leftIndex >= (((int) (ts.size()))-1)) {
			//We must extrapolate the value on the right - use nearest neighbor
			new_Xs[newDataIndex++] = Xs[ts.size()-1];
		}
		else if (t <= ts[leftIndex+1]) {
			//Interpolate using leftIndex and leftIndex+1
			new_Xs[newDataIndex++] = interp1(ts[leftIndex], Xs[leftIndex], ts[leftIndex+1], Xs[leftIndex+1], t);
		}
		else {
			leftIndex++;
		}	
	}
}

double interp1(double t1, double x1, double t2, double x2, double t) {
	//Linearly interpolate between 2 data points (t1,x1) and (t2,x2). t must be between t1 and t2 (possibly equal to one of them).
	double T = t2 - t1;
	if (T < 0.0000001) return(0.5*x1 + 0.5*x2);
	else return(((t2-t)*x1 + (t-t1)*x2)/T);
}

