#include "opencv2/opencv.hpp"
#include <unistd.h>
#include <sys/time.h>
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"

#include <iostream>
#include <fstream>
#include <iomanip> 

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_GYROPATH "/dev/ttyACM0"
#define DEFAULT_VIDEOCAM_NUM 1

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//If this program is run without arguments, it looks to a hard-coded default path for the gyro module: /dev/ttyACM0
//If this program is run with a single argument, that argument should be the device path of the serial gyro module.

//Forward Declarations
void displayHelp(void);
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);
void estimateKtilde(list<Matrix3d> &RList, list<Matrix3d> &HList, Matrix3d &Ktilde);
void generateGFromHandRPairs(list<Matrix3d> &RList, list<Matrix3d> &HList, MatrixXd &G);
void refineHandRPairs(list<Matrix3d> &RList, list<double> &thetaList, list<Matrix3d> &HList, list<double> &homographyFitErrorList, list<double> &fractionOfCorrespondencesUsedList);
void computeHomographyRecoveryResiduals(list<Matrix3d> &RList, list<Matrix3d> &HList, Matrix3d &Ktilde, vector<double> &residuals);

//Constants
const double pi = 3.141592653589793;

namespace parameters {
	int  trackingScheme = 1;
	bool replaceLostFeatures = true;
}

int main(int argc, char** argv) {
     string GyroDevicePath;
     int cameraOpenCVIndex = 0;
     int key, retval;
     Mat temp, frame, frameBW, frameWithFeatures;
     ImageStack frameStack;
     
     //Parse command-line options
	int gflag = 0;
	int vflag = 0;
	int hflag = 0;
	char *gval = NULL;
	char *vval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "g:v:h")) != -1)
		switch (c) {
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'v':
				vflag = 1;
				vval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}

	if (gflag == 1)
		GyroDevicePath = string(gval);
	else
		GyroDevicePath = string(DEFAULT_GYROPATH);
	
	if (vflag == 1)
		cameraOpenCVIndex = atoi(vval);
	else
		cameraOpenCVIndex = DEFAULT_VIDEOCAM_NUM;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
     
     //Initialize Main Output Window
     namedWindow("Main Output", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	//Start Gyro Subsystem
	int returnCode;
	returnCode = gyro::openPort(GyroDevicePath.c_str());
	if (returnCode != 0) {
		fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Set Gyro divider
	usleep(100000U); //Wait 1/10 of a second.
	returnCode = gyro::setDivider(1U);
	if (returnCode != 0) {
		fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Check that sensor is giving us data
	usleep(100000U); //Wait 1/10 of a second.
	//sleep(1U); //Wait 1 second.
	if (!gyro::receivingGyroData()) {
		fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Disable autofocus or auto-exposure for camera (if desired) - Linux and V4L only
	Utilities::adjustCameraSettings(cap, "Main Output", cameraOpenCVIndex);
	
	//Load Gyro calibration data
	string gyroCalibrationFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Loading Gyro Calibration Data from file: %s\n\r", gyroCalibrationFileName.c_str());
	Vector3d b;
	double Lgyro;
	Matrix3d Ktilde;
	double sigma;
	loadCalibrationData(gyroCalibrationFileName, b, Ktilde, sigma, Lgyro);
	
	//Set gyro bias and latency corrections
	fprintf(stderr,"Setting gyro bias and latency corrections... ");
	gyro::setGyroBiasCorrections(b(0), b(1), b(2));
	gyro::setGyroLatencyCorrection(Lgyro);
	fprintf(stderr,"Done.\n\r");
	
	//Step 2: Record a video with rotation and estimate H and R between pairs of consecutive frames.
	waitForButtonPress(cap, "Main Output", 5, string("Press space to record cal. seq."), string("Press space again to stop."));
	fprintf(stderr,"Recording Frame Sequence for H and R estimation... ");
	
	struct timeval oldTime, newTime;
	Mat oldFrame, newFrame;
	getFrameWithTimestamp(cap, newFrame, newTime);
	double rotationAngle, fractionOfCorrespondencesUsed, homographyFitError;
	Matrix3d R, H;
	list<Matrix3d> RList;
	list<double>   thetaList;
	list<Matrix3d> HList;
	list<double> homographyFitErrorList;
	list<double> fractionOfCorrespondencesUsedList;
	while (true) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		
		//Estimate the homography between frames
		estimateHomographyBetweenFrames(oldFrame, newFrame, H, fractionOfCorrespondencesUsed, homographyFitError);
		//fprintf(stderr,"homographyFitError: %f\n\r", homographyFitError);
		//printf("\n\r\n\r");
		//cout << H;
		
		//Get rotation between frames - do this after homography estimation so the gyro measurement buffer has had time to
		//collect measurements from after second frame arrival time.
		retval = gyro::getRotationBetweenTwoTimes(oldTime, newTime, R, rotationAngle, true);
		//printf("\n\r\n\r");
		//cout << R;
		//fprintf(stderr,"Rotation Angle: %f (deg)\n\r", rotationAngle*(180.0/pi));
		if (retval != 0) {
			fprintf(stderr,"gyro::getRotationBetweenTwoTimes failed with return code %d", retval);
			gyro::closePort();
			return(-1);
		}
		
		//Save H, R, and H quality estimates
		RList.push_back(R);
		HList.push_back(H);
		thetaList.push_back(rotationAngle);
		homographyFitErrorList.push_back(homographyFitError);
		fractionOfCorrespondencesUsedList.push_back(fractionOfCorrespondencesUsed);
		
		//Show frame and check for stop condition (space bar)
		imshow("Main Output", newFrame);
		key = waitKey(1);
		if (key == 32) break;
	}
	fprintf(stderr,"Done.\n\r");
	
	//Save raw (H,R) pairs and quality data to file for Matlab/Octave algorithm development
	fprintf(stderr,"Exporting raw data for high-level algorithm development...");
	ofstream myfile;
	myfile.open ("RawCalibrationData.txt");
	list<Matrix3d>::iterator RList_it                             = RList.begin();
	list<double>::iterator   thetaList_it                         = thetaList.begin();
	list<Matrix3d>::iterator HList_it                             = HList.begin();
	list<double>::iterator   homographyFitErrorList_it            = homographyFitErrorList.begin();
	list<double>::iterator   fractionOfCorrespondencesUsedList_it = fractionOfCorrespondencesUsedList.begin();
	while (RList_it != RList.end()) {
		H = *HList_it;
		R = *RList_it;
		myfile << setprecision(15) << H(0,0) << "," << H(0,1) << "," << H(0,2) << ",";
		myfile << setprecision(15) << H(1,0) << "," << H(1,1) << "," << H(1,2) << ",";
		myfile << setprecision(15) << H(2,0) << "," << H(2,1) << "," << H(2,2) << ",";
		myfile << setprecision(15) << R(0,0) << "," << R(0,1) << "," << R(0,2) << ",";
		myfile << setprecision(15) << R(1,0) << "," << R(1,1) << "," << R(1,2) << ",";
		myfile << setprecision(15) << R(2,0) << "," << R(2,1) << "," << R(2,2) << ",";
		myfile << setprecision(15) << *thetaList_it << ",";
		myfile << setprecision(15) << *homographyFitErrorList_it << ",";
		myfile << setprecision(15) << *fractionOfCorrespondencesUsedList_it << "\n";
		
		RList_it++;
		thetaList_it++;
		HList_it++;
		homographyFitErrorList_it++;
		fractionOfCorrespondencesUsedList_it++;
	}
	myfile.close();
	fprintf(stderr,"Done.\n\r");
	
	//Step 3: Filter out bad (H,R) pairs
	int numPairsBeforeFiltering = (int) (RList.size());
	fprintf(stderr,"Filtering out poor and unhelpful (H,R) pairs... ");
	refineHandRPairs(RList, thetaList, HList, homographyFitErrorList, fractionOfCorrespondencesUsedList);
	int numPairsAfterFiltering = (int) (RList.size());
	fprintf(stderr,"Done\n\r");
	fprintf(stderr,"Retained %d out of %d pairs (%.1f %%)\n\r", numPairsAfterFiltering, numPairsBeforeFiltering, 100.0*((double)numPairsAfterFiltering)/((double)numPairsBeforeFiltering));
	
	//Step 4: Estimate Ktilde from surviving (H,R) pairs
	fprintf(stderr,"Estimating Ktilde from retained (H,R) pairs... ");
	estimateKtilde(RList, HList, Ktilde);
	fprintf(stderr,"Done\n\r");
	fprintf(stderr,"Ktilde was estimated from %d inlier (H,R) pairs.\n\r", ((int) RList.size()));
	Matrix3d KtildeInv = Ktilde.inverse(); //Precompute inverse
	
	//Step 5: Estimate Confidence in Ktilde based on new measurements
	waitForButtonPress(cap, "Main Output", 5, string("Press space to record test seq."), string("Press space again to stop."));
	fprintf(stderr,"Recording Frame Sequence for Ktilde confidence detirmination... ");
	getFrameWithTimestamp(cap, newFrame, newTime);
	vector<Point2f> points1, points2;
	vector<uchar> trackedFlags;
	int numberOfGoodCorrespondences;
	vector<double> predictionErrorsBetween2Frames;
	list<double> flowPredictionErrors;
	Vector3d point1Hom, point2Hom;
	point1Hom(2) = 1.0;
	Point2f predictedPoint;
	Scalar color_measured(255,0,0);
	Scalar color_predicted(0,0,255);
	Mat newFrameColor;
	while (true) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		
		getPointCorrespondences(oldFrame, newFrame, points1, points2, trackedFlags, numberOfGoodCorrespondences);
		
		//Get rotation between frames - do this after homography estimation so the gyro measurement buffer has had time to
		//collect measurements from after second frame arrival time.
		retval = gyro::getRotationBetweenTwoTimes(oldTime, newTime, R, rotationAngle, true);
		//printf("\n\r\n\r");
		//cout << R;
		//fprintf(stderr,"Rotation Angle: %f (deg)\n\r", rotationAngle*(180.0/pi));
		if (retval != 0) {
			fprintf(stderr,"gyro::getRotationBetweenTwoTimes failed with return code %d", retval);
			gyro::closePort();
			return(-1);
		}
		
		//Estimate Homography between 2 frames
		H = Ktilde*R*KtildeInv;
		
		//Measure flow prediction errors for each feature
		computePointCorrespondenceErrorsBasedOnHomography(H, points1, points2, predictionErrorsBetween2Frames);
		
		//Add flow prediction errors to cumulative errors vector
		for (int i=0; i<((int) predictionErrorsBetween2Frames.size()); i++)
			if (trackedFlags[i] == 1U)
				flowPredictionErrors.push_back(predictionErrorsBetween2Frames[i]);
		
		//Render measured and predicted flow to newFrame (in color)
		cvtColor(newFrame, newFrameColor, CV_GRAY2RGB);
		for (int i=0; i<((int) points1.size()); i++) {
			if (trackedFlags[i] == 1U) {
				point1Hom(0) = points1[i].x;
				point1Hom(1) = points1[i].y;
			
				point2Hom = H*point1Hom;
				predictedPoint.x = point2Hom(0)/point2Hom(2);
				predictedPoint.y = point2Hom(1)/point2Hom(2);
			
				//Go to un-normalized coordinates for display purposes
				denormalizeImageCoords(newFrame.cols, newFrame.rows, points1[i]);
				denormalizeImageCoords(newFrame.cols, newFrame.rows, points2[i]);
				denormalizeImageCoords(newFrame.cols, newFrame.rows, predictedPoint);
			
				//Render Measured Flow
				//line(newFrameColor, points1[i], points2[i], color_measured, 2);
			
				//Render Predicted Flow
				line(newFrameColor, points1[i], predictedPoint, color_predicted, 2);
			
				//Render Circle at starting point
				circle(newFrameColor, points1[i], 3, color_measured, 2);
			}
		}
		
		//Show frame and check for stop condition (space bar)
		imshow("Main Output", newFrameColor);
		key = waitKey(1);
		if (key == 32) break;
	}
	fprintf(stderr,"Done.\n\r");
	
	vector<double> flowPredictionErrorsVec(flowPredictionErrors.begin(), flowPredictionErrors.end());
	vector<uchar> underThresholdFlags;
	underThresholdFlags.resize((int) flowPredictionErrorsVec.size());
	sigma = coputeStandardDeviation(flowPredictionErrorsVec);
	double threshold = sigma;
	for (int i=0; i<((int) flowPredictionErrorsVec.size()); i++) {
		if (flowPredictionErrorsVec[i] < threshold) {
			underThresholdFlags[i] = 1U;
		}
		else {
			underThresholdFlags[i] = 0U;
		}
	}
	sigma = coputeStandardDeviation(flowPredictionErrorsVec, underThresholdFlags);
	
	fprintf(stderr,"sigma = %.15f  (normalized coords)\n\r", sigma);
	
	//Write calibration constants to file
	
	fprintf(stderr,"Calibration Finished\n\r");
	
	string outputFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Writing Calibration Data to file: %s ... ", outputFileName.c_str());
	double bx, by, bz;
	gyro::getGyroBiasCorrections(bx, by, bz);
	b(0) = bx;
	b(1) = by;
	b(2) = bz;
	retval = saveCalibrationData(outputFileName, b, Ktilde, sigma, Lgyro);
	if (retval == 0) fprintf(stderr,"Success\n\r");
	else fprintf(stderr,"Failure\n\r");
	
	//Re-run gyro bias estimation to check that biases are decreased after applying corrections
	/*waitForButtonPress(cap, "Main Output", 5, string("Press any key to start Gyro Bias"), string("Estimation. Then Do Not Touch!"));
	if (countDown(cap, "Main Output", 4, string("Bias estimation takes ")+to_string(numberOfSecondsForBiasEstimation)+string(" seconds."), string("Starting. Do Not Touch!")) == -1) return(0);
	
	numberOfSamplesUsed = gyro::averageReadings(numberOfSecondsForBiasEstimation, Xavg, Yavg, Zavg);
	fprintf(stderr,"Gyro Bias Estimation Complete - Biases based on %d measurements over %d seconds\n\r",numberOfSamplesUsed,numberOfSecondsForBiasEstimation);
	fprintf(stderr,"Gyro Bias: (% .10f, % .10f, % .10f)\n\r", Xavg, Yavg, Zavg);*/
	
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"******************************     Gyro-Optical Calibration     ****************************\n\r");
	fprintf(stderr,"This program runs the gyro-optical calibration sequence. This does not include gyro bias\n\r");
	fprintf(stderr,"or relative latency estimation. Certain parameters can be specified through command-line\n\r");
	fprintf(stderr,"options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file.\n\r");
	fprintf(stderr,"           Defalt value is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking. Default is %d.\n\r\n\r", DEFAULT_VIDEOCAM_NUM);
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	cvtColor(frameColor, frame, CV_RGB2GRAY);
}

void refineHandRPairs(list<Matrix3d> &RList, list<double> &thetaList, list<Matrix3d> &HList, list<double> &homographyFitErrorList, list<double> &fractionOfCorrespondencesUsedList) {
	//Filter out bad (H,R) pairs. Here, we remove a pair (H,R) from our set if we have low confidence in our estimate of H, or if the rotation was through too small
	//of an angle for the pair to be useful for estimating Ktilde.
	
	double minFractionOfCorrespondences = 0.75;
	double maxHFitError;
	
	//First, we must evaluate our required homography fit error (upper bound)
	double cumsum = 0.0;
	for (list<double>::iterator it = homographyFitErrorList.begin(); it != homographyFitErrorList.end(); it++) {
		cumsum += (*it)*(*it);
	}
	maxHFitError = sqrt(cumsum/(homographyFitErrorList.size()));
	
	//Generate a min rotation angle threshold
	vector<double> sortedThetaList(thetaList.begin(), thetaList.end());
	sort(sortedThetaList.begin(), sortedThetaList.end()); //Sort into ascending order
	//Grab the theta corresponding to the 90'th percentile
	int index = (int)(0.9*((double) (sortedThetaList.size())));
	if (index > (((int) sortedThetaList.size()) - 1)) index = ((int) sortedThetaList.size()) - 1;
	double value90Percentile = sortedThetaList[index];
	double minTheta = value90Percentile/5.0;
	
	//Throw out any pair where one of the following conditions is met:
	//1: The fraction of correspondences used to build H was less than minFractionOfCorrespondences 
	//2: The homography fit error exceeds maxHFitError
	//3: The angle of rotations was below angleThreshold
	
	list<Matrix3d>::iterator RList_it                             = RList.begin();
	list<double>::iterator   thetaList_it                         = thetaList.begin();
	list<Matrix3d>::iterator HList_it                             = HList.begin();
	list<double>::iterator   homographyFitErrorList_it            = homographyFitErrorList.begin();
	list<double>::iterator   fractionOfCorrespondencesUsedList_it = fractionOfCorrespondencesUsedList.begin();
	
	while (RList_it != RList.end()) {
		if ((*fractionOfCorrespondencesUsedList_it < minFractionOfCorrespondences) /*|| (*homographyFitErrorList_it > maxHFitError) || (*thetaList_it < minTheta)*/) {
			RList_it                             = RList.erase(RList_it);
			thetaList_it                         = thetaList.erase(thetaList_it);
			HList_it                             = HList.erase(HList_it);
			homographyFitErrorList_it            = homographyFitErrorList.erase(homographyFitErrorList_it);
			fractionOfCorrespondencesUsedList_it = fractionOfCorrespondencesUsedList.erase(fractionOfCorrespondencesUsedList_it);
		}
		else {
			RList_it++;
			thetaList_it++;
			HList_it++;
			homographyFitErrorList_it++;
			fractionOfCorrespondencesUsedList_it++;
		}
	}
}

void estimateKtilde(list<Matrix3d> &RList, list<Matrix3d> &HList, Matrix3d &Ktilde) {
	//Estimate Ktilde from (H,R) pairs. This function performs 2 iterations. After the first pass,
	//outlier pairs are rejected. Upon completion, RList and HList will have been modified to contain
	//only inlier matrices (ones used for estimting Ktilde).
	
	MatrixXd G;
	
	int minPairs = 5;
	if (((int) RList.size()) < minPairs) {
		fprintf(stderr,"Warning: Estimating Ktilde from fewer than %d pairs. Results may be bad.\n\r", minPairs);
	}
	
	//Build G for linear least squares
	generateGFromHandRPairs(RList, HList, G);
	
	//Perform homogeneous linear least squares
	VectorXd Ktilde_vec, sigma;
	homogeneousLinearLeastSquares(G, Ktilde_vec, sigma);
	
	fprintf(stderr,"\n\rDone Estimating Ktilde. singular values where:\n\r");
	cerr << sigma;
	fprintf(stderr,"\n\r\n\r");
	
	//De-vectorize Ktilde
	Ktilde(0,0)=Ktilde_vec(0);  Ktilde(0,1)=Ktilde_vec(1);  Ktilde(0,2)=Ktilde_vec(2);
	Ktilde(1,0)=Ktilde_vec(3);  Ktilde(1,1)=Ktilde_vec(4);  Ktilde(1,2)=Ktilde_vec(5);
	Ktilde(2,0)=Ktilde_vec(6);  Ktilde(2,1)=Ktilde_vec(7);  Ktilde(2,2)=Ktilde_vec(8);
	
	//Reconstruct each H from Ktilde and the corresponding R and measure homography recovery residuals.
	vector<double> residuals;
	computeHomographyRecoveryResiduals(RList, HList, Ktilde, residuals);
	
	//Compute standard deviation of recovery residuals and set residual threshold
	double sd = coputeStandardDeviation(residuals);
	double threshold = sd;
	
	//Throw out outlier (H,R) pairs
	list<Matrix3d>::iterator R_iter = RList.begin();
	list<Matrix3d>::iterator H_iter = HList.begin();
	int index = 0;
	while (R_iter != RList.end()) {
		if (residuals[index] > threshold) {
			R_iter = RList.erase(R_iter);
			H_iter = HList.erase(H_iter);
			index++;
		}
		else {
			R_iter++;
			H_iter++;
			index++;
		}
	}
	
	//Re-compute homography
	generateGFromHandRPairs(RList, HList, G);
	homogeneousLinearLeastSquares(G, Ktilde_vec, sigma);
	
	//De-vectorize Ktilde
	Ktilde(0,0)=Ktilde_vec(0);  Ktilde(0,1)=Ktilde_vec(1);  Ktilde(0,2)=Ktilde_vec(2);
	Ktilde(1,0)=Ktilde_vec(3);  Ktilde(1,1)=Ktilde_vec(4);  Ktilde(1,2)=Ktilde_vec(5);
	Ktilde(2,0)=Ktilde_vec(6);  Ktilde(2,1)=Ktilde_vec(7);  Ktilde(2,2)=Ktilde_vec(8);
}

void generateGFromHandRPairs(list<Matrix3d> &RList, list<Matrix3d> &HList, MatrixXd &G) {
	int numPairs = (int) (RList.size());
	G = MatrixXd(9*numPairs,9);
	Matrix3d H,R,Rt;
	
	list<Matrix3d>::iterator R_iter = RList.begin();
	list<Matrix3d>::iterator H_iter = HList.begin();
	
	int startRow = 0;
	while (R_iter != RList.end()) {
		H = *H_iter;
		R = *R_iter;
		Rt = R.transpose();
		
		G.block(startRow  ,0,3,3) = H(0,0)*MatrixXd::Identity(3, 3) - Rt;
		G.block(startRow  ,3,3,3) = H(0,1)*MatrixXd::Identity(3, 3);
		G.block(startRow  ,6,3,3) = H(0,2)*MatrixXd::Identity(3, 3);
		
		G.block(startRow+3,0,3,3) = H(1,0)*MatrixXd::Identity(3, 3);
		G.block(startRow+3,3,3,3) = H(1,1)*MatrixXd::Identity(3, 3) - Rt;
		G.block(startRow+3,6,3,3) = H(1,2)*MatrixXd::Identity(3, 3);
		
		G.block(startRow+6,0,3,3) = H(2,0)*MatrixXd::Identity(3, 3);
		G.block(startRow+6,3,3,3) = H(2,1)*MatrixXd::Identity(3, 3);
		G.block(startRow+6,6,3,3) = H(2,2)*MatrixXd::Identity(3, 3) - Rt;
		
		R_iter++;
		H_iter++;
		startRow += 9;
	}
}

void computeHomographyRecoveryResiduals(list<Matrix3d> &RList, list<Matrix3d> &HList, Matrix3d &Ktilde, vector<double> &residuals) {
	//This function computes the homography recovery residual for each (H,R) pair provided.
	//This is a measure of how well H can be reconstructed from R and Ktilde.
	
	residuals.resize(RList.size());
	list<Matrix3d>::iterator R_iter = RList.begin();
	list<Matrix3d>::iterator H_iter = HList.begin();
	int index = 0;
	
	Matrix3d recoveredH;
	Matrix3d KtildeInv = Ktilde.inverse();
	
	while (R_iter != RList.end()) {
		recoveredH = Ktilde*(*R_iter)*KtildeInv;
		residuals[index] = ((*H_iter) - recoveredH).norm();
		
		R_iter++;
		H_iter++;
		index++;
	}
}



