//External Includes
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

//Local Includes
#include "calibrationLoadSave.h"

using namespace std;
using namespace Eigen;
using boost::property_tree::ptree;

int loadCalibrationData(const string &filename, Vector3d &b, Matrix3d &Ktilde, double &sigma, double &Lgyro) {
	//Loads calibration data from a JSON file with the provided filename and stores it in the
	//provided variables. If the desired JSON file cannot be read (perhaps it does not exist),
	//then this function fails and returns -1. If the file does not exist, or if certain fields are missing,
	//the relavent default values will be returned. If the file exists and contains even 1 of the fields
	//we look for, then this function returns success (0).
	
	bool readSuccessful = false;
	ptree pt; //Create Empty Property Tree
	
	//Try to read the specified JSON file.
	try {
     	read_json(filename, pt); //Load JSON file
	}
	catch (...) {
		b(0) = 0.0;
		b(1) = 0.0;
		b(2) = 0.0;
		Lgyro = 0.0;
		Ktilde(0,0) = 1.0;
		Ktilde(0,1) = 0.0;
		Ktilde(0,2) = 0.0;
		Ktilde(1,0) = 0.0;
		Ktilde(1,1) = 1.0;
		Ktilde(1,2) = 0.0;
		Ktilde(2,0) = 0.0;
		Ktilde(2,1) = 0.0;
		Ktilde(2,2) = 1.0;
		sigma = NAN;
		return(-1);
	}
	
	try {
		b(0) = pt.get<double>("GyroOpticalCalibration.bx");
		b(1) = pt.get<double>("GyroOpticalCalibration.by");
		b(2) = pt.get<double>("GyroOpticalCalibration.bz");
		readSuccessful = true;
	}
	catch (...) {
		b(0) = 0.0;
		b(1) = 0.0;
		b(2) = 0.0;
	}
	
	try {
		Lgyro = pt.get<double>("GyroOpticalCalibration.Lgyro");
		readSuccessful = true;
	}
	catch (...) {
		Lgyro = 0.0;
	}
	
	try {
		Ktilde(0,0) = pt.get<double>("GyroOpticalCalibration.Ktilde11");
		Ktilde(0,1) = pt.get<double>("GyroOpticalCalibration.Ktilde12");
		Ktilde(0,2) = pt.get<double>("GyroOpticalCalibration.Ktilde13");
		Ktilde(1,0) = pt.get<double>("GyroOpticalCalibration.Ktilde21");
		Ktilde(1,1) = pt.get<double>("GyroOpticalCalibration.Ktilde22");
		Ktilde(1,2) = pt.get<double>("GyroOpticalCalibration.Ktilde23");
		Ktilde(2,0) = pt.get<double>("GyroOpticalCalibration.Ktilde31");
		Ktilde(2,1) = pt.get<double>("GyroOpticalCalibration.Ktilde32");
		Ktilde(2,2) = pt.get<double>("GyroOpticalCalibration.Ktilde33");
		readSuccessful = true;
	}
	catch (...) {
		Ktilde(0,0) = 1.0;
		Ktilde(0,1) = 0.0;
		Ktilde(0,2) = 0.0;
		Ktilde(1,0) = 0.0;
		Ktilde(1,1) = 1.0;
		Ktilde(1,2) = 0.0;
		Ktilde(2,0) = 0.0;
		Ktilde(2,1) = 0.0;
		Ktilde(2,2) = 1.0;
	}
	
	try {
		sigma = pt.get<double>("GyroOpticalCalibration.sigma");
	}
	catch (...) {
		sigma = NAN;
	}
	
	if (readSuccessful) return(0);
	else return(-1);
}


int saveCalibrationData(const string &filename, Vector3d &b, Matrix3d &Ktilde, double &sigma, double &Lgyro) {
	//This function saves the provided calibration coefficients to the JSON file with the given
	//filename. If this operation fails (perhaps the path is not writable), the function fails and
	//returns -1. Upon success, the function returns 0.
	
	ptree root, GyroOpticalCalibration;
	
	GyroOpticalCalibration.put<double>("bx",b(0));
	GyroOpticalCalibration.put<double>("by",b(1));
	GyroOpticalCalibration.put<double>("bz",b(2));
	
	GyroOpticalCalibration.put<double>("Lgyro",Lgyro);
	
	GyroOpticalCalibration.put<double>("Ktilde11",Ktilde(0,0));
	GyroOpticalCalibration.put<double>("Ktilde12",Ktilde(0,1));
	GyroOpticalCalibration.put<double>("Ktilde13",Ktilde(0,2));
	GyroOpticalCalibration.put<double>("Ktilde21",Ktilde(1,0));
	GyroOpticalCalibration.put<double>("Ktilde22",Ktilde(1,1));
	GyroOpticalCalibration.put<double>("Ktilde23",Ktilde(1,2));
	GyroOpticalCalibration.put<double>("Ktilde31",Ktilde(2,0));
	GyroOpticalCalibration.put<double>("Ktilde32",Ktilde(2,1));
	GyroOpticalCalibration.put<double>("Ktilde33",Ktilde(2,2));
	
	GyroOpticalCalibration.put<double>("sigma",sigma);
	
	root.put_child("GyroOpticalCalibration", GyroOpticalCalibration);
	
	try {
		write_json(filename, root); //Write JSON file
		return(0);
	}
	catch (...) {return(-1);}
}
