#include "opencv2/opencv.hpp"
#include <getopt.h>

using namespace cv;
using namespace std;

//Light Degradation Profile
#define LOW_DEGRADATION_PROFILE_MULTIPLIER       0.9
#define LOW_DEGRADATION_PROFILE_NOISE1_MEAN      0.0
#define LOW_DEGRADATION_PROFILE_NOISE1_SIGMA     15.0
#define LOW_DEGRADATION_PROFILE_SIGMA_X          1.5
#define LOW_DEGRADATION_PROFILE_SIGMA_Y          1.5
#define LOW_DEGRADATION_PROFILE_NOISE2_MEAN      0.0
#define LOW_DEGRADATION_PROFILE_NOISE2_SIGMA     1.5

//Medium Degredation Profile
#define MEDIUM_DEGRADATION_PROFILE_MULTIPLIER    0.8
#define MEDIUM_DEGRADATION_PROFILE_NOISE1_MEAN   0.0
#define MEDIUM_DEGRADATION_PROFILE_NOISE1_SIGMA  30.0
#define MEDIUM_DEGRADATION_PROFILE_SIGMA_X       3.0
#define MEDIUM_DEGRADATION_PROFILE_SIGMA_Y       3.0
#define MEDIUM_DEGRADATION_PROFILE_NOISE2_MEAN   0.0
#define MEDIUM_DEGRADATION_PROFILE_NOISE2_SIGMA  3.0

//High Degredation Profile
#define HIGH_DEGRADATION_PROFILE_MULTIPLIER      0.7
#define HIGH_DEGRADATION_PROFILE_NOISE1_MEAN     4.0
#define HIGH_DEGRADATION_PROFILE_NOISE1_SIGMA    60.0
#define HIGH_DEGRADATION_PROFILE_SIGMA_X         5.0
#define HIGH_DEGRADATION_PROFILE_SIGMA_Y         5.0
#define HIGH_DEGRADATION_PROFILE_NOISE2_MEAN     0.0
#define HIGH_DEGRADATION_PROFILE_NOISE2_SIGMA    6.0

//Extreme Degredation Profile
#define EXTREME_DEGRADATION_PROFILE_MULTIPLIER   0.6
#define EXTREME_DEGRADATION_PROFILE_NOISE1_MEAN  6.0
#define EXTREME_DEGRADATION_PROFILE_NOISE1_SIGMA 80.0
#define EXTREME_DEGRADATION_PROFILE_SIGMA_X      7.0
#define EXTREME_DEGRADATION_PROFILE_SIGMA_Y      7.0
#define EXTREME_DEGRADATION_PROFILE_NOISE2_MEAN  0.0
#define EXTREME_DEGRADATION_PROFILE_NOISE2_SIGMA 8.0

void addNoise(Mat &Image, float mu, float sigma, uint64 seed) {
	Mat NoiseMat(Image.rows, Image.cols, Image.type());
	RNG rng(seed);
	rng.fill(NoiseMat, cv::RNG::NORMAL, mu, sigma);
	Image += NoiseMat;
}

unsigned int generateHash(std::string inputString) {
	const char * myStr = inputString.c_str();
	size_t len = inputString.length();
	
	unsigned int hash = 0;
	for(size_t i = 0; i < len; ++i) 
		hash = 65599 * hash + myStr[i];
	return hash ^ (hash >> 16);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"********************************     Image Degradation Program     *******************************\n\r");
	fprintf(stderr,"This program degrades a single image according to the following algorithm:\n\r");
	fprintf(stderr,"1: Multiply each pixel by 'Multiplier' \n\r");
	fprintf(stderr,"2: Add i.i.d. Gaussian noise to each pixel with mean and sd (noise1Mean, noise1Sigma)\n\r");
	fprintf(stderr,"3: Apply Gaussian blur where convolution mask has std. deviations (blurSigmaX, blurSigmaY)\n\r");
	fprintf(stderr,"4: Add i.i.d. Gaussian noise to each pixel with mean and sd (noise2Mean, noise2Sigma)\n\r\n\r");
	fprintf(stderr,"Each of the above parameters can be set manually at the command line, or a pre-set degradation\n\r");
	fprintf(stderr,"profile can be selected. This is done through command-line options. Supported options are:\n\r\n\r");
	fprintf(stderr,"--low             Select low degradation profile\n\r\n\r");
	fprintf(stderr,"--medium          Select medium degradation profile\n\r\n\r");
	fprintf(stderr,"--high            Select high degradation profile\n\r\n\r");
	fprintf(stderr,"--extreme         Select extreme degradation profile\n\r\n\r");
	fprintf(stderr,"Custom degradation parameters can be set through the following options. If any of these are set,\n\r");
	fprintf(stderr,"then un-set parameters will default to values which do not modify the imagery.\n\r");
	fprintf(stderr,"--multiplier=n    Set parameter Multiplier to n\n\r");
	fprintf(stderr,"--noise1Mean=n    Set parameter noise1Mean to n\n\r");
	fprintf(stderr,"--noise1Sigma=n   Set parameter noise1Sigma to n\n\r");
	fprintf(stderr,"--blurSigmaX=n    Set parameter blurSigmaX to n\n\r");
	fprintf(stderr,"--blurSigmaY=n    Set parameter blurSigmaY to n\n\r");
	fprintf(stderr,"--noise2Mean=n    Set parameter noise2Mean to n\n\r");
	fprintf(stderr,"--noise2Sigma=n   Set parameter noise2Sigma to n\n\r\n\r");
	fprintf(stderr,"-h                Display this help message.\n\r\n\r");
	fprintf(stderr,"This program also has one mandatory argument, which is the path to the file which will be degraded.\n\r");
	fprintf(stderr,"The degradation occurs in-place (input file will be over-written.\n\r\n\r");
}

int main(int argc, char** argv) {
     //Setup for command-line option parsing
	static struct option long_options[] = {
		{"help",        no_argument,       0, 'h'},
		{"low",         no_argument,       0, 256},
		{"medium",      no_argument,       0, 257},
		{"high",        no_argument,       0, 258},
		{"extreme",     no_argument,       0, 259},
		{"multiplier",  required_argument, 0, 260},
		{"noise1Mean",  required_argument, 0, 261},
		{"noise1Sigma", required_argument, 0, 262},
		{"blurSigmaX",  required_argument, 0, 263},
		{"blurSigmaY",  required_argument, 0, 264},
		{"noise2Mean",  required_argument, 0, 265},
		{"noise2Sigma", required_argument, 0, 266},
		{0, 0, 0, 0}
	};
	
	enum degradationProfile {LOW, MEDIUM, HIGH, EXTREME, CUSTOM};
	
	//Set default parameters
	degradationProfile profile = LOW;
	float  multiplier  = 1.0;
	float  noise1Mean  = 0.0;
	float  noise1Sigma = 0.0;
	double sigma_x     = 0.0;
	double sigma_y     = 0.0;
	float  noise2Mean  = 0.0;
	float  noise2Sigma = 0.0;
	
	//Parse command-line options and override defaults with provided values.
	int c, option_index;
	while ((c = getopt_long (argc, argv, "h", long_options, &option_index)) != -1)
		switch (c) {
			case 'h':
				displayHelp();
				return(0);
			case 256:
				profile = LOW;
				break;
			case 257:
				profile = MEDIUM;
				break;
			case 258:
				profile = HIGH;
				break;
			case 259:
				profile = EXTREME;
				break;
			case 260:
				profile = CUSTOM;
				multiplier = atof(optarg);
				break;
			case 261:
				profile = CUSTOM;
				noise1Mean = atof(optarg);
				break;
			case 262:
				profile = CUSTOM;
				noise1Sigma = atof(optarg);
				break;
			case 263:
				profile = CUSTOM;
				sigma_x = atof(optarg);
				break;
			case 264:
				profile = CUSTOM;
				sigma_y = atof(optarg);
				break;
			case 265:
				profile = CUSTOM;
				noise2Mean = atof(optarg);
				break;
			case 266:
				profile = CUSTOM;
				noise2Sigma = atof(optarg);
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}
	
	//Check for and process required arguments
	if (argc - optind != 1) {
		fprintf(stderr,"%s has 1 mandatory argument. Run %s -h for complete usage.\n\r", argv[0], argv[0]);
		return(-1);
	}
	string path(argv[optind]);
     
     //Use profile or provided options to set degradation parameters
     switch (profile) {
     	case LOW:
     		multiplier  = LOW_DEGRADATION_PROFILE_MULTIPLIER;
     		noise1Mean  = LOW_DEGRADATION_PROFILE_NOISE1_MEAN;
     		noise1Sigma = LOW_DEGRADATION_PROFILE_NOISE1_SIGMA;
     		sigma_x     = LOW_DEGRADATION_PROFILE_SIGMA_X;
     		sigma_y     = LOW_DEGRADATION_PROFILE_SIGMA_Y;
     		noise2Mean  = LOW_DEGRADATION_PROFILE_NOISE2_MEAN;
     		noise2Sigma = LOW_DEGRADATION_PROFILE_NOISE2_MEAN;
     		break;
     	case MEDIUM:
     		multiplier  = MEDIUM_DEGRADATION_PROFILE_MULTIPLIER;
     		noise1Mean  = MEDIUM_DEGRADATION_PROFILE_NOISE1_MEAN;
     		noise1Sigma = MEDIUM_DEGRADATION_PROFILE_NOISE1_SIGMA;
     		sigma_x     = MEDIUM_DEGRADATION_PROFILE_SIGMA_X;
     		sigma_y     = MEDIUM_DEGRADATION_PROFILE_SIGMA_Y;
     		noise2Mean  = MEDIUM_DEGRADATION_PROFILE_NOISE2_MEAN;
     		noise2Sigma = MEDIUM_DEGRADATION_PROFILE_NOISE2_MEAN;
     		break;
     	case HIGH:
     		multiplier  = HIGH_DEGRADATION_PROFILE_MULTIPLIER;
     		noise1Mean  = HIGH_DEGRADATION_PROFILE_NOISE1_MEAN;
     		noise1Sigma = HIGH_DEGRADATION_PROFILE_NOISE1_SIGMA;
     		sigma_x     = HIGH_DEGRADATION_PROFILE_SIGMA_X;
     		sigma_y     = HIGH_DEGRADATION_PROFILE_SIGMA_Y;
     		noise2Mean  = HIGH_DEGRADATION_PROFILE_NOISE2_MEAN;
     		noise2Sigma = HIGH_DEGRADATION_PROFILE_NOISE2_MEAN;
     		break;
     	case EXTREME:
     		multiplier  = EXTREME_DEGRADATION_PROFILE_MULTIPLIER;
     		noise1Mean  = EXTREME_DEGRADATION_PROFILE_NOISE1_MEAN;
     		noise1Sigma = EXTREME_DEGRADATION_PROFILE_NOISE1_SIGMA;
     		sigma_x     = EXTREME_DEGRADATION_PROFILE_SIGMA_X;
     		sigma_y     = EXTREME_DEGRADATION_PROFILE_SIGMA_Y;
     		noise2Mean  = EXTREME_DEGRADATION_PROFILE_NOISE2_MEAN;
     		noise2Sigma = EXTREME_DEGRADATION_PROFILE_NOISE2_MEAN;
     		break;
     	case CUSTOM:
     		break;
     }
     
     Mat Image_Uint = imread(path.c_str());
     Mat Image;
     
     uint64 seed = (uint64) generateHash(path);
     
     Image_Uint.convertTo(Image, CV_32F);
     
     //First, darken and add noise
     Image = Image * multiplier;
     addNoise(Image, noise1Mean, noise1Sigma, seed++);
     
     //Blur Image
     GaussianBlur(Image, Image, Size(0,0), sigma_x, sigma_y);
     
     //Add more noise
     addNoise(Image, noise2Mean, noise2Sigma, seed++);
     
     //Save image
     imwrite(path.c_str(), Image);
}
