#include "opencv2/opencv.hpp"
#include <eigen3/Eigen/Eigen>
#include <sys/time.h>
#include <cstdint>
#include <stdio.h>
#include <pthread.h>
#include <sys/select.h>
#include <termios.h>
#include <queue>
#include <fcntl.h>
#include <unistd.h>
#include <array>

//Local Includes
#include "gyroManager.h"
#include "Utilities.h"
#include "Benchmark.h"

using namespace std;
using namespace cv;
using namespace Eigen;
using namespace Utilities;

//#define DEBUG
//#define GYRO_SIGN_FLIP
#define BAUDRATE B115200
#define IntegrationIntervals 50

namespace gyro{
	//POD struct for holding a single gyro measurement
	struct measurement {
		double gyroX;
		double gyroY;
		double gyroZ;
		double HWTime;
		struct timeval compRXTime;
	};
	
	//Union used in bit-pattern reinterpretation utility function.
	union reinterpretationUnion_uint64_double {
		uint64_t data_uint64;
		double data_double;
	};
	
	//Union used in bit-pattern reinterpretation utility function.
	union reinterpretationUnion_uint32_int32 {
		uint32_t data_uint32;
		int32_t  data_int32;
	};
	
	//Union used in bit-pattern reinterpretation utility function.
	union reinterpretationUnion_uint16_int16 {
		uint16_t data_uint16;
		int16_t  data_int16;
	};
	
	int portFD;  //Handle for serial port
	bool gyroSystemRunning = false;
	
	//Raw buffers and thread management variables
	queue<unsigned char> RXQueue; //Buffer will always be kept at just a few bytes in size.
     pthread_t RXManagerThread;
     pthread_mutex_t measurementListMUTEX;
     bool KillRXThreadRequest = false;
     pthread_mutex_t KillRXThreadRequestMUTEX;
	
	//Measurement Buffer
	//This message list acts as a buffer for recent gyro measurements. New messages are pushed onto the front of this list.
	//When the list reaches the max size, old messages are deleted off the back when new messages pushed onto the front.
	//Rotations can only be computed over time spans that are entirely contained within this buffer at the time of request.
	//Thus, this buffer should be large enough to allow for the worst-case latancy in processing.
	int measurementListSize = 500;
	list<struct measurement> measurementList;
	
	//Storage buffers for saving and returning raw measurements (for latency estimation)
	bool recordingInProgress = false;
	list<Vector3d> savedMeasurements;
	list<struct timeval> savedMeasurementSampleTimes;
	
	//Gyro bias corrections - these are applied when a measurement is constructed, not when it is used.
	//These values are added to raw measurements to generate corrected measurements.
	double calibration_bx = 0.0;
	double calibration_by = 0.0;
	double calibration_bz = 0.0;
	double calibration_latency = 0.0;
	
	//Forward declarations for private functions
	int sendPacket(unsigned char * buffer, int bytesToSend);
	int launchRXManager(void);
	void * RXManager(void *);
	void processRXBuffer(void);
	double reinterpret_uint64_double(uint64_t data);
	int32_t reinterpret_uint32_int32(uint32_t data);
	int16_t reinterpret_uint16_int16(uint16_t data);
	int getNumberOfReadingsInBuffer(void);
	void resample(vector<double>& ts, vector<double>& Xs, vector<double>& Ys, vector<double>& Zs, array<double, IntegrationIntervals>& new_ts, array<double, IntegrationIntervals>& new_Xs, array<double, IntegrationIntervals>& new_Ys, array<double, IntegrationIntervals>& new_Zs);
	double interp1(double t1, double x1, double t2, double x2, double t);
	void interp3(double t1, double x1, double y1, double z1, double t2, double x2, double y2, double z2, double t, double& x, double& y, double& z);
	
	double reinterpret_uint64_double(uint64_t data) {
		//Utility function for re-interpreting a bit pattern. This re-interprets data in a uint64 as a double.
		reinterpretationUnion_uint64_double tempdata;
		tempdata.data_uint64 = data;
		return(tempdata.data_double);
	}
	
	int32_t reinterpret_uint32_int32(uint32_t data) {
		//Utility function for re-interpreting a bit pattern. This re-interprets data in a uint32 as an int32.
		reinterpretationUnion_uint32_int32 tempdata;
		tempdata.data_uint32 = data;
		return(tempdata.data_int32);
	}
	
	int16_t reinterpret_uint16_int16(uint16_t data) {
		//Utility function for re-interpreting a bit pattern. This re-interprets data in a uint16 as an int16.
		reinterpretationUnion_uint16_int16 tempdata;
		tempdata.data_uint16 = data;
		return(tempdata.data_int16);
	}
	
	void setGyroBiasCorrections(double bx, double by, double bz) {
		//Set new gyro bias corrections. We need to lock the measurement list. The same mutex controls
		//access to calibration values.
		pthread_mutex_lock(&measurementListMUTEX);
		calibration_bx = bx;
		calibration_by = by;
		calibration_bz = bz;
		pthread_mutex_unlock(&measurementListMUTEX);
	}
	
	void getGyroBiasCorrections(double &bx, double &by, double &bz) {
		//Read and return gyro bias corrections. We need to lock the measurement list. The same mutex controls
		//access to calibration values.
		pthread_mutex_lock(&measurementListMUTEX);
		bx = calibration_bx;
		by = calibration_by;
		bz = calibration_bz;
		pthread_mutex_unlock(&measurementListMUTEX);
	}
	
	double getGyroLatencyCorrection(void) {
		//Read and return gyro system latency estimate being used. We need to lock the measurement list.
		//The same mutex controls access to calibration values.
		double latched_calibration_latency;
		pthread_mutex_lock(&measurementListMUTEX);
		latched_calibration_latency = calibration_latency;
		pthread_mutex_unlock(&measurementListMUTEX);
		return(latched_calibration_latency);
	}
	
	void setGyroLatencyCorrection(double systemLatency) {
		//Set new gyro system latency estimate. We need to lock the measurement list.
		//The same mutex controls access to calibration values.
		pthread_mutex_lock(&measurementListMUTEX);
		calibration_latency = systemLatency;
		pthread_mutex_unlock(&measurementListMUTEX);
	}
	
	void startRecordingReadings(void) {
		//Start recording raw measurements. Don't forget to  stop recording.
		//We must lock the measurement list. The same mutex governs access to the recording control variable.
		pthread_mutex_lock(&measurementListMUTEX);
		recordingInProgress = true;
		savedMeasurements.clear();
		savedMeasurementSampleTimes.clear();
		pthread_mutex_unlock(&measurementListMUTEX);
	}
	
	void stopAndReturnMeasurements(std::vector<struct timeval> &sampleTimes, std::vector<Eigen::Vector3d> &measurements) {
		//Stop recording raw measurements (don't halt gyro subsystem - just stop saving old measurements).
		//Return buffered measurements. We must lock the measurement list. The same mutex governs access to the
		//recording control variable and the measurement buffers.
		pthread_mutex_lock(&measurementListMUTEX);
		sampleTimes  = vector<struct timeval>(savedMeasurementSampleTimes.begin(), savedMeasurementSampleTimes.end());
		measurements = vector<Vector3d>(savedMeasurements.begin(), savedMeasurements.end());
		recordingInProgress = false;
		pthread_mutex_unlock(&measurementListMUTEX);
	}
	
	int openPort(const char * devicePath) {
		//Open the serial port and start collecting/decoding gyro readings
		//devicePath is the file path of the serial port. The baudrate is specified in a define statement.
		struct termios config;

		portFD = open(devicePath, O_RDWR | O_NOCTTY);
		if(portFD == -1) return(-1);

		bzero(&config, sizeof(config));

		config.c_cflag = (BAUDRATE | CS8 | CLOCAL | CREAD | INPCK | ISTRIP);
		config.c_iflag = 0;
		config.c_oflag = 0;
		config.c_lflag = 0;
		config.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);	//Ensure Device is in RAW mode

		config.c_cc[VMIN] = 0;				//Set Port to timeout control
		config.c_cc[VTIME] = 1;				//Set read timeout to 1/10 seconds

		tcflush(portFD, TCIFLUSH);			//Flush Buffers
		tcsetattr(portFD, TCSANOW, &config);	//Apply New Settings Immediately

		pthread_mutex_init(&measurementListMUTEX, NULL);
		pthread_mutex_init(&KillRXThreadRequestMUTEX, NULL);

		launchRXManager();
		gyroSystemRunning = true;

		return(0);                         //Return Success Code
	}
	
	int closePort(void) {
		//Close port and stop receiving/decoding gyro measurements.
		if (gyroSystemRunning) {
			pthread_mutex_lock(&KillRXThreadRequestMUTEX);
			KillRXThreadRequest = true;
			pthread_mutex_unlock(&KillRXThreadRequestMUTEX);
			pthread_join(RXManagerThread, NULL);
			close(portFD);
		}
		return(0);                         //Return Success Code
	}
	
	int sendPacket(unsigned char * buffer, int bytesToSend) {
		//This function sends a packet out over the serial port.
		int numberOfBytesWritten;
		numberOfBytesWritten = write(portFD, buffer, bytesToSend);
		if (numberOfBytesWritten == bytesToSend) return(0);         //Return Success Code
		else return(-1);                                            //Return Failure Code
	}
	
	int resetDevice(void) {
		//Reset gyro module. Port must be open when this function is called.
		unsigned char command[] = "\xA1\x1F\xF4\x4B\x00";
		return(sendPacket(command, 5));
	}
	
	int setDivider(unsigned char divider) {
		//Set gyro Rate divider (1 gives fastest rate, 2 gives half the fatsest rate, etc.)
		//Port must be open when this function is called.
		if (divider == 0) return(-1);
		unsigned char command[] = "\xA1\x1F\xF4\x4B\x00";
		command[4] = divider;
		return(sendPacket(command, 5));
	}
	
	int launchRXManager(void) {
          //This function launches the RX manager thread. The function returns immediately.
          int returnCode;
          returnCode = pthread_create(&RXManagerThread, NULL, RXManager, NULL);
          return(returnCode);
     }

     void * RXManager(void *) {
          //This function is meant to be run in its own thread of execution. It reads data from the serial port and
          //buffers it in a queue. Examining and processing data in this queue should be done elsewhere.
          int bytesRead;
          bool KillFlag = false;
          unsigned char byte;
          int local_portFD = portFD;
          //fd_set FDset;            //A set containing only our file descriptor. Used for the select function (sleep until data ready)
          //FD_ZERO(&FDset);         //Clear the set
          //FD_SET(portFD, &FDset);  //Add our file descriptor to the set
          printf("Launching RX Manager Thread...\n\r");
          while(1) {  //This loop runs until the program terminates or until an exit request is made
               //First check to see if a kill request has been made
               pthread_mutex_lock(&KillRXThreadRequestMUTEX);
               KillFlag = KillRXThreadRequest;
               pthread_mutex_unlock(&KillRXThreadRequestMUTEX);
               if (KillFlag) break;
               //Read a byte from the serial port
               bytesRead = read(local_portFD, (void *)(&byte), 1);
               if (bytesRead == 1) {
            	   RXQueue.push(byte);
            	   processRXBuffer();
               }
          }
          printf("RX Manager Thread Exiting...\n\r");
          return(NULL);
     }
     
     void processRXBuffer(void) {
     	//This function parses the raw RX buffer and inserts measurements into the measuremnt list.
     	//This list is controlled with a mutex so it can be accessed from multiple threads.
     	static unsigned char packetAssemblyBuf[22];
     	static int index = 0;
     	
     	while (!RXQueue.empty()) {
     		unsigned char byte = RXQueue.front();
     		
     		if (index == 0) {
     			if (byte == (unsigned char)0x30) {
     				packetAssemblyBuf[index++] = byte;
     				RXQueue.pop();
     			}
     			else
     				RXQueue.pop();
     		}
     		else if (index == 1) {
     			if (byte == (unsigned char)0x66) {
     				packetAssemblyBuf[index++] = byte;
     				RXQueue.pop();
     			}
     			else
     				index = 0;
     		}
     		else if (index == 2) {
     			if (byte == (unsigned char)0x3E) {
     				packetAssemblyBuf[index++] = byte;
     				RXQueue.pop();
     			}
     			else
     				index = 0;
     		}
     		else if (index == 3) {
     			if (byte == (unsigned char)0xC6) {
     				packetAssemblyBuf[index++] = byte;
     				RXQueue.pop();
     			}
     			else
     				index = 0;
     		}
     		else {
     			packetAssemblyBuf[index++] = byte;
     			RXQueue.pop();
     		}
     		
     		if (index == 22) {
     			//Assembly buffer is full with a new measurement
     			struct timeval compRXTime;
     			gettimeofday(&compRXTime, NULL);
     			
     			struct measurement newMeasurement;
     			
     			uint16_t temp16;
     			uint32_t temp32;
     			uint64_t multTemp;
     			
     			double HWTime, gyroX, gyroY, gyroZ, multiplier;
     			
     			//Decode sensor-reported hardware time
     			temp32  =  (uint32_t) (packetAssemblyBuf[4]);
     			temp32 += ((uint32_t) (packetAssemblyBuf[5])) << 8;
     			temp32 += ((uint32_t) (packetAssemblyBuf[6])) << 16;
     			temp32 += ((uint32_t) (packetAssemblyBuf[7])) << 24;
     			HWTime = ((double)temp32)*0.001;
     			
     			//Decode X measurement
     			temp16  =  (uint16_t) (packetAssemblyBuf[8]);
     			temp16 += ((uint16_t) (packetAssemblyBuf[9])) << 8;
     			gyroX = (double)reinterpret_uint16_int16(temp16);
     			
     			//Decode Y measurement
     			temp16  =  (uint16_t) (packetAssemblyBuf[10]);
     			temp16 += ((uint16_t) (packetAssemblyBuf[11])) << 8;
     			gyroY = (double)reinterpret_uint16_int16(temp16);
     			
     			//Decode Z measurement
     			temp16  =  (uint16_t) (packetAssemblyBuf[12]);
     			temp16 += ((uint16_t) (packetAssemblyBuf[13])) << 8;
     			gyroZ = (double)reinterpret_uint16_int16(temp16);
     			
     			//Decode multiplier
     			multTemp  =  (uint64_t) (packetAssemblyBuf[14]);
     			multTemp += ((uint64_t) (packetAssemblyBuf[15])) << 8;
     			multTemp += ((uint64_t) (packetAssemblyBuf[16])) << 16;
     			multTemp += ((uint64_t) (packetAssemblyBuf[17])) << 24;
     			multTemp += ((uint64_t) (packetAssemblyBuf[18])) << 32;
     			multTemp += ((uint64_t) (packetAssemblyBuf[19])) << 40;
     			multTemp += ((uint64_t) (packetAssemblyBuf[20])) << 48;
     			multTemp += ((uint64_t) (packetAssemblyBuf[21])) << 56;
     			multiplier = reinterpret_uint64_double(multTemp);
     			
     			#ifdef GYRO_SIGN_FLIP
     			multiplier = (-1.0)*multiplier;
     			#endif
     			
     			//Populate new measurement and append to measurement list
     			//We must apply the multiplier here and apply calibration corrections.
     			pthread_mutex_lock(&measurementListMUTEX);
     			newMeasurement.HWTime     = HWTime;
     			newMeasurement.gyroX      = gyroX*multiplier + calibration_bx;
     			newMeasurement.gyroY      = gyroY*multiplier + calibration_by;
     			newMeasurement.gyroZ      = gyroZ*multiplier + calibration_bz;
     			newMeasurement.compRXTime = compRXTime;
     			
     			//Correct for estimated system latency
				moveBackwardsInTime(newMeasurement.compRXTime, calibration_latency);
     			
     			measurementList.push_front(newMeasurement);
     			if ((int)(measurementList.size()) > measurementListSize)
     				measurementList.pop_back();
     			
     			if (recordingInProgress) {
     				savedMeasurements.push_back(Vector3d(newMeasurement.gyroX, newMeasurement.gyroY, newMeasurement.gyroZ));
					savedMeasurementSampleTimes.push_back(newMeasurement.compRXTime);
     			}
     				
     			pthread_mutex_unlock(&measurementListMUTEX);
     			
     			index = 0; //Reset assembly buffer
     			
     			#ifdef DEBUG
     			fprintf(stderr,"(% 7.3f, % 8.6f, % 8.6f, % 8.6f)\n\r", HWTime, newMeasurement.gyroX, newMeasurement.gyroY, newMeasurement.gyroZ);
     			#endif
     		}
     	}
     }
     
	int getRotationBetweenTwoTimes(struct timeval start, struct timeval end, Matrix3d& R, double& rotationAngle, bool wait) {
		//This function re-samples the gyro data stream between the provided start and end times
		//and performs quaternion ODE integration to estimate the total rotation between the start
		//and end times. The rotation angle is returned along with a direction-cosine matrix for the rotation.
		//If R is not a 3x3 matrix with type CV_64F, it will be re-allocated as such.
		//If wait is true, then we will wait until the measurement buffer has at least 1 measurement from after the end time
		//before processing. If wait is false, we will immediately begin processing, using extrapolation if necessary to fill
		//in for measurements that have not yet been received. If measurement buffer does not contain measurements older than
		//the start time, then the measurement buffer is too small. The function will fail and return -1.
		//If the provided start time comes after the provided end time, the function fails and returns -2.
		//If the measurement buffer is empty, this function will wait for up to 1 second to get new measurements. If none are
		//received the function will fail and return -3. If wait=true and after waiting a full second, all measurements in the
		//measurement buffer are still all older than the provided end time, the function will fail and return -4.
		//Other errors will return -5.
		//
		//If R is returned by this function, and v1 is a vector in the coordinate frame of the gyro at time 1, then R*v1
		//gives the same vector in a coordinate frame with the same origin as the first, but with axes aligned with the
		//gyro axes at time 2.
		
		struct timeval newestMeasurementTime;
		struct timeval oldestMeasurementTime;
		
		//Make sure we have some measurements. Wait for up to 1 second to get measurements
		if (getNumberOfReadingsInBuffer() < 2) {
			bool waitSuccessful = false;
			for (int waitTimer = 0; waitTimer < 100; waitTimer++) {
				usleep(10000U); //Wait 1/100 of a second.
				if (getNumberOfReadingsInBuffer() >= 2) {
					waitSuccessful = true;
					break;
				}
			}
			if (!waitSuccessful) {
				fprintf(stderr,"getRotationBetweenTwoTimes: Measurement buffer has less than 2 items and did not receive any new measurements for 1 second.\n\r");
				return(-3);
			}
		}
		
		//Check to make sure we have messages older than start time
		pthread_mutex_lock(&measurementListMUTEX);
		oldestMeasurementTime = measurementList.back().compRXTime;
		pthread_mutex_unlock(&measurementListMUTEX);
		if (isFirstNewer(oldestMeasurementTime,start)) {
			fprintf(stderr,"requesting rotation over timespan outside of current buffer.\n\r");
			fprintf(stderr,"Desired start time is %f seconds older than oldest measurement in buffer.\n\r", timeDiff(oldestMeasurementTime, start));
			fprintf(stderr,"There are %d measurements in the buffer.\n\r", getNumberOfReadingsInBuffer());
			return(-1);
		}
		
		//Check that end time comes after start time
		if (isFirstNewer(start,end)) return(-2);
		
		if (wait) {
			bool waitSuccessful = false;
			pthread_mutex_lock(&measurementListMUTEX);
			newestMeasurementTime = measurementList.front().compRXTime;
			pthread_mutex_unlock(&measurementListMUTEX);
			
			if (isFirstNewer(end, newestMeasurementTime)) {
				//Wait for newer measurements - wait for up to a second
				for (int waitTimer = 0; waitTimer < 100; waitTimer++) {
					usleep(5000U); //Wait 1/200 of a second.
					pthread_mutex_lock(&measurementListMUTEX);
					newestMeasurementTime = measurementList.front().compRXTime;
					pthread_mutex_unlock(&measurementListMUTEX);
					
					if (isFirstNewer(newestMeasurementTime, end)) {
						waitSuccessful = true;
						break;
					}
				}
				if (!waitSuccessful) {
					fprintf(stderr,"getRotationBetweenTwoTimes: Waited 1 second for newer gyro readings, but never reached desired end time.\n\r");
					return(-4);
				}
			}
			
			//If we make it this far, we can proceed with pure interpolation
			//Extract sublist with relavent measurements
			list<struct measurement> relaventMeasurements;
			bool copyInProgress = false;
			pthread_mutex_lock(&measurementListMUTEX);
			
			//The measurement list should have at least 2 elements in it, so the following is safe.
			list<struct measurement>::iterator it = measurementList.begin();
			list<struct measurement>::iterator nextit = it;
			nextit++;
			
			while((it != measurementList.end()) && (nextit != measurementList.end())) {
				if (!copyInProgress) {
					nextit = it;
					nextit++;
					if (isFirstNewer(it->compRXTime, end) && (!isFirstNewer(nextit->compRXTime, end))) {
						relaventMeasurements.push_front(*it);
						copyInProgress = true;
					}
				}
				else {
					if (isFirstNewer(start, it->compRXTime)) {
						relaventMeasurements.push_front(*it);
						break;
					}
					else {
						relaventMeasurements.push_front(*it);
					}
				}
				it++;
				nextit++;
			}
			pthread_mutex_unlock(&measurementListMUTEX);
			
			//Pack sublist into vector with shifted timebase so first measurement corresponds to time 0
			struct timeval time0 = relaventMeasurements.front().compRXTime;
			
			vector<double> measured_ts(relaventMeasurements.size());
			vector<double> measured_Xs(relaventMeasurements.size());
			vector<double> measured_Ys(relaventMeasurements.size());
			vector<double> measured_Zs(relaventMeasurements.size());
			
			int arrayIndex = 0;
			for (list<struct measurement>::iterator it = relaventMeasurements.begin(); it != relaventMeasurements.end(); it++) {
				measured_ts[arrayIndex] = timeDiff(it->compRXTime, time0);
				measured_Xs[arrayIndex] = it->gyroX;
				measured_Ys[arrayIndex] = it->gyroY;
				measured_Zs[arrayIndex] = it->gyroZ;
			}
			
			//Re-sample between start and end to get lots of uniformly spaced measurements for quaternion integration
			array<double, IntegrationIntervals> ts;
			array<double, IntegrationIntervals> Xs;
			array<double, IntegrationIntervals> Ys;
			array<double, IntegrationIntervals> Zs;
			
			//We resample in the middle of each integration interval.
			double deltaT = timeDiff(end, start)/IntegrationIntervals;
			double startTime = timeDiff(start, time0) + deltaT/2;
			for (int i=0; i<IntegrationIntervals; i++) {
				ts[i] = startTime + i*deltaT;
			}
			resample(measured_ts, measured_Xs, measured_Ys, measured_Zs, ts, Xs, Ys, Zs);
			
			//Initialize quaternion and integrate resampled measurements to get final rotaion quaternion
			Quaternion<double> Q(1.0, 0.0, 0.0, 0.0);
			Quaternion<double> P_rb_b;
			for (int i=0; i<IntegrationIntervals; i++) {
				P_rb_b = Quaternion<double>(0.0, Xs[i], Ys[i], Zs[i]);
				Q.coeffs() += (0.5)*deltaT*(P_rb_b*Q).coeffs();
				Q.normalize();
			}
			
			//Extract direction-cosine matrix from final quaternion and rotation angle.
			R = Q.toRotationMatrix().transpose();
			rotationAngle = 2.0*atan2(sqrt(Q.x()*Q.x() + Q.y()*Q.y() + Q.z()*Q.z()),Q.w());
		}
		else {
			//Proceed with interpolation and extrapolation (if necessary)
			fprintf(stderr,"getRotationBetweenTwoTimes: wait=false is not yet supported. Run with wait=true.\n\r");
			return(-5);
		}
		return(0);
	}
	
	void resample(vector<double>& ts, vector<double>& Xs, vector<double>& Ys, vector<double>& Zs, array<double, IntegrationIntervals>& new_ts, array<double, IntegrationIntervals>& new_Xs, array<double, IntegrationIntervals>& new_Ys, array<double, IntegrationIntervals>& new_Zs) {
		//Perform linear interpolation to re-sample a time series (in R^3) at new time instants. Input arrays must not be empty or the function will fail and do nothing.
		//If resample times do not fall within between 2 values in "ts", then they are extrapolated with nearest neighbor extrapolation.
		//ts and new_ts must both be weakly monotonically increasing, or the output will be incorrect (although the function will execute and return normally).
		if (ts.size() == 0) {
			fprintf(stderr,"resample: Function called with empty input arrays. Aborting function.\n\r");
			return;
		}
		
		int newDataIndex = 0;
		int leftIndex = 0;
		
		while (newDataIndex < ((int) (new_ts.size()))) {
			double t = new_ts[newDataIndex];
			if (t < ts[leftIndex]) {
				//We must extrapolate the value on the left - use nearest neighbor
				new_Xs[newDataIndex] = Xs[0];
				new_Ys[newDataIndex] = Ys[0];
				new_Zs[newDataIndex] = Zs[0];
				newDataIndex++;
			}
			else if (leftIndex >= (((int) (ts.size()))-1)) {
				//We must extrapolate the value on the right - use nearest neighbor
				new_Xs[newDataIndex] = Xs[ts.size()-1];
				new_Ys[newDataIndex] = Ys[ts.size()-1];
				new_Zs[newDataIndex] = Zs[ts.size()-1];
				newDataIndex++;
			}
			else if (t <= ts[leftIndex+1]) {
				//Interpolate using leftIndex and leftIndex+1
				interp3(ts[leftIndex],   Xs[leftIndex],        Ys[leftIndex],        Zs[leftIndex],
				        ts[leftIndex+1], Xs[leftIndex+1],      Ys[leftIndex+1],      Zs[leftIndex+1],
				        t,               new_Xs[newDataIndex], new_Ys[newDataIndex], new_Zs[newDataIndex]);
				newDataIndex++;
			}
			else {
				leftIndex++;
			}	
		}
	}
	
	double interp1(double t1, double x1, double t2, double x2, double t) {
		//Linearly interpolate between 2 data points (t1,x1) and (t2,x2). t must be between t1 and t2 (possibly equal to one of them).
		double T = t2 - t1;
		if (T < 0.0000001) return(0.5*x1 + 0.5*x2);
		else return(((t2-t)*x1 + (t-t1)*x2)/T);
	}
	
	void interp3(double t1, double x1, double y1, double z1, double t2, double x2, double y2, double z2, double t, double& x, double& y, double& z) {
		//Linearly interpolate between 2 data points (range in R^3). t must be between t1 and t2 (possibly equal to one of them).
		double T = t2 - t1;
		double a, b;
		if (T < 0.0000001) {
			a = 0.5;
			b = 0.5;
		}
		else {
			a = (t2-t)/T;
			b = (t-t1)/T;
		}
		x = a*x1 + b*x2;
		y = a*y1 + b*y2;
		z = a*z1 + b*z2;
	}
	
	int averageReadings(int numberOfSeconds, double& Xavg, double& Yavg, double& Zavg) {
		//Flush measurement buffer
		pthread_mutex_lock(&measurementListMUTEX);
		measurementList.clear();
		pthread_mutex_unlock(&measurementListMUTEX);
		
		struct timeval startTime, currentTime;
		gettimeofday(&startTime, NULL);
		
		//Initialize variables
		int numberOfMeasurements = 0;
		Xavg = 0.0;
		Yavg = 0.0;
		Zavg = 0.0;
		
		while (true) {
			gettimeofday(&currentTime, NULL);
			if (timeDiff(currentTime,startTime) >= ((double) numberOfSeconds)) break;
			
			pthread_mutex_lock(&measurementListMUTEX);
			for (list<struct measurement>::iterator it=measurementList.begin(); it != measurementList.end(); it++) {
				Xavg += it->gyroX;
				Yavg += it->gyroY;
				Zavg += it->gyroZ;
				numberOfMeasurements++;
			}
			measurementList.clear();
			pthread_mutex_unlock(&measurementListMUTEX);
			
			usleep(10000U); //Wait 1/100 of a second.
		}
		Xavg = Xavg/numberOfMeasurements;
		Yavg = Yavg/numberOfMeasurements;
		Zavg = Zavg/numberOfMeasurements;
		
		return(numberOfMeasurements);
	}
	
	int getNumberOfReadingsInBuffer(void) {
		//Returns the number of readings in the measurement buffer
		int size;
		pthread_mutex_lock(&measurementListMUTEX);
		size = measurementList.size();
		pthread_mutex_unlock(&measurementListMUTEX);
		return(size);
	}
	
	bool receivingGyroData(void) {
		//Returns true if we have received new data in the last second, and false otherwise.
		struct timeval newestMeasurementTime, currentTime;
		gettimeofday(&currentTime, NULL);
		
		pthread_mutex_lock(&measurementListMUTEX);
		newestMeasurementTime = measurementList.front().compRXTime;
		pthread_mutex_unlock(&measurementListMUTEX);
		
		if (fabs(timeDiff(newestMeasurementTime,currentTime)) <= 1.0) return(true);
		else return(false);
	}
}
