#include "opencv2/opencv.hpp"
#include <eigen3/Eigen/Eigen>
#include <sys/time.h>

namespace gyro{
	//Set/get the bias corrections for the gyro. These values are added to the raw measurements to generate corrected measurements.
	//Changes to the bias corrections take effect immediately. Warning: This overwrites the previous bias corrections - it does not
	//add to them. Thus, if you plan on running this function more than once during the execution of the program, then on all
	//calibration runs after the first, first clear bias corrections (using this function), then average collected measurements,
	//then use this function to apply new terms. If you do not clear the bias corrections first, you will not be observing raw
	//measurements when you estimate new corrections.
	void setGyroBiasCorrections(double  bx, double  by, double  bz);
	void getGyroBiasCorrections(double &bx, double &by, double &bz);
	
	//Set/get estimated gyro system latency. The RX timestamps of received measurements will be set backwards in time by this many
	//seconds when they are recorded. The HWTime for each measurement is not altered. If estimating system latency, it is important
	//that this correction be cleared first.
	double getGyroLatencyCorrection(void);
	void   setGyroLatencyCorrection(double systemLatency);
	
	//Open the serial port and start receiving/decoding gyro readings.
	int openPort(const char * devicePath);
	
	//Close port and stop receiving/decoding gyro readings.
	int closePort(void);
	
	//Reset the gyro. Port must be open when this function is called.
	int resetDevice(void);
	
	//Set gyro divider. Setting this to 1U results in the maximum possible measurement rate. 0U is not allowed.
	//The port must be open when this function is called.
	int setDivider(unsigned char divider);
     
     //This function re-samples the gyro data stream between the provided start and end times and performs quaternion ODE integration
     //to estimate the total rotation between the start and end times. The rotation angle is returned along with a direction-cosine
     //matrix for the rotation. If R is not a 3x3 matrix with type CV_64F, it will be re-allocated as such.
	int getRotationBetweenTwoTimes(struct timeval start, struct timeval end, Eigen::Matrix3d& R, double& rotationAngle, bool wait);
	
	//Record and average together (time-average) gyro readings for the provided number of seconds. Upon completion, return the number
	//of measurements that were collected and averaged. This function starts by flushing the measurement buffer so only readings
	//received after this function is called are used. The function will only return after "numberOfSeconds" seconds have elapsed.
	int averageReadings(int numberOfSeconds, double& Xavg, double& Yavg, double& Zavg);
	
	//Returns true if we have received new data in the last second, and false otherwise.
	bool receivingGyroData(void);
	
	//The following functions record and give access to raw timestamped measurements. These are used, for example, to
	//estimate the gyro HW/SW system latency. Functions are also provided to set system latency corrections.
	void startRecordingReadings(void);
	void stopAndReturnMeasurements(std::vector<struct timeval> &sampleTimes, std::vector<Eigen::Vector3d> &measurements);
}
