//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>

//Internal Includes
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_GYROPATH "/dev/ttyACM0"
#define DEFAULT_VIDEOCAM_NUM 1

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//If this program is run without arguments, it looks to a hard-coded default path for the gyro module: /dev/ttyACM0
//If this program is run with a single argument, that argument should be the device path of the serial gyro module.

//Forward Declarations
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);
void displayHelp(void);

//Constants
const double pi = 3.141592653589793;

namespace parameters {
	int  trackingScheme = 1;
	bool replaceLostFeatures = true;
}

int main(int argc, char** argv) {
     int key, retval;
     Mat temp, frame, frameBW, frameWithFeatures;
     
     string GyroDevicePath;
     int cameraOpenCVIndex = 0;
     
     //Parse command-line options
	int gflag = 0;
	int vflag = 0;
	int hflag = 0;
	char *gval = NULL;
	char *vval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "g:v:h")) != -1)
		switch (c) {
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'v':
				vflag = 1;
				vval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}

	if (gflag == 1)
		GyroDevicePath = string(gval);
	else
		GyroDevicePath = string(DEFAULT_GYROPATH);
	
	if (vflag == 1)
		cameraOpenCVIndex = atoi(vval);
	else
		cameraOpenCVIndex = DEFAULT_VIDEOCAM_NUM;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
     
     //Initialize Tracker Output Window
     namedWindow("Live Display", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	//Start Gyro Subsystem
	int returnCode;
	returnCode = gyro::openPort(GyroDevicePath.c_str());
	if (returnCode != 0) {
		fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Set Gyro divider
	usleep(100000U); //Wait 1/10 of a second.
	returnCode = gyro::setDivider(1U);
	if (returnCode != 0) {
		fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Check that sensor is giving us data
	usleep(100000U); //Wait 1/10 of a second.
	//sleep(1U); //Wait 1 second.
	if (!gyro::receivingGyroData()) {
		fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Disable autofocus or auto-exposure for camera (if desired) - Linux and V4L only
	Utilities::adjustCameraSettings(cap, "Live Display", cameraOpenCVIndex);
	
	//Load Gyro calibration data
	string CalibrationFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Loading Calibration Data from file: %s... ", CalibrationFileName.c_str());
	Vector3d b;
	double Lgyro;
	Matrix3d Ktilde, KtildeInv;
	double sigma;
	retval = loadCalibrationData(CalibrationFileName, b, Ktilde, sigma, Lgyro);
	KtildeInv = Ktilde.inverse();
	if (retval == 0) fprintf(stderr,"Success\n\r");
	else {
		fprintf(stderr,"Failure -> Aborting...\n\r");
		gyro::closePort();
		return(-1);
	}
	
	//Set gyro bias and latency corrections
	fprintf(stderr,"Setting gyro bias and latency corrections... ");
	gyro::setGyroBiasCorrections(b(0), b(1), b(2));
	gyro::setGyroLatencyCorrection(Lgyro);
	fprintf(stderr,"Done.\n\r");
	
	//Render display with feature locations estimated based on gyro readings only
	fprintf(stderr,"Rendering live display. Press i to re-initialize features.\n\rPress spacebar to exit.\n\r");
	struct timeval oldTime, newTime;
	Mat oldFrame, newFrame;
	getFrameWithTimestamp(cap, newFrame, newTime);
	vector<Point2f> points1, points2;
	Vector3d point1Hom, point2Hom;
	Matrix3d H, R;
	double rotationAngle;
	point1Hom(2) = 1.0;
	Point2f pointA, pointB;
	Scalar color_point(255,0,0);
	Scalar color_line(0,0,255);
	Mat newFrameColor;
	int maxNumberOfFeatures = 200;
	
	//Extract Features
	goodFeaturesToTrack(newFrame, points2, maxNumberOfFeatures, 0.001, 3.0);
	normalizeImageCoords(newFrame.cols, newFrame.rows, points2);
	points1.resize((int) points2.size());
	
	while (true) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		points1.swap(points2);
		getFrameWithTimestamp(cap, newFrame, newTime);
		
		retval = gyro::getRotationBetweenTwoTimes(oldTime, newTime, R, rotationAngle, true);
		if (retval != 0) {
			fprintf(stderr,"gyro::getRotationBetweenTwoTimes failed with return code %d", retval);
			gyro::closePort();
			return(-1);
		}
		
		//Estimate Homography between 2 frames
		H = Ktilde*R*KtildeInv;
		
		//propogate feature positions based on estimated homographies (in color)
		for (int i=0; i<((int) points1.size()); i++) {
			point1Hom(0) = points1[i].x;
			point1Hom(1) = points1[i].y;
		
			point2Hom = H*point1Hom;
			points2[i].x = point2Hom(0)/point2Hom(2);
			points2[i].y = point2Hom(1)/point2Hom(2);
		}
		
		//Render predicted flow
		cvtColor(newFrame, newFrameColor, CV_GRAY2RGB);
		for (int i=0; i<((int) points1.size()); i++) {
			pointA = points1[i];
			pointB = points2[i];
			
			denormalizeImageCoords(newFrame.cols, newFrame.rows, pointA);
			denormalizeImageCoords(newFrame.cols, newFrame.rows, pointB);
			
			//Render Circle at current point and flow line
			circle(newFrameColor, pointB, 3, color_point, 2);
			//line(newFrameColor, pointA, pointB, color_line, 2);
		}
		
		//Show frame and check for stop condition (space bar)
		imshow("Live Display", newFrameColor);
		key = waitKey(1);
		if (key == ' ') break;
		else if (key == 'i') {
			//Re-initialize features
			goodFeaturesToTrack(newFrame, points2, maxNumberOfFeatures, 0.001, 3.0);
			normalizeImageCoords(newFrame.cols, newFrame.rows, points2);
			points1.resize((int) points2.size());
			fprintf(stderr,"Features Re-initialized.\n\r");
		}
	}
	
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"******************************     Gyro-Predicted Flow Display     ****************************\n\r");
	fprintf(stderr,"This program initializes features with a camera and propogates their gyro-predicted flows. Their\n\r");
	fprintf(stderr,"propogated positions are super-imposed on the video stream. Certain parameters can be specified\n\r");
	fprintf(stderr,"through command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file. Default is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking. Default is %d\n\r\n\r", DEFAULT_VIDEOCAM_NUM);
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	cvtColor(frameColor, frame, CV_RGB2GRAY);
}
