//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include "boost/filesystem.hpp"

//Internal Includes
//#include "Tracker.h"
//#include "ImageStack.h"
//#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"
#include "rawData.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_GYROPATH "/dev/ttyACM0"
#define DEFAULT_VIDEOCAM_NUM 1

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;
using namespace boost::filesystem;

//If this program is run without arguments, it looks to a hard-coded default path for the gyro module: /dev/ttyACM0
//If this program is run with a single argument, that argument should be the device path of the serial gyro module.

//Forward Declarations
void displayHelp(void);
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);

//Constants
const double pi = 3.141592653589793;

namespace parameters {
	int  trackingScheme = 1;
	bool replaceLostFeatures = true;
}

int main(int argc, char** argv) {
     int key, retval;
     Mat temp, frame, frameBW, frameWithFeatures;
     string GyroDevicePath;
     int cameraOpenCVIndex = 0;
     
     //Parse command-line options
	int gflag = 0;
	int vflag = 0;
	int hflag = 0;
	char *gval = NULL;
	char *vval = NULL;

	int c;
	opterr = 0; //Do not print an error when unrecognized arguments are encountered

	while ((c = getopt (argc, argv, "g:v:h")) != -1)
		switch (c) {
			case 'g':
				gflag = 1;
				gval = optarg;
				break;
			case 'v':
				vflag = 1;
				vval = optarg;
				break;
			case 'h':
				hflag = 1;
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}

	if (gflag == 1)
		GyroDevicePath = string(gval);
	else
		GyroDevicePath = string(DEFAULT_GYROPATH);
	
	if (vflag == 1)
		cameraOpenCVIndex = atoi(vval);
	else
		cameraOpenCVIndex = DEFAULT_VIDEOCAM_NUM;
	
	//If help flag was passed, print help text and exit imediately
	if (hflag == 1) {
		displayHelp();
		return(0);
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
     
     //Initialize Tracker Output Window
     namedWindow("Live Display", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	//Start Gyro Subsystem
	int returnCode;
	returnCode = gyro::openPort(GyroDevicePath.c_str());
	if (returnCode != 0) {
		fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Set Gyro divider
	usleep(100000U); //Wait 1/10 of a second.
	returnCode = gyro::setDivider(1U);
	if (returnCode != 0) {
		fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Check that sensor is giving us data
	usleep(100000U); //Wait 1/10 of a second.
	//sleep(1U); //Wait 1 second.
	if (!gyro::receivingGyroData()) {
		fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
		gyro::closePort();
		return -1;
	}
	
	//Initialize Target Data Folder
	fprintf(stderr,"Creating target data folder and initializing data saver... ");
     rawDataSaver dataSaver;
     string CWD = current_path().string();
     path slash("/");
	string filesep = slash.make_preferred().native();
     string targetFolderPath = dataSaver.getNextAvailableFolderPath(CWD + filesep + string("Recordings") + filesep + string("Recording"));
     if (dataSaver.init(targetFolderPath) != 0) {
     	fprintf(stderr,"Failed\n\r");
     	fprintf(stderr,"Error initializing raw data saver with target data directory: %s\n\r",targetFolderPath.c_str());
     	return(-1);
     }
     fprintf(stderr,"Done\n\r");
	
	//Disable autofocus or auto-exposure for camera (if desired) - Linux and V4L only
	Utilities::adjustCameraSettings(cap, "Live Display", cameraOpenCVIndex);
	
	//Load Gyro calibration data
	string CalibrationFileName("GyroOpticalCalibration.json");
	fprintf(stderr,"Loading Calibration Data from file: %s... ", CalibrationFileName.c_str());
	Vector3d b;
	double Lgyro;
	Matrix3d Ktilde, KtildeInv;
	double sigma;
	retval = loadCalibrationData(CalibrationFileName, b, Ktilde, sigma, Lgyro);
	KtildeInv = Ktilde.inverse();
	if (retval == 0) fprintf(stderr,"Success\n\r");
	else {
		fprintf(stderr,"Failure -> Aborting...\n\r");
		gyro::closePort();
		return(-1);
	}
	
	//Copy current gyro-optical calibration data into target data folder
	fprintf(stderr,"Copying current gyro-optical calibration file to target data folder... ");
	string targetCalibrationPath = targetFolderPath + filesep + string("GyroOpticalCalibration.json");
	saveCalibrationData(targetCalibrationPath, b, Ktilde, sigma, Lgyro);
	fprintf(stderr,"Done.\n\r");
	
	//Set gyro bias and latency corrections
	fprintf(stderr,"Setting gyro bias and latency corrections... ");
	gyro::setGyroBiasCorrections(b(0), b(1), b(2));
	gyro::setGyroLatencyCorrection(Lgyro);
	fprintf(stderr,"Done.\n\r");
	
	//Wait until user is ready to start recording
	key = waitForButtonPress(cap, "Live Display", 5, string("Press any key to start recording"), string(""));
	
	//Render display with feature locations estimated based on gyro readings only
	fprintf(stderr,"Recording video and gyro data. Press spacebar to exit.\n\r");
	struct timeval oldTime, newTime;
	Mat oldFrame, newFrame;
	getFrameWithTimestamp(cap, newFrame, newTime);
	dataSaver.saveFrame(0, newFrame);
	Matrix3d H, R;
	double rotationAngle;
	
	for (int frameNumber=1; ; frameNumber++) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		
		//Save frame
     	dataSaver.saveFrame(frameNumber, newFrame);
     	
     	//Get rotation between frames
     	retval = gyro::getRotationBetweenTwoTimes(oldTime, newTime, R, rotationAngle, true);
		if (retval != 0) {
			fprintf(stderr,"gyro::getRotationBetweenTwoTimes failed with return code %d", retval);
			gyro::closePort();
			return(-1);
		}
		
		//Save rotation
		dataSaver.saveRotation(frameNumber-1, frameNumber, R);
		
		//Update display
     	imshow("Live Display", newFrame);
		
		key = waitKey(1);
		if (key == ' ') break;
	}
	
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	printf("\n\r***************************  Gyro-Optical Recording Program  ***************************\n\r");
	fprintf(stderr,"This program records video and synchronized gyro data. Between-frame rotations are\n\r");
	fprintf(stderr,"saved instead of raw measurements, so it is important that the gyro and optical systems\n\r");
	fprintf(stderr,"both be calibrated prior to recording. Certain parameters can be specified through\n\r");
	fprintf(stderr,"command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file.\n\r");
	fprintf(stderr,"           Defalt value is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking. Default is %d.\n\r\n\r", DEFAULT_VIDEOCAM_NUM);
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	frame = frameColor;
	//cvtColor(frameColor, frame, CV_RGB2GRAY);
}
