#include "singleFeatureEnergy.h"
#include "parameters.h"
#include <cstdint>
#include <cmath>
#include "Utilities.h"
#include <thread>

using namespace cv;
using namespace std;
using namespace Utilities;

double computeBlendedEnergy(ImageStack& Template, ImageStack& frameStack, Vec2f location, const double blendingCoefficients[]) {
	const double denominator = parameters::n*parameters::n*parameters::maxIntensity;
	Mat I;
	double energy = 0.0;
	Vec2f location_thisLevel;
	for (int m=0; m<parameters::M; m++) {
		if (blendingCoefficients[m] > 0.000000001) {
			location_thisLevel = location/pow(2,m);
			//frameStack.convertImageCoordinates(location, location_thisLevel, 0, m);
			if ((location_thisLevel[0] >= 0) &&
			    (location_thisLevel[0] < frameStack.Images[m].rows) &&
			    (location_thisLevel[1] >= 0) &&
			    (location_thisLevel[1] < frameStack.Images[m].cols))
				getRectSubPix(frameStack.Images[m], Size(parameters::n,parameters::n), Point2f(location_thisLevel[1], location_thisLevel[0]), I);
			else
				I = Mat::zeros(parameters::n, parameters::n, frameStack.Images[m].type());
			energy += blendingCoefficients[m]*Utilities::computeL1Diff_SIMD(Template.Images[m], I);
		}
	}
	return(energy/denominator);
}

void computeBlendedEnergy(ImageStack& Template, ImageStack& frameStack, Vec2f location, const double blendingCoefficients[], double &energy, double &Dx, double &Dy) {
	//Conpute blended energy and derivative in a slow yet simple way.
	
	Vec2f location_xminus = location - Vec2f(0.25, 0.0);
	Vec2f location_xplus  = location + Vec2f(0.25, 0.0);
	Vec2f location_yminus = location - Vec2f(0.0, 0.25);
	Vec2f location_yplus  = location + Vec2f(0.0, 0.25);
	
	energy = computeBlendedEnergy(Template, frameStack, location, blendingCoefficients);
	Dx = 2.0*(computeBlendedEnergy(Template, frameStack, location_xplus, blendingCoefficients) - computeBlendedEnergy(Template, frameStack, location_xminus, blendingCoefficients));
	Dy = 2.0*(computeBlendedEnergy(Template, frameStack, location_yplus, blendingCoefficients) - computeBlendedEnergy(Template, frameStack, location_yminus, blendingCoefficients));
}

void computeBlendedEnergy(ImageStack& Template, ImageStack& frameStack, ImageStack& DIDX, ImageStack& DIDY, Vec2f location, const double blendingCoefficients[], double &energy, double &Dx, double &Dy) {
	//Conpute blended energy and derivative using the chain rule and pre-computed image gradients.
	//This function is not yet implemented
	
	fprintf(stderr,"\nAccelerated computeBlendedEnergy function is not yet implemented!\n");
}

double computeSingleFeatureEnergy(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location) {
	//Compute the single-feature energy. The input location must be provided at the base (full resolution) pyramid level.
	const double denominator = parameters::n*parameters::n*parameters::maxIntensity;
	Mat I;
	double energy;
	Vec2f location_thisLevel;
	
	location_thisLevel = location/pow(2,resolutionLevel);
	if ((location_thisLevel[0] >= 0) &&
	    (location_thisLevel[0] < frameStack.Images[resolutionLevel].rows) &&
	    (location_thisLevel[1] >= 0) &&
	    (location_thisLevel[1] < frameStack.Images[resolutionLevel].cols))
		getRectSubPix(frameStack.Images[resolutionLevel], Size(parameters::n,parameters::n), Point2f(location_thisLevel[1], location_thisLevel[0]), I);
	else
		I = Mat::zeros(parameters::n, parameters::n, frameStack.Images[resolutionLevel].type());
	energy = Utilities::computeL1Diff_SIMD(Template.Images[resolutionLevel], I)/denominator;
	return(energy);
}

void computeSingleFeatureEnergy(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location, double& energy, double& Dx, double& Dy) {
	//The input locations must be at the base resolution level (pyramid level 0).
	//It is important when numerically computing the derivative by sampling symetrically about a point like this that the
	//sampling distance be reasonable *relative* to the source imagery. The source imagery is selected in the energy evaluation
	//function according to "resolutionLevel" and the sample location is scaled appropriately. We must account for that here by
	//enlarging the sampling distance on higher pyramid levels.
	float sampleDelta = (float)(0.25*pow(2.0, resolutionLevel));
	Vec2f location_xminus = location - Vec2f(sampleDelta, 0.0);
	Vec2f location_xplus  = location + Vec2f(sampleDelta, 0.0);
	Vec2f location_yminus = location - Vec2f(0.0, sampleDelta);
	Vec2f location_yplus  = location + Vec2f(0.0, sampleDelta);
	
	energy =  computeSingleFeatureEnergy(Template, frameStack, resolutionLevel, location);
	Dx = (computeSingleFeatureEnergy(Template, frameStack, resolutionLevel, location_xplus) - computeSingleFeatureEnergy(Template, frameStack, resolutionLevel, location_xminus))/(2.0*sampleDelta);
	Dy = (computeSingleFeatureEnergy(Template, frameStack, resolutionLevel, location_yplus) - computeSingleFeatureEnergy(Template, frameStack, resolutionLevel, location_yminus))/(2.0*sampleDelta);
}

double computeSingleFeatureEnergy_regularized(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location, Vec2f locationPrior, double lambda) {
	//Compute the single-feature energy with regularization term. The input location must be provided at the base (full resolution) pyramid level.
	const double denominator = parameters::n*parameters::n*parameters::maxIntensity;
	Mat I;
	double regularizationTerm, energy;
	Vec2f location_thisLevel;
	
	location_thisLevel = location/pow(2,resolutionLevel);
	if ((location_thisLevel[0] >= 0) &&
	    (location_thisLevel[0] < frameStack.Images[resolutionLevel].rows) &&
	    (location_thisLevel[1] >= 0) &&
	    (location_thisLevel[1] < frameStack.Images[resolutionLevel].cols))
		getRectSubPix(frameStack.Images[resolutionLevel], Size(parameters::n,parameters::n), Point2f(location_thisLevel[1], location_thisLevel[0]), I);
	else
		I = Mat::zeros(parameters::n, parameters::n, frameStack.Images[resolutionLevel].type());
	regularizationTerm = lambda*log(parameters::alpha*norm(location-locationPrior) + 1.0)/log(parameters::alpha*parameters::xmax + 1.0);
	energy = Utilities::computeL1Diff_SIMD(Template.Images[resolutionLevel], I)/denominator + regularizationTerm;
	return(energy);
}

void computeSingleFeatureEnergy_regularized(ImageStack& Template, ImageStack& frameStack, int resolutionLevel, Vec2f location, Vec2f locationPrior, double lambda, double& energy, double& Dx, double& Dy) {
	//The input locations must be at the base resolution level (pyramid level 0).
	//It is important when numerically computing the derivative by sampling symetrically about a point like this that the
	//sampling distance be reasonable *relative* to the source imagery. The source imagery is selected in the energy evaluation
	//function according to "resolutionLevel" and the sample location is scaled appropriately. We must account for that here by
	//enlarging the sampling distance on higher pyramid levels.
	float sampleDelta = (float)(0.25*pow(2.0, resolutionLevel));
	Vec2f location_xminus = location - Vec2f(sampleDelta, 0.0);
	Vec2f location_xplus  = location + Vec2f(sampleDelta, 0.0);
	Vec2f location_yminus = location - Vec2f(0.0, sampleDelta);
	Vec2f location_yplus  = location + Vec2f(0.0, sampleDelta);
	
	energy =  computeSingleFeatureEnergy_regularized(Template, frameStack, resolutionLevel, location, locationPrior, lambda);
	Dx = (computeSingleFeatureEnergy_regularized(Template, frameStack, resolutionLevel, location_xplus, locationPrior, lambda) - computeSingleFeatureEnergy_regularized(Template, frameStack, resolutionLevel, location_xminus, locationPrior, lambda))/(2.0*sampleDelta);
	Dy = (computeSingleFeatureEnergy_regularized(Template, frameStack, resolutionLevel, location_yplus, locationPrior, lambda) - computeSingleFeatureEnergy_regularized(Template, frameStack, resolutionLevel, location_yminus, locationPrior, lambda))/(2.0*sampleDelta);
}


