//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/time.h>

//Internal Includes
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "gyroManager.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"
#include "singleFeatureEnergy.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

//Set default option values
#define DEFAULT_TRACKERID 0
#define DEFAULT_GYROPATH "/dev/ttyACM0"
#define DEFAULT_VIDEOCAM_NUM 1

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//Run this program with the -h command-line option for usage options.

//Forward Declarations
void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime);
void displayHelp(void);
void renderFeature(ImageStack& frameStack);
Mat computeBlendedEnergySurfaceForActiveFeature(int X, int Y, ImageStack& frameStack, const double blendingCoefficients[]);

//Constants
const double pi = 3.141592653589793;

//Pointer to "Active Feature" for rendering live energy function
int activeFeatureID = -1;
Feature *activeFeaturePtr = NULL;


static void onMouse(int event, int x, int y, int, void* userData) {
	if (event != EVENT_LBUTTONDOWN) return;
	if (userData == NULL) return;
	
	//Identify feature nearest the mouse click
	Vec2f click((float)y, (float)x);
	Tracker *myTrackerPtr = (Tracker*)userData;
	if (myTrackerPtr->features.Features.size() > 0) {
		activeFeaturePtr = &(myTrackerPtr->features.Features.front());
		activeFeatureID = activeFeaturePtr->featureID;
		double lowestDist = norm(activeFeaturePtr->CurrentLocation - click);
		for (list<Feature>::iterator it = myTrackerPtr->features.Features.begin(); it != myTrackerPtr->features.Features.end(); it++) {
			double dist = norm(it->CurrentLocation - click);
			if (dist < lowestDist) {
				activeFeaturePtr = &(*it);
				activeFeatureID = activeFeaturePtr->featureID;;
				lowestDist = dist;
			}
		}
		//fprintf(stderr,"Click: (%f, %f)\n\r", click[0], click[1]);
		//fprintf(stderr,"Nearest Feasture: (%f, %f)\n\r", activeFeaturePtr->CurrentLocation[0], activeFeaturePtr->CurrentLocation[1]);
	}
}

int main(int argc, char** argv) {
     //Setup for command-line option parsing
	static struct option long_options[] = {
		{"tracker",       required_argument, 0, 't'},
		{"gyropath",      required_argument, 0, 'g'},
		{"cameraID",      required_argument, 0, 'v'},
		{"help",          no_argument,       0, 'h'},
		{"p1",            required_argument, 0, 256},
		{"p2",            required_argument, 0, 257},
		{0, 0, 0, 0}
	};
	
	//Set default parameters
	string GyroDevicePath(DEFAULT_GYROPATH);
     int cameraOpenCVIndex          = DEFAULT_VIDEOCAM_NUM;
     int trackerID                  = DEFAULT_TRACKERID;
	bool trackerParameter1Provided = false;
	bool trackerParameter2Provided = false;
	double trackerParameter1       = 0.0;
	double trackerParameter2       = 0.0;
	
	//Parse command-line options and override defaults with provided values.
	int c, option_index;
	while ((c = getopt_long (argc, argv, "t:g:v:h", long_options, &option_index)) != -1)
		switch (c) {
			case 't': trackerID = atoi(optarg);          break;
			case 'g': GyroDevicePath = string(optarg);   break;
			case 'v': cameraOpenCVIndex = atoi(optarg);  break;
			case 'h': displayHelp();                     return(0);
			case 256:
				trackerParameter1Provided = true;
				trackerParameter1 = atof(optarg);
				break;
			case 257:
				trackerParameter2Provided = true;
				trackerParameter2 = atof(optarg);
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}
     
     int key, retval;
     Mat temp, frame, frameBW, frameWithFeatures;
     Matrix3d Ktilde, KtildeInv;
	
	//Decide whether or not the gyro system will be needed
	bool useGyroSystem = false;
	if ((trackerID == 1) || (trackerID == 3) || (trackerID == 4) || (trackerID == 12) || (trackerID == 13)) {
		fprintf(stderr,"Selected tracker requires gyro module. Real-time gyro processing system will be enabled.\n\r");
		useGyroSystem = true;
	}
     
     //TODO:Disable for non-debug mode
     feenableexcept(FE_INVALID | FE_OVERFLOW);
     
     //Initialize Tracker Output Window
     namedWindow("Live Display", CV_WINDOW_AUTOSIZE|CV_GUI_NORMAL);
     
     //Open Video Camera
	VideoCapture cap(cameraOpenCVIndex); //Open the camera
	if(!cap.isOpened())                  //Check if we succeeded
		return -1;
	
	if (useGyroSystem) {
		//Start Gyro Subsystem
		int returnCode;
		returnCode = gyro::openPort(GyroDevicePath.c_str());
		if (returnCode != 0) {
			fprintf(stderr,"Failed to open serial port for gyro module. Aborting...\n\r");
			gyro::closePort();
			return -1;
		}
	
		//Set Gyro divider
		usleep(100000U); //Wait 1/10 of a second.
		returnCode = gyro::setDivider(1U);
		if (returnCode != 0) {
			fprintf(stderr,"Failed to set gyro divider. Aborting...\n\r");
			gyro::closePort();
			return -1;
		}
	
		//Check that sensor is giving us data
		usleep(100000U); //Wait 1/10 of a second.
		//sleep(1U); //Wait 1 second.
		if (!gyro::receivingGyroData()) {
			fprintf(stderr,"Not receiving data from gyro module. Aborting...\n\r");
			gyro::closePort();
			return -1;
		}
		
		//Load Gyro calibration data
		string CalibrationFileName("GyroOpticalCalibration.json");
		fprintf(stderr,"Loading Calibration Data from file: %s... ", CalibrationFileName.c_str());
		Vector3d b;
		double Lgyro;
		double sigma;
		retval = loadCalibrationData(CalibrationFileName, b, Ktilde, sigma, Lgyro);
		KtildeInv = Ktilde.inverse();
		if (retval == 0) fprintf(stderr,"Success\n\r");
		else {
			fprintf(stderr,"Failure -> Aborting...\n\r");
			gyro::closePort();
			return(-1);
		}
	
		//Set gyro bias and latency corrections
		fprintf(stderr,"Setting gyro bias and latency corrections... ");
		gyro::setGyroBiasCorrections(b(0), b(1), b(2));
		gyro::setGyroLatencyCorrection(Lgyro);
		fprintf(stderr,"Done.\n\r");
	}
	
	//Disable autofocus or auto-exposure for camera (if desired) - Linux and V4L only
	Utilities::adjustCameraSettings(cap, "Live Display", cameraOpenCVIndex);
	
	fprintf(stderr,"Rendering live display.\n\rPress spacebar to exit.\n\r");
	struct timeval oldTime, newTime;
	Mat oldFrame, newFrame;
	getFrameWithTimestamp(cap, newFrame, newTime);
	Matrix3d H, R;
	double rotationAngle;
	ImageStack frameStack;
     Tracker myTracker;
     
     //Build Multi-Feature tracker options struct and set up appropriately
     MultiTrackerParameters myMultiFeatureTrackerParams;
     myMultiFeatureTrackerParams.L = 5;
     switch(trackerID) {
     	case 5:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.1;     //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 6:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;    //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 7:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.001;   //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 8:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0005;  //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 9:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0015;  //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 10: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.002;   //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 13: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;    //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     }
     
     //If trackerParameter1 is provided at run-time, use this for the constrained tracker penalty coefficient
     if (trackerParameter1Provided && ((trackerID == 5) || (trackerID == 6) || (trackerID == 7) || (trackerID == 8) || (trackerID == 9) || (trackerID == 10) || (trackerID == 13))) {
     	myMultiFeatureTrackerParams.penaltyCoefficient = trackerParameter1;
     }
	
	//Initialize Tracker
	frameStack.BuildFromImage(newFrame, parameters::M);
	myTracker.initialize(0, frameStack, 0);
	myTracker.multiTrackerParametersPtr = &myMultiFeatureTrackerParams;
	
     fprintf(stderr,"\nTracker initialized - Tracking input stream.\n");
     
     //Register Mouse Callback
     setMouseCallback("Live Display", onMouse, (void *)(&myTracker));
	
	for (int frameNumber=1; ; frameNumber++) {
		oldTime = newTime;
		oldFrame = newFrame;
		newFrame = Mat();
		getFrameWithTimestamp(cap, newFrame, newTime);
		tic(); //Start timer for detirmining processing frame rate.
		
		//Build frame stack for new frame
		frameStack.BuildFromImage(newFrame, parameters::M);
     	
     	//Get rotation between frames
     	if (useGyroSystem) {
			retval = gyro::getRotationBetweenTwoTimes(oldTime, newTime, R, rotationAngle, true);
			if (retval != 0) {
				fprintf(stderr,"gyro::getRotationBetweenTwoTimes failed with return code %d", retval);
				gyro::closePort();
				return(-1);
			}
		
			//Estimate Homography between 2 frames
			H = Ktilde*R*KtildeInv;
		}
     	
     	//Track features to new frame
     	switch (trackerID) {
     		case 0:  myTracker.trackToNextFrame_KLT                             (frameNumber, frameStack);    break;
     		case 1:  myTracker.trackToNextFrame_KLT_HomInit                     (frameNumber, frameStack, H); break;
     		case 2:  myTracker.trackToNextFrame_1stOrderDescent_GlobalMotionInit(frameNumber, frameStack);    break;
     		case 3:  myTracker.trackToNextFrame_1stOrderDescent_HomInit         (frameNumber, frameStack, H); break;
     		case 4:  myTracker.trackToNextFrame_GyroOptical_Experimental_1      (frameNumber, frameStack, H); break;
     		case 5:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 6:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 7:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 8:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 9:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 10: myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
			case 11: myTracker.trackToNextFrame_1stOrderDescent_Blended         (frameNumber, frameStack);    break;
			case 12:	if (trackerParameter1Provided) {
						myTracker.trackToNextFrame_1stOrderDescent_GyroPrior (frameNumber, frameStack, H, trackerParameter1);
						break;
					}
					else {
						myTracker.trackToNextFrame_1stOrderDescent_GyroPrior (frameNumber, frameStack, H);
						break;
					}
     		case 13:	if (trackerParameter2Provided) {
						myTracker.trackToNextFrame_multiTrack_GyroPrior      (frameNumber, frameStack, H, trackerParameter2);
						break;
					}
					else {
						myTracker.trackToNextFrame_multiTrack_GyroPrior      (frameNumber, frameStack, H);
						break;
					}
     		default: fprintf(stderr,"Un-implemented tracker ID.");
     		         if (useGyroSystem)
     		              gyro::closePort();
     		         return(-1);
     	}
     	
     	fprintf(stderr,"\rDone tracking through frame %d (Processing framerate: %f)", frameNumber, 1.0/toc());
     	
     	//Render Current Features and check for stop condition (space bar)
     	cvtColor(newFrame, frameWithFeatures, CV_GRAY2RGB);
     	myTracker.features.renderToImage(frameWithFeatures);
     	imshow("Live Display", frameWithFeatures);
     	
     	if (activeFeaturePtr != NULL) {
     		//Check to see if active feature is still alive
     		activeFeaturePtr = NULL;
     		for (list<Feature>::iterator it = myTracker.features.Features.begin(); it != myTracker.features.Features.end(); it++) {
				if (it->featureID == activeFeatureID) {
					activeFeaturePtr = &(*it);
					break;
				}
			}
			if (activeFeaturePtr != NULL) {
				renderFeature(frameStack);
			}
     	}
		
		key = waitKey(1);
		if (key == ' ') break;
	}
	
	fprintf(stderr,"\nExiting.\n");
	gyro::closePort();
     return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"************************************     Real-Time Tracker     **********************************\n\r");
	fprintf(stderr,"This program runs one of multiple feature trackers in real-time on streaming video. This program\n\r");
	fprintf(stderr,"supports gyro-integrated trackers, provided a gyro module is available. Certain parameters are\n\r");
	fprintf(stderr,"specified through command-line options. The supported options are:\n\r\n\r");
	fprintf(stderr,"-t n       Selects the tracker to be used. Default is %d. Supported values for n are:\n\r", DEFAULT_TRACKERID);
	fprintf(stderr,"           0:  KLT\n\r");
	fprintf(stderr,"           1:  KLT with Gyro initialization\n\r");
	fprintf(stderr,"           2:  1st-order decent\n\r");
	fprintf(stderr,"           3:  1st-order decent with Gyro initialization\n\r");
	fprintf(stderr,"           4:  Experimental Gyro-integrated tracker #1\n\r");
	fprintf(stderr,"           5:  Multi-tracker with rank penalty - Emp Dim  - uncentered M\n\r");
	fprintf(stderr,"           6:  Multi-tracker with rank penalty - Emp Dim  - centered M\n\r");
	fprintf(stderr,"           7:  Multi-tracker with rank penalty - Nuc Norm - uncentered M\n\r");
	fprintf(stderr,"           8:  Multi-tracker with rank penalty - Nuc Norm - centered M\n\r");
	fprintf(stderr,"           9:  Multi-tracker with rank penalty - Exp Fact - uncentered M\n\r");
	fprintf(stderr,"           10: Multi-tracker with rank penalty - Exp Fact - centered M\n\r");
	fprintf(stderr,"           11: Blended-energy tracker - 1st order descent\n\r");
	fprintf(stderr,"           12: 1st-order decent with gyro prior (lambda = --p1)\n\r");
	fprintf(stderr,"           13: Multi-tracker with rank penalty - Emp Dim  - centered M - and gyro prior.\n\r");
	fprintf(stderr,"               (penaltyCoef = --p1, lambda = --p2)\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for gyro device file. Only used if a gyro-integrated tracker is run.\n\r");
	fprintf(stderr,"           Defalt value is %s\n\r\n\r", DEFAULT_GYROPATH);
	fprintf(stderr,"-v n       Selects the video camera (OpenCV index) to use for tracking. Default is %d.\n\r\n\r", DEFAULT_VIDEOCAM_NUM);
	fprintf(stderr,"Some trackers allow tuning parameters to be set at runtime through command-line options. These are\n\r");
	fprintf(stderr,"never required - if omitted the program will use default values. Refer to the documentation on ths -t\n\r");
	fprintf(stderr,"option to see which trackers support run-time parameters and their meanings.\n\r");
	fprintf(stderr,"--p1=x     Set tracker parameter p1 to x. The meaning of p1 depends on the selected tracker (-t)\n\r");
	fprintf(stderr,"--p2=x     Set tracker parameter p2 to x. The meaning of p2 depends on the selected tracker (-t)\n\r");
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
}

void getFrameWithTimestamp(VideoCapture &cap, Mat &frame, struct timeval &RXTime) {
	Mat frameColor;
	cap >> frameColor;
	gettimeofday(&RXTime, NULL);
	cvtColor(frameColor, frame, CV_RGB2GRAY);
}

void renderFeature(ImageStack& frameStack) {
	//Render Template(s)
	Mat largeTemplate, EnergySurface;
	for (int m=0; m<parameters::M; m++) {
		char str[32];
		sprintf(str, "Template m=%d", m);
		resize(activeFeaturePtr->Template.Images[m], largeTemplate, Size(8*parameters::n,8*parameters::n), 0, 0, INTER_NEAREST);
		imshow(str, largeTemplate);
	}
	
	const double blendingCoefficients1[] = {0.0, 0.0, 0.0, 1.0};
	const double blendingCoefficients2[] = {1.0, 0.0, 0.0, 0.0};
	const double blendingCoefficients3[] = {1.0, 0.0, 0.0, 2.0};
	
	int X = 45;
	int Y = 45;
	EnergySurface = computeBlendedEnergySurfaceForActiveFeature(X, Y, frameStack, blendingCoefficients1);
	imshow("Energy Surface 1", EnergySurface);
	
	EnergySurface = computeBlendedEnergySurfaceForActiveFeature(X, Y, frameStack, blendingCoefficients2);
	imshow("Energy Surface 2", EnergySurface);
	
	EnergySurface = computeBlendedEnergySurfaceForActiveFeature(X, Y, frameStack, blendingCoefficients3);
	imshow("Energy Surface 3", EnergySurface);
}

Mat computeBlendedEnergySurfaceForActiveFeature(int X, int Y, ImageStack& frameStack, const double blendingCoefficients[]) {
	Vec2f location;
	Mat energySurface(X,Y,CV_32FC1);
	for (int x=0; x<X; x++) {
		for (int y=0; y<Y; y++) {
			location = activeFeaturePtr->CurrentLocation - Vec2f(((float)X - 1)/2, ((float)Y - 1)/2) + Vec2f(x, y);
			energySurface.at<float>(x,y) = (float)computeBlendedEnergy(activeFeaturePtr->Template, frameStack, location, blendingCoefficients);
		}
	}
	Mat energySurface_UINT8;
	double minVal;
	double maxVal;
	minMaxLoc(energySurface, &minVal, &maxVal, NULL, NULL);
	energySurface.convertTo(energySurface_UINT8, CV_8UC1, 255.0/maxVal, 0);
	return(energySurface_UINT8);
}
