//External Includes
#include "opencv2/opencv.hpp"
#include <stdio.h>
#include <getopt.h>
#include <sys/time.h>
#include "boost/filesystem.hpp"

//Internal Includes
#include "Tracker.h"
#include "ImageStack.h"
#include "parameters.h"
#include "Benchmark.h"
#include "Utilities.h"
#include "calibrationLoadSave.h"
#include "rawData.h"

//#define _GNU_SOURCE 1
#include <fenv.h>

using namespace cv;
using namespace std;
using namespace Eigen;
using namespace Utilities;

//This program performs tracking on a pre-recorded data set. It can use one of multiple trackers and it can run
//in different modes. Run with -h command-line switch for a list of run-time options.

//Set default option values
#define DEFAULT_TRACKERID 0
#define DEFAULT_MODE 0
#define DEFAULT_GROUNDTRUTH_PATH ""
#define DEFAULT_INITIALFEATURES_PATH ""
#define DEFAULT_REINIT_TOLERANCE 10.0

//Forward Declarations
void displayHelp(void);
void tracker_Init(Tracker &myTracker, int mode, double reinitTolerance, rawDataLoader &dataLoader, string initFeaturesPath, string groundTruthPath);
void tracker_TrackToNextFrame(Tracker &myTracker, int frameNumber, rawDataLoader &dataLoader, int trackerID, Matrix3d &Ktilde, Matrix3d &KtildeInv,
						bool renderLive, bool trackerParameter1Provided, bool trackerParameter2Provided, double trackerParameter1, double trackerParameter2);
Mat toGray(Mat image);

//Constants
const double pi = 3.141592653589793;

int main(int argc, char** argv) {
	string sourceFolderPath;
	string targetFilePath;
	
	//Setup for command-line option parsing
	static struct option long_options[] = {
		{"tracker",             required_argument, 0, 't'},
		{"mode",                required_argument, 0, 'm'},
		{"groundtruthpath",     required_argument, 0, 'g'},
		{"initialfeaturespath", required_argument, 0, 'i'},
		{"tolerance",           required_argument, 0, 'e'},
		{"help",                no_argument,       0, 'h'},
		{"render",              no_argument,       0, 'r'},
		{"p1",                  required_argument, 0, 256},
		{"p2",                  required_argument, 0, 257},
		{0, 0, 0, 0}
	};
	
	//Set default parameters
	int stopFrame = -1; //If >= 0, stop tracking when this frame is reached
	int trackerID = DEFAULT_TRACKERID;
	int mode = DEFAULT_MODE;
	string groundTruthFilePath(DEFAULT_GROUNDTRUTH_PATH);
	string initFeaturesFilePath(DEFAULT_INITIALFEATURES_PATH);
	double reinitTolerance = DEFAULT_REINIT_TOLERANCE;
	bool renderLive = false;
	bool trackerParameter1Provided = false;
	bool trackerParameter2Provided = false;
	double trackerParameter1 = 0.0;
	double trackerParameter2 = 0.0;
	
	//Parse command-line options and override defaults with provided values.
	int c, option_index;
	while ((c = getopt_long (argc, argv, "t:s:m:g:i:e:hr", long_options, &option_index)) != -1)
		switch (c) {
			case 't': trackerID = atoi(optarg);              break;
			case 'm': mode = atoi(optarg);                   break;
			case 'g': groundTruthFilePath = string(optarg);  break;
			case 'i': initFeaturesFilePath = string(optarg); break;
			case 'e': reinitTolerance = atof(optarg);        break;
			case 'h': displayHelp();                         return(0);
			case 'r': renderLive = true;                     break;
			case 's': stopFrame = atoi(optarg);              break;
			case 256:
				trackerParameter1Provided = true;
				trackerParameter1 = atof(optarg);
				break;
			case 257:
				trackerParameter2Provided = true;
				trackerParameter2 = atof(optarg);
				break;
			case '?':
				fprintf (stderr, "Unrecognized Option: -%c found. Aborting.\n\r", optopt);
				return(-1);
			default:
				fprintf(stderr, "Unexpected value returned by getopt. Aborting.\n\r");
				return(-1);
	}
	
	//Check for and process required arguments
	if (argc - optind != 2) {
		fprintf(stderr,"%s has 2 mandatory arguments. Run %s -h for complete usage.\n\r", argv[0], argv[0]);
		return(-1);
	}
	sourceFolderPath = string(argv[optind]);
	targetFilePath = string(argv[optind+1]);
	
     //Done parsing parameters - Try to initialize a raw data loader for the provided data folder and make sure it has valid data
     rawDataLoader dataLoader;
     if (dataLoader.init(sourceFolderPath) != 0) {
     	fprintf(stderr,"Error initializing raw data loader for provided data folder. Aborting.\n\r");
     	return(-1);
     }
     if ((trackerID == 1) || (trackerID == 3) || (trackerID == 4) || (trackerID == 12) || (trackerID == 13)) {
     	if (!dataLoader.containsRotations()) {
     		fprintf(stderr,"The selected tracker requires gyro data, which does not appear to be available for the provided data set. Aborting.\n\r");
     		return(-1);
     	}
     }
     
     //Load gyro-optical calibration data
     //Make sure sourceFolderPath has a trailing filesep.
	if ((sourceFolderPath.back() != '\\') && (sourceFolderPath.back() != '/')) {
		boost::filesystem::path slash("/");
		std::string filesep = slash.make_preferred().native();
		sourceFolderPath = sourceFolderPath + filesep;
	}
     string CalibrationFileName(sourceFolderPath + "GyroOpticalCalibration.json");
	Vector3d b;   //Not neaded
	double Lgyro; //Not Neaded
	Matrix3d Ktilde, KtildeInv;
	double sigma; //Not Neaded
	if ((trackerID == 1) || (trackerID == 3) || (trackerID == 4) || (trackerID == 12) || (trackerID == 13)) {
		int retval = loadCalibrationData(CalibrationFileName, b, Ktilde, sigma, Lgyro);
		if (retval != 0) {
			fprintf(stderr,"Failure loading calibration data. -> Aborting...\n\r");
			return(-1);
		}
		KtildeInv = Ktilde.inverse();
	}

	//If we make it this far, we should be ready to track. numFrames will hold the number of frames to track through.
	int numFrames = dataLoader.getNumberOfFramesAvailable();
	if ((stopFrame >= 0) && (numFrames >= stopFrame+1)) {
		numFrames = stopFrame+1;
	}
	Tracker myTracker;
	
	//Build Multi-Feature tracker options struct and set up appropriately
     MultiTrackerParameters myMultiFeatureTrackerParams;
     myMultiFeatureTrackerParams.L = 10;
     switch(trackerID) {
     	case 5:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.1;     //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 6:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;    //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 7:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.001;   //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 8:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::NuclearNorm;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0005;  //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 9:  myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.0015;  //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = false;
     	         break;
     	case 10: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::ExplicitFactorization;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.002;   //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     	case 13: myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::EmpiricalDimension;
     	         myMultiFeatureTrackerParams.penaltyCoefficient = 0.15;    //Reasonable default value for lightly degraded videos
     	         myMultiFeatureTrackerParams.centerHistoryMatrix = true;
     	         break;
     }
     
     //If trackerParameter1 is provided at run-time, use this for the constrained tracker penalty coefficient
     if (trackerParameter1Provided && ((trackerID == 5) || (trackerID == 6) || (trackerID == 7) || (trackerID == 8) || (trackerID == 9) || (trackerID == 10) || (trackerID == 13))) {
     	myMultiFeatureTrackerParams.penaltyCoefficient = trackerParameter1;
     }
     
     //If the rank penalty term is disabled by setting its coeficient to 0, then disable rank penalty evaluation when computing energy and gradients
     if (myMultiFeatureTrackerParams.penaltyCoefficient == 0.0) {
     	myMultiFeatureTrackerParams.penaltyForm = MultiTrackerParameters::penaltyType::None;
     }
	
	//Initialize Tracker
	fprintf(stderr,"Initializing tracker... ");
	tracker_Init(myTracker, mode, reinitTolerance, dataLoader, initFeaturesFilePath, groundTruthFilePath);
	myTracker.multiTrackerParametersPtr = &myMultiFeatureTrackerParams;
	fprintf(stderr,"Done\n\r");
	
	//Track through all available frames, or until stop frame
	double totalTime = 0.0;
	fprintf(stderr,"Beginning to track\n\r");
	for (int frameNumber = 1; frameNumber < numFrames; frameNumber++) {
		tic();
		tracker_TrackToNextFrame(myTracker, frameNumber, dataLoader, trackerID, Ktilde, KtildeInv,
							renderLive, trackerParameter1Provided, trackerParameter2Provided, trackerParameter1, trackerParameter2);
		totalTime += toc();
		fprintf(stderr,"\rTracking to frame %d", frameNumber);
	}
	fprintf(stderr,"\n\rTracking Complete.\n\r");
	
	//We are done tracking. Move all current features to dead-tracks feature set
	list< Feature >::iterator featureIter = myTracker.features.Features.begin();
	while (featureIter != myTracker.features.Features.end()) {
		myTracker.deadFeatures.addFeature(*featureIter);
		myTracker.features.removeFeature(featureIter);
	}
	
	if (mode == 2) {
		//Render formated tracker results to stdout. The head and tail tokens are used by other programs to extract the formatted results section. These tokens
		//should not be changed.
		int featureFrameCounter = 0;
		for (list<Feature>::iterator featureIter = myTracker.deadFeatures.Features.begin(); featureIter != myTracker.deadFeatures.Features.end(); featureIter++)
			featureFrameCounter += (int)(featureIter->History.size());
		
		const char * headToken = "******************   Begin Results   ******************\n";
		const char * tailToken = "*******************   End Results   *******************\n";
		fprintf(stdout, "%s", headToken);
		fprintf(stdout, "Total Re-initializations:             %d\n", myTracker.reinitializationCounter);
		fprintf(stdout, "Total feature frame count:            %d\n", featureFrameCounter);
		fprintf(stdout, "Average frames per re-initialization: %f\n", ((double)featureFrameCounter)/((double)(myTracker.reinitializationCounter)));
		fprintf(stdout, "Re-initialization tolerance (pixels): %f\n", reinitTolerance);
		fprintf(stdout, "Mean processing frame-rate (fps):     %f\n", (numFrames-1)/totalTime);
		fprintf(stdout, "%s", tailToken);
	}
	
	//Save tracker output
	if (saveTrajectories(myTracker.deadFeatures, targetFilePath) != 0)
		fprintf(stderr,"Error saving final feature trajectories.\n\r");
	
	return(0);
}

void displayHelp(void) {
	fprintf(stderr,"\n\r");
	fprintf(stderr,"**********************************     Post-Processing Tracker     *********************************\n\r");
	fprintf(stderr,"This program runs one of multiple feature trackers on a pre-recorded video. This program supports\n\r");
	fprintf(stderr,"gyro-integrated trackers, provided the data set you try to run on contains inter-frame rotation\n\r");
	fprintf(stderr,"data. This program has 2 mandatory arguments. The first is the absolute path of the source video\n\r");
	fprintf(stderr,"folder on which feature tracking will be run. The second is the absolute path of the output file\n\r");
	fprintf(stderr,"where tracker results will be saved. Other parameters are specified through command-line options.\n\r");
	fprintf(stderr,"The supported options are:\n\r\n\r");
	fprintf(stderr,"-t n       Selects the tracker to be used. Supported values for n are:\n\r");
	fprintf(stderr,"           0:  KLT\n\r");
	fprintf(stderr,"           1:  KLT with Gyro initialization\n\r");
	fprintf(stderr,"           2:  1st-order decent\n\r");
	fprintf(stderr,"           3:  1st-order decent with Gyro initialization\n\r");
	fprintf(stderr,"           4:  Experimental Gyro-integrated tracker #1\n\r");
	fprintf(stderr,"           5:  Multi-tracker with rank penalty - Emp Dim  - uncentered M (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           6:  Multi-tracker with rank penalty - Emp Dim  - centered M   (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           7:  Multi-tracker with rank penalty - Nuc Norm - uncentered M (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           8:  Multi-tracker with rank penalty - Nuc Norm - centered M   (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           9:  Multi-tracker with rank penalty - Exp Fact - uncentered M (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           10: Multi-tracker with rank penalty - Exp Fact - centered M   (penaltyCoef = --p1)\n\r");
	fprintf(stderr,"           11: Blended-energy tracker - 1st order descent\n\r");
	fprintf(stderr,"           12: 1st-order decent with gyro prior (lambda = --p1)\n\r");
	fprintf(stderr,"           13: Multi-tracker with rank penalty - Emp Dim  - centered M - and gyro prior.\n\r");
	fprintf(stderr,"               (penaltyCoef = --p1, lambda = --p2)\n\r\n\r");
	fprintf(stderr,"-m n       Selects the tracker feature management mode. Supported values for n are:\n\r");
	fprintf(stderr,"           0: Detect and manage features autonomously. Replace features that leave field of view\n\r");
	fprintf(stderr,"           1: Load initial feature locations from a file. Do not replace features that leave field\n\r");
	fprintf(stderr,"              of view. If this mode is selected, then an initial features file must be specified\n\r");
	fprintf(stderr,"              through the -i option.\n\r");
	fprintf(stderr,"           2: Load ground truth trajectories from file. When a feature leaves the field of view pre-\n\r");
	fprintf(stderr,"              maturely, or wanders too far from ground truth, the feature is re-initialized to the\n\r");
	fprintf(stderr,"              ground truth location. The number of re-initializations can be used as a performance\n\r");
	fprintf(stderr,"              metric for the tracker.\n\r\n\r");
	fprintf(stderr,"-g path    Sets path for ground truth file (containing ground truth trajectories). This option is\n\r");
	fprintf(stderr,"           needed when the feature management mode is set to 2.\n\r\n\r");
	fprintf(stderr,"-i path    Sets path for initial features file (containing initial locations of features in frame 0)\n\r");
	fprintf(stderr,"           This option is needed when the feature management mode is set to 1.\n\r\n\r");
	fprintf(stderr,"-e tol     Sets allowed amount that a feature can drift from ground truth before being re-\n\r");
	fprintf(stderr,"           initialized when the feature management mode is set to 2.\n\r");
	fprintf(stderr,"           This value is specified in units: pixels\n\r\n\r");
	fprintf(stderr,"-s n       Stop tracking after tracking through frame n. Frames are numbered starting with 0.\n\r\n\r");
	fprintf(stderr,"Some trackers allow tuning parameters to be set at runtime through command-line options. These are\n\r");
	fprintf(stderr,"never required - if omitted the program will use default values. Refer to the documentation on ths -t\n\r");
	fprintf(stderr,"option to see which trackers support run-time parameters and their meanings.\n\r");
	fprintf(stderr,"--p1=x     Set tracker parameter p1 to x. The meaning of p1 depends on the selected tracker (-t)\n\r");
	fprintf(stderr,"--p2=x     Set tracker parameter p2 to x. The meaning of p2 depends on the selected tracker (-t)\n\r");
	fprintf(stderr,"-h         Display this help message.\n\r\n\r");
	fprintf(stderr,"-r         Render tracker results to window while tracking.\n\r\n\r");
}

void tracker_Init(Tracker &myTracker, int mode, double reinitTolerance, rawDataLoader &dataLoader, string initFeaturesPath, string groundTruthPath) {
	//Grab frame 0
	Mat frame0;
	dataLoader.getFrame(0, frame0);
	frame0 = toGray(frame0);
	
	ImageStack frameStack;
	frameStack.BuildFromImage(frame0, parameters::M);
	
	switch(mode) {
		case 0:  myTracker.initialize(0, frameStack, mode, string(""));       break;
		case 1:  myTracker.initialize(0, frameStack, mode, initFeaturesPath); break;
		case 2:  myTracker.initialize(0, frameStack, mode, groundTruthPath);  break;
		default: fprintf(stderr,"Unsupported mode. Failed to initialize");
	}
	
	if (mode == 2) {
		//We also need to set the feature re-initialization tolerance prior to tracking
		myTracker.featureReinitTolerance = reinitTolerance;
	}
}

void tracker_TrackToNextFrame(Tracker &myTracker, int frameNumber, rawDataLoader &dataLoader, int trackerID, Matrix3d &Ktilde, Matrix3d &KtildeInv,
						bool renderLive, bool trackerParameter1Provided, bool trackerParameter2Provided, double trackerParameter1, double trackerParameter2) {
	Matrix3d H, R;
	
	//Get new frame of imagery
	Mat frame, grayFrame;
	dataLoader.getFrame(frameNumber, frame);
	grayFrame = toGray(frame);
	
	ImageStack frameStack;
	frameStack.BuildFromImage(grayFrame, parameters::M);
	
	//If relavent, grab inter-frame rotation matirx and compute between-frame homography
	if ((trackerID == 1) || (trackerID == 3) || (trackerID == 4) || (trackerID == 12) || (trackerID == 13)) {
		if (dataLoader.getRotation(frameNumber-1, frameNumber, R) != 0) {
			fprintf(stderr,"Error loading rotation between frames %d and %d.\n\r", frameNumber-1, frameNumber);
			return;
		}
		H = Ktilde*R*KtildeInv;
	}
	
	//Step tracker forward to new frame
	switch (trackerID) {
		case 0:  myTracker.trackToNextFrame_KLT                             (frameNumber, frameStack);    break;
		case 1:  myTracker.trackToNextFrame_KLT_HomInit                     (frameNumber, frameStack, H); break;
		case 2:  myTracker.trackToNextFrame_1stOrderDescent_GlobalMotionInit(frameNumber, frameStack);    break;
		case 3:  myTracker.trackToNextFrame_1stOrderDescent_HomInit         (frameNumber, frameStack, H); break;
		case 4:  myTracker.trackToNextFrame_GyroOptical_Experimental_1      (frameNumber, frameStack, H); break;
		case 5:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 6:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 7:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 8:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 9:  myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 10: myTracker.trackToNextFrame_multiTrack_GlobalMotionInit     (frameNumber, frameStack);    break;
		case 11: myTracker.trackToNextFrame_1stOrderDescent_Blended         (frameNumber, frameStack);    break;
		case 12:	if (trackerParameter1Provided) {
					myTracker.trackToNextFrame_1stOrderDescent_GyroPrior (frameNumber, frameStack, H, trackerParameter1);
					break;
				}
				else {
					myTracker.trackToNextFrame_1stOrderDescent_GyroPrior (frameNumber, frameStack, H);
					break;
				}
		case 13:	if (trackerParameter2Provided) {
					myTracker.trackToNextFrame_multiTrack_GyroPrior      (frameNumber, frameStack, H, trackerParameter2);
					break;
				}
				else {
					myTracker.trackToNextFrame_multiTrack_GyroPrior      (frameNumber, frameStack, H);
					break;
				}
		default: fprintf(stderr,"Error tracking to frame %d. Tracker ID not recognized.\n\r", frameNumber);
	}
	
	//If requested, render tracker results
	if (renderLive) {
		//We are not going to use "frame" for anything else, so we can write directly to the origional image buffer.
		myTracker.features.renderToImage(frame);
		imshow("Live Display", frame);
		waitKey(1);
	}
}

Mat toGray(Mat image) {
	Mat grayImage;
	if (image.type() == CV_8UC3) {
		cvtColor(image, grayImage, CV_RGB2GRAY);
	}
	else if (image.type() == CV_8UC1){
		grayImage = image;
	}
	else {
		fprintf(stderr,"Function toGray only supports 1 or 3-channel 8-bit images as input.\n\r");
		grayImage = Mat(0,0,CV_8UC1);
	}
	return(grayImage);
}

