#!/usr/bin/python
import os
import glob
import numpy as np
import cv
import cv2
import sys
#from vsgui.api import *
import subprocess
import math

#This program allows the user to view and edit the "ground-truth" feature trajectories for a recording.

#In this program, a trajectory is represented by a trajectory object. A collection of trajectories is
#stored as a list of trajectory objects

circleSize = 9
inactiveFeatureColor = (100,150,255)
activeFeatureColor   = (255,0,0)

folderPath = ''
inputFilePath = ''
outputFilePath = ''

trajectories = list()
currentFrame = 0
activeFeature = None
currentFrameImage = cv.CreateMat(1, 1, cv.CV_8UC3)
discontinuityThreshold = 50

#Initialize Font for overlaying text on images
font = cv.InitFont(cv.CV_FONT_HERSHEY_SIMPLEX, 0.7, 0.7, shear=0, thickness=2, lineType=8)

class trajectory:
	def __init__(self):
		self.startFrame = 0
		self.endFrame = 0
		self.trajectory = list()
		self.discontinuityIgnoreSet = set()
		
	def getLocation(self, frameNumber):
		if (frameNumber < self.startFrame) or (frameNumber > self.endFrame):
			return(None)
		for cell in self.trajectory:
			if cell[0] == frameNumber:
				return((cell[1],cell[2]))
		return(None)
	
	def truncate(self, frameNumber):
		#remove all entries in trajectory that occur at or later than the provided frame number are deleted
		cellNumber = 0
		while cellNumber < len(self.trajectory):
			cell = self.trajectory[cellNumber]
			if cell[0] >= frameNumber:
				self.trajectory.pop(cellNumber)
			else:
				cellNumber += 1
		if self.endFrame >= frameNumber:
			self.endFrame = frameNumber - 1
	
	def extend(self, frameNumber):
		#The features final position is repeated through the provided frame number
		finalCell = self.trajectory[len(self.trajectory) - 1]
		finalFrame = finalCell[0]
		if finalFrame < frameNumber:
			#Extend trajectory
			cell = [finalCell[0], finalCell[1], finalCell[2]]
			cell[0] += 1
			while cell[0] <= frameNumber:
				self.trajectory.append([cell[0], cell[1], cell[2]])
				cell[0] += 1
			self.endFrame = frameNumber
	
	def isEmpty(self):
		#Returns true if feature contains no valid history, and false otherwise
		if len(self.trajectory) == 0:
			return(True)
		else:
			return(False)

def readTrajectories(filePath):
	trajectories = list()
	dataFile = open(filePath,'r')
	for line in dataFile:
		line = line.lstrip(' ')
		line = line.rstrip('\n\r')
		if (len(line) > 0) and (line[0] != '#'):
			cells = line.split(':')
			newTrajectory = trajectory()
			for cell in cells:
				cell = cell.lstrip('(')
				cell = cell.rstrip(')')
				numbers = cell.split(',')
				if len(numbers) != 3:
					print "Error parsing input file. Invalid cell. Aborting."
					return
				frameNumber = int(numbers[0])
				x = float(numbers[1])
				y = float(numbers[2])
				newTrajectory.trajectory.append([frameNumber,x,y])
			newTrajectory.startFrame = (newTrajectory.trajectory[0])[0]
			newTrajectory.endFrame = (newTrajectory.trajectory[len(newTrajectory.trajectory) - 1])[0]
			trajectories.append(newTrajectory)
	dataFile.close()
	return(trajectories)


def writeTrajectories(filePath, trajectories):
	dataFile = open(filePath,'w')
	for feature in trajectories:
		#Write all but last cell with trailing colon
		for cellNumber in range(len(feature.trajectory)-1):
			cell = feature.trajectory[cellNumber]
			dataFile.write("({0:d},{1:.15f},{2:.15f}):".format(cell[0],cell[1],cell[2]))
		#Write last cell with trailing new line character
		cell = feature.trajectory[len(feature.trajectory)-1]
		dataFile.write("({0:d},{1:.15f},{2:.15f})\n".format(cell[0],cell[1],cell[2]))
	dataFile.close()


def grabFrame(folderPath, frameNumber):
	#Make sure folderPath ends in a file separator
	if folderPath[len(folderPath)-1] != os.sep:
		folderPath = folderPath + os.sep
	filename = "Frame_" + str(frameNumber) + ".png"
	filePath = folderPath + filename
	frame = cv.LoadImageM(filePath)
	return(frame)
	

def updateDisplay(rawFrameImage, features):
	#Render current frame with current features. If activeFeature is not None, it will be
	#rendered in an alternate color.
	vis = cv.CreateMat(rawFrameImage.rows, rawFrameImage.cols, cv.CV_8UC3)
	cv.Copy(rawFrameImage, vis)
	for feature in features:
		if (currentFrame >= feature.startFrame) and (currentFrame <= feature.endFrame):
			#Feature currently exists and needs to be rendered
			currentLocation = [0,0]
			for cell in feature.trajectory:
				if cell[0] == currentFrame:
					currentLocation[0] = cell[1]
					currentLocation[1] = cell[2]
					break
			x = int(currentLocation[1])
			y = int(currentLocation[0])
			cv.Circle(vis, (x,y), circleSize, inactiveFeatureColor, thickness=3, lineType=8, shift=0)
	if activeFeature != None:
		if (currentFrame >= activeFeature.startFrame) and (currentFrame <= activeFeature.endFrame):
			currentLocation = [0,0]
			for cell in activeFeature.trajectory:
				if cell[0] == currentFrame:
					currentLocation[0] = cell[1]
					currentLocation[1] = cell[2]
					break
			x = int(currentLocation[1])
			y = int(currentLocation[0])
			cv.Circle(vis, (x,y), circleSize, activeFeatureColor, thickness=3, lineType=8, shift=0)
		else:
			cv.PutText(vis,"DNE",(4,20),font,activeFeatureColor)
	cv.ShowImage("Display",vis)


def onMouse(event, x, y, flags, param):
	if event == cv.CV_EVENT_LBUTTONDOWN:
		#Find point closest to mouse click
		global activeFeature
		initialized = False
		minDist = 0
		for feature in trajectories:
			currentLoc = feature.getLocation(currentFrame)
			if currentLoc != None:
				delta_x = x - currentLoc[1]
				delta_y = y - currentLoc[0]
				dist = math.sqrt(delta_x*delta_x + delta_y*delta_y)
				if not(initialized):
					minDist = dist
					activeFeature = feature
					initialized = True
				elif dist < minDist:
					minDist = dist
					activeFeature = feature
		if initialized:
			loc = activeFeature.getLocation(currentFrame)
			print "Activated Feature at location (%f,%f)"%(loc[0],loc[1])
		updateDisplay(currentFrameImage, trajectories)


def adjustFeatureLeft(feature, frameNumber):
	if (feature != None) and (frameNumber >= feature.startFrame) and (frameNumber <= feature.endFrame):
		for cell in feature.trajectory:
			if cell[0] >= frameNumber:
				cell[2] -= 1


def adjustFeatureRight(feature, frameNumber):
	if (feature != None) and (frameNumber >= feature.startFrame) and (frameNumber <= feature.endFrame):
		for cell in feature.trajectory:
			if cell[0] >= frameNumber:
				cell[2] += 1


def adjustFeatureUp(feature, frameNumber):
	if (feature != None) and (frameNumber >= feature.startFrame) and (frameNumber <= feature.endFrame):
		for cell in feature.trajectory:
			if cell[0] >= frameNumber:
				cell[1] -= 1


def adjustFeatureDown(feature, frameNumber):
	if (feature != None) and (frameNumber >= feature.startFrame) and (frameNumber <= feature.endFrame):
		for cell in feature.trajectory:
			if cell[0] >= frameNumber:
				cell[1] += 1



def searchForFeatureWithDiscontinuity(features, threshold):
	#Return a feature with a trajectory discontinuity, or None if none could be found.
	for feature in features:
		firstDiscontinuity = findDiscontinuity(feature, threshold)
		if firstDiscontinuity != None:
			return(feature)
	return(None)


def findDiscontinuity(feature, threshold):
	#Find the first time the feature jumps more than the provided threshold between frames. Return the frame number
	#right before the jump. Return None if there are no such events.
	lastx = feature.trajectory[0][1]
	lasty = feature.trajectory[0][2]
	for cellNumber in range(1,len(feature.trajectory)):
		cell = feature.trajectory[cellNumber]
		frameNumber = cell[0]
		x = cell[1]
		y = cell[2]
		dist = math.sqrt((x-lastx)*(x-lastx) + (y-lasty)*(y-lasty))
		if (dist > threshold) and ((frameNumber-1) not in feature.discontinuityIgnoreSet):
			return(frameNumber-1)
		else:
			lastx = x
			lasty = y
	return(None)


def repair(feature, frameNumber):
	#Compute the jump from frameNumber-1 to frameNumber and shift all locations starting at frameNumber by -1 times
	#this amount. This is used to correct for large jumps that occur when a feature is mis-registered.
	oldLoc = feature.getLocation(frameNumber - 1)
	newLoc = feature.getLocation(frameNumber)
	if (oldLoc != None) and (newLoc != None):
		correction = [oldLoc[0]-newLoc[0], oldLoc[1]-newLoc[1]]
		for cell in feature.trajectory:
			if cell[0] >= frameNumber:
				cell[1] += correction[0]
				cell[2] += correction[1]
	

def truncateFeaturesThatLeaveView(features):
	#Go through a list of features and truncate each feature when it first leaves the field of view.
	for feature in features:
		truncateFeatureWhenItLeavesView(feature)

def eliminateRedundantFeatures(features, numFrames):
	#Delete redundant features
	for frameNumber in range(numFrames):
		eliminateRedundantFeaturesStartingAtSameTime(features, frameNumber)
	for frameNumber in range(1,numFrames):
		eliminateRedundantFeaturesThatDontStartAtSameTime(features, frameNumber)
	

def eliminateRedundantFeaturesStartingAtSameTime(features, frameNumber=0):
	#Look for features that are too close and both start at the given frame number.
	closenessThreshold = 2
	global activeFeature
	startPositions = list()
	conflicts = list() #A list of positions which multiple featueres are close to
	for feature in features:
		if feature.startFrame == frameNumber:
			startPosition = feature.getLocation(frameNumber)
			for loc in startPositions:
				if math.sqrt(math.pow(loc[0] - startPosition[0],2) + math.pow(loc[1] - startPosition[1],2)) < closenessThreshold:
					conflicts.append(startPosition)
			startPositions.append(startPosition)
	#Resolve conflicts
	for loc in conflicts:
		conflictingFeatures = list()
		for feature in features:
			if feature.startFrame == frameNumber:
				startPosition = feature.getLocation(frameNumber)
				if math.sqrt(math.pow(loc[0] - startPosition[0],2) + math.pow(loc[1] - startPosition[1],2)) < closenessThreshold:
					conflictingFeatures.append(feature)
		#Keep the feature that has the longest lifespan
		maxTrackLen = 0
		for feature in conflictingFeatures:
			if (feature.endFrame - feature.startFrame > maxTrackLen):
				maxTrackLen = feature.endFrame - feature.startFrame
		for feature in conflictingFeatures:
			if (feature.endFrame - feature.startFrame < maxTrackLen):
				conflictingFeatures.remove(feature)
				features.remove(feature)
				if activeFeature == feature:
					activeFeature = None
				print("Removing redundant feature from frame %d"%(frameNumber))
		if len(conflictingFeatures) > 1:
			for i in range(1,len(conflictingFeatures)):
				features.remove(conflictingFeatures[i])
				if activeFeature == conflictingFeatures[i]:
					activeFeature = None
				print("Removing redundant feature from frame %d"%(frameNumber))
	
def eliminateRedundantFeaturesThatDontStartAtSameTime(features, frameNumber=0):
	#Look for features that start too close to an existing feature
	closenessThreshold = 2
	global activeFeature
	existingFeatureLocations = list()
	for feature in features:
		if (frameNumber > feature.startFrame) and (frameNumber <= feature.endFrame):
			existingFeatureLocations.append(feature.getLocation(frameNumber))
	for feature in features:
		if feature.startFrame == frameNumber:
			startPosition = feature.getLocation(frameNumber)
			#See if feature starts too close to an existing feature
			for loc in existingFeatureLocations:
				if math.sqrt(math.pow(loc[0] - startPosition[0],2) + math.pow(loc[1] - startPosition[1],2)) < closenessThreshold:
					features.remove(feature)
					if activeFeature == feature:
						activeFeature = None
					print("Removing redundant feature from frame %d because it is born too close to an existing feature."%(frameNumber))
					break

def truncateFeatureWhenItLeavesView(feature):
	#Check if a single feature ever leaves the field of view. If it does, truncate it at the first such occurance.
	X = currentFrameImage.rows
	Y = currentFrameImage.cols
	for cellNumber in range(len(feature.trajectory)):
		cell = feature.trajectory[cellNumber]
		frameNumber = cell[0]
		x = cell[1]
		y = cell[2]
		if (x < 0) or (x >= X) or (y < 0) or (y >= Y):
			print "Truncating feature"
			feature.truncate(frameNumber)
			break


#Select the recording to edit
recordingPaths = glob.glob("Recordings" + os.sep + "*")
recordingPaths.sort()
command = "zenity --list --title='Choose a recording' --column='Available Recordings'"
for i in range(len(recordingPaths)):
	command += " '" + recordingPaths[i] + "'"
output = subprocess.check_output(command, shell=True)
output = output[0:(len(output)-1)]
folderPath = os.path.abspath(output)
if (folderPath[len(folderPath) - 1] != os.sep):
	folderPath += os.sep
inputFilePath  = folderPath + "GroundTruth.txt"
outputFilePath = folderPath + "GroundTruth.txt"

#Count frames in data folder
frameFilenames = glob.glob(folderPath + "Frame_*.png")
numFrames = len(frameFilenames)

#Load the source file
trajectories = readTrajectories(inputFilePath)

#Load frame 0
currentFrameImage = grabFrame(folderPath, currentFrame)

cv.NamedWindow("Display",cv.CV_WINDOW_AUTOSIZE)
cv.SetMouseCallback("Display", onMouse, 0)
updateDisplay(currentFrameImage, trajectories)

print("Press 'c' to clean all feature trajectories. This truncates features thar leave the FOV.")
print("It also eliminates redundant features - i.e. features that appear very close to existing features.")
print("Press 'h' to hunt for a feature and frame with a discontinuity.")
print("")
print("***************     Click a feature to activate it     ***************")
print("Use the arrow keys to adjust the activated feature.")
print("Press 'f' to find and move to the first discontinuity in the current features track.")
print("Press 'i' to add the current discontinuity (between last and current frame) to the ignore set")
print("Press 'r' to repair a jump (remove the displacement from the previous to the current frame)")
print("Press 'd' to delete an entire feature track.")
print("Press 't' to terminate the active track on the current frame.")
print("Press 'e' to extend the active track to the current frame (if applicable)")
print("Use   'n' and 'p' or '<' and '>' to navigate frames.")
print("Press 'b' to jump to the beginning on the video sequence")
print("")
print("Press 's' to save the current trajectories.")
print("Press spacebar to exit.")

while True:
	key = cv.WaitKey()
	if key == 65361:   #Left
		adjustFeatureLeft(activeFeature, currentFrame)
	elif key == 65363: #Right
		adjustFeatureRight(activeFeature, currentFrame)
	elif key == 65362: #Up
		adjustFeatureUp(activeFeature, currentFrame)
	elif key == 65364: #Down
		adjustFeatureDown(activeFeature, currentFrame)
	elif key == 100: #'d'
		if activeFeature != None:
			trajectories.remove(activeFeature)
			activeFeature = None
			print "Feature Deleted"
	elif key == ord('t'):
		if activeFeature != None:
			activeFeature.truncate(currentFrame)
			if activeFeature.isEmpty():
				trajectories.remove(activeFeature)
				activeFeature = None
				print "Feature Deleted because it was truncated out of existence."
			else:
				print "Feature truncated to frame " + str(currentFrame)
	elif key == ord('e'):
		if activeFeature != None:
			activeFeature.extend(currentFrame)
			print "Active feature extended to current frame."
	elif key == ord('f'):
		if activeFeature != None:
			jumpFrame = findDiscontinuity(activeFeature, discontinuityThreshold)
			if jumpFrame == None:
				print "No discontinuity in trajectory found."
			else:
				print "Moving to first discontinuity. At frame %d."%(jumpFrame)
				currentFrame = jumpFrame
				currentFrameImage = grabFrame(folderPath, currentFrame)
	elif key == ord('r'):
		if activeFeature != None:
			print "Attempting to repair discontinuity"
			repair(activeFeature, currentFrame)
	elif key == ord('c'):
		print "Cleaning all features (truncate when leaving FOV)."
		truncateFeaturesThatLeaveView(trajectories)
		eliminateRedundantFeatures(trajectories, numFrames)
	elif key == ord('h'):
		troubledFeature = searchForFeatureWithDiscontinuity(trajectories, discontinuityThreshold)
		if troubledFeature == None:
			print "No features with discontinuities found."
		else:
			activeFeature = troubledFeature
			jumpFrame = findDiscontinuity(activeFeature, discontinuityThreshold)
			currentFrame = jumpFrame
			currentFrameImage = grabFrame(folderPath, currentFrame)
			oldLoc = activeFeature.getLocation(currentFrame)
			newLoc = activeFeature.getLocation(currentFrame+1)
			deltax = newLoc[0] - oldLoc[0]
			deltay = newLoc[1] - oldLoc[1]
			dist = math.sqrt(deltax*deltax + deltay*deltay)
			print "Feature with discontinuity of size %f found"%(dist)
	elif key == ord('i'):
		if activeFeature != None:
			activeFeature.discontinuityIgnoreSet.add(currentFrame-1)
			print "Ignoring discontinuity of active feature between last and current frame"
	elif key == 115:   #'s'
		writeTrajectories(outputFilePath, trajectories)
		print("Modified Trajectories saved to file " + outputFilePath)
	elif (key == 110) or (key == 46): #'n' or '>'
		currentFrame = currentFrame + 1
		if currentFrame >= numFrames:
			currentFrame = numFrames - 1
		currentFrameImage = grabFrame(folderPath, currentFrame)
	elif (key == 112) or (key == 44): #'p' or '<'
		currentFrame = currentFrame - 1
		if currentFrame < 0:
			currentFrame = 0
		currentFrameImage = grabFrame(folderPath, currentFrame)
	elif key == ord('b'):
		currentFrame = 0
		currentFrameImage = grabFrame(folderPath, currentFrame)
	elif key == 32:    #Space
		break
	updateDisplay(currentFrameImage, trajectories)
