#!/usr/bin/python
import sys
import os
import shutil
import glob
import getopt
import re

#This function copies a subset of available recordings (in the Recordings folder) into the
#"Training Videos" folder and another (disjoint) subset of videos into the "Test Videos"
#folder. Each frame in each copied video is degraded by calling the program ./degradeImage
#on it. By default, this program will skip copying and degrading any video for which the
#target folder already exists. If the -f command-line option is provided, this will force
#a replacement of the target video whenever conflicts arise.

#Specify default list of test videos
TrainingVideos = [0,1,4,7]
TestVideos = [2,3,6,8,9,10,11,12]

def degradeRecording(recordingPath, profile):
	imagePaths = glob.glob(recordingPath + os.sep + "*.png")
	imagePaths.sort()
	if profile != "none":
		optionString = "--%s"%(profile)
		for path in imagePaths:
			print "\rDegrading Image: %s"%(os.path.basename(path)),
			sys.stdout.flush()
			command = "./degradeImage %s '%s'"%(optionString, path)
			retvalue = os.system(command)
			if (retvalue != 0):
				print "Failed!"
				print "Command was: %s"%(command)
				print "Aborting."
				break
		print ""

def printUsage():
	print "print sys.argv[0] Generates (possibly degraded) copies of source videos for tracker training and testing."
	print "Options:"
	print "-f       Force overwriting of existing target videos. By default, if a target video folder"
	print "         already exists, that video is skipped and the target is left unchanged."
	print "-h       Print this usage message."
	print "--train  Generate training videos."
	print "--test   Generate testing videos."
	print ""
	print "By default, the videos are not synthetically degraded."
	print "This behavior can be controlled through the following options:"
	print "--none     Generate videos with no degradation (just copy)"
	print "--low      Generate videos using low degradation profile"
	print "--medium   Generate videos using medium degradation profile"
	print "--high     Generate videos using high degradation profile"
	print "--extreme  Generate videos using extreme degradation profile"
	print ""
	print "If neither of the options --train or --test is given, then both sets are produced."

#Parse command-line options
try:
	opts, args = getopt.getopt(sys.argv[1:],"fh",["train", "test", "none", "low", "medium", "high", "extreme"])
except getopt.GetoptError:
	printUsage()
	sys.exit(2)
forceFlag = False
videoSetSpecified = False
processTrainingVids = False
processTestingVids = False
degradationProfile = "none"
degradationProfileSet = False
for o, a in opts:
	if o == "-f":
		forceFlag = True
	elif o == "-h":
		printUsage()
		sys.exit(0)
	elif o == "--train":
		videoSetSpecified = True
		processTrainingVids = True
	elif o == "--test":
		videoSetSpecified = True
		processTestingVids = True
	elif o == "--none":
		if degradationProfileSet:
			print "Degradation profile already set by previous option. Ignoring option: %s"%o
		else:
			degradationProfileSet = True
			degradationProfile = "none"
	elif o == "--low":
		if degradationProfileSet:
			print "Degradation profile already set by previous option. Ignoring option: %s"%o
		else:
			degradationProfileSet = True
			degradationProfile = "low"
	elif o == "--medium":
		if degradationProfileSet:
			print "Degradation profile already set by previous option. Ignoring option: %s"%o
		else:
			degradationProfileSet = True
			degradationProfile = "medium"
	elif o == "--high":
		if degradationProfileSet:
			print "Degradation profile already set by previous option. Ignoring option: %s"%o
		else:
			degradationProfileSet = True
			degradationProfile = "high"
	elif o == "--extreme":
		if degradationProfileSet:
			print "Degradation profile already set by previous option. Ignoring option: %s"%o
		else:
			degradationProfileSet = True
			degradationProfile = "extreme"
	else:
		print "Unhandled option - Ignored"
if not videoSetSpecified:
	processTrainingVids = True
	processTestingVids = True


#Enumerate videos and detirmine paths for training and test videos
pattern = re.compile("Recording [0-9]+")
recordingPaths = glob.glob("Recordings" + os.sep + "*")
TrainingVideoPaths = list()
TestVideoPaths = list()
for path in recordingPaths:
	recordingName = os.path.basename(path)
	if pattern.match(recordingName) != None:
		parts = recordingName.split(' ')
		recordingNumber = int(parts[1])
		if recordingNumber in TrainingVideos:
			TrainingVideoPaths += (os.path.abspath(path),)
		elif recordingNumber in TestVideos:
			TestVideoPaths += (os.path.abspath(path),)

#Generate Training Videos
if processTrainingVids:
	print "Generating Training Videos"
	for recordingPath in TrainingVideoPaths:
		recordingName = os.path.basename(recordingPath)
		targetPath = os.path.abspath("Training Videos" + os.sep + recordingName)
	
		if (os.path.isdir(targetPath) and (forceFlag == False)):
			sys.stderr.write("Target for %s already exists - skipping.\n\r"%(recordingName))
		else:
			if os.path.isdir(targetPath):
				print "Target for %s exists - It will be replaced."%(recordingName)
				shutil.rmtree(targetPath, True)
			else:
				print "No target version of %s found. One will be created."%(recordingName)
			print "Copying recording... ",
			shutil.copytree(recordingPath, targetPath)
			print "Done"
			print "Degrading frames."
			degradeRecording(targetPath, degradationProfile)
			print "Done degrading recording."

#Generate Test Videos
if processTestingVids:
	print "Generating Test Videos"
	for recordingPath in TestVideoPaths:
		recordingName = os.path.basename(recordingPath)
		targetPath = os.path.abspath("Test Videos" + os.sep + recordingName)
	
		if (os.path.isdir(targetPath) and (forceFlag == False)):
			sys.stderr.write("Target for %s already exists - skipping.\n\r"%(recordingName))	
		else:
			if os.path.isdir(targetPath):
				print "Target for %s exists - It will be replaced."%(recordingName)
				shutil.rmtree(targetPath, True)
			else:
				print "No target version of %s found. One will be created."%(recordingName)
			print "Copying recording... ",
			shutil.copytree(recordingPath, targetPath)
			print "Done"
			print "Degrading frames."
			degradeRecording(targetPath, degradationProfile)
			print "Done degrading recording."
