#!/usr/bin/python
import cv, cv2
import sys, os
import glob
import subprocess
import getopt

#This function goes through each recording in the folder "Test Videos" and executes a collection
#of trackers on each one with either feature management mode set to 1 or 2.
#If run with mode 1:
#Features will be initialized in frame 0 using ground truth. No new features will be initialized
#throughout the run. Features will be terminated when they leave the view. A single file will be
#created: Tracker_n_OutputTrajectories_mode1.txt. This will contain the trajectories output by the tracker.
#
#If run with mode 2:
#Features will be re-initialized using ground truth when they leave the scene early or wander too
#far fround ground truth (threshold set in this file). 2 files will be created for each tracker and
#each data set. The first is Tracker_n_OutputTrajectories_mode2.txt and will contain feature trajectories.
#The second is Tracker_n_mode2_Results.txt. This will contain performance measures for the tracker.
#
#Which trackers are run and which mode is used can be specified on the command line.

#Command line options:
#-t x,y,z     Run trackers with the given tracker IDs. If ommited, all trackers are run.
#-f           "Force" - run and overwrite results if they already exist. By default, old results
#             are not overwritten.
#-m x         Run trackers with feature management mode x (supported modes are 1 and 2).

#Specify re-initialization tolerance for tracking against ground truth
reinitTolerance = 10

#Specify default options
trackers = [0,1,2,3,6,12,13]
#trackers = [12,13]
forceFlag = False
mode = 2

#Specify additional options on a per-tracker basis. If additional options are not needed use an empty string.
#Otherwise, the string should have a leading space. It will be inserted into the tracking system call before
#the main arguments.
additionalOptions = {0:"",
				 1:"",
				 2:"",
				 3:"",
				 4:"",
				 5:" --p1=0.45", #0.45
				 6:" --p1=0.45", #0.45
				 7:"",
				 8:"",
				 9:"",
				 10:"",
				 11:"",
				 12:" --p1=0.0125", #0.0125
				 13:" --p1=0.4 --p2=0.00003"} #Gyro and rank terms
				 #13:" --p1=0.0 --p2=0.005"} #Gyro term only

def runTracker_mode1(trackerID, folderPath, force):
	#First check to see if results already exist
	if (folderPath[len(folderPath) - 1] != os.sep):
		folderPath += os.sep
	TrajectoriesPath = folderPath + "Tracker_%d_OutputTrajectories_mode1.txt"%(trackerID)
	if os.path.isfile(TrajectoriesPath) and not(force):
		print "Trajectories file already exist - skipping."
	else:
		groundTruthPath = folderPath + "GroundTruth.txt"
		command = "./trackPostProcess -t %d"%(trackerID) + additionalOptions[trackerID] + " -m 1 -i '" + groundTruthPath + "' '" + folderPath + "' '" + TrajectoriesPath + "'"
		output = subprocess.check_output(command, shell=True)

def runTracker_mode2(trackerID, folderPath, force):
	#First check to see if results already exist
	if (folderPath[len(folderPath) - 1] != os.sep):
		folderPath += os.sep
	TrajectoriesPath = folderPath + "Tracker_%d_OutputTrajectories_mode2.txt"%(trackerID)
	ResultsPath = folderPath + "Tracker_%d_mode2_Results.txt"%(trackerID)
	if os.path.isfile(TrajectoriesPath) and os.path.isfile(ResultsPath) and not(force):
		print "Results and Trajectories files already exist - skipping."
	else:
		groundTruthPath = folderPath + "GroundTruth.txt"
		command = "./trackPostProcess -t %d"%(trackerID) + additionalOptions[trackerID] + " -m 2 -g '" + groundTruthPath + "' -e %f '"%(reinitTolerance) + folderPath + "' '" + TrajectoriesPath + "'"
		output = subprocess.check_output(command, shell=True)
		#Pan through the output for tracker results
		headToken = "******************   Begin Results   ******************\n"
		tailToken = "*******************   End Results   *******************\n"
		startIndex = output.find(headToken)
		endIndex   = output.find(tailToken, startIndex)
		if (startIndex != -1) and (endIndex != -1):
			results = output[(startIndex + len(headToken)):endIndex]
			FID = open(ResultsPath, 'w')
			FID.write(results)
			FID.close()
		else:
			print "No properly formatted results section in tracker output. Tracker_n_Results.txt will not be created."

#Parse command-line options
try:
	opts, args = getopt.getopt(sys.argv[1:],"t:fm:")
except getopt.GetoptError:
	print sys.argv[0] + ' -t x,y,z -f -m m'
	sys.exit(2)

for opt in opts:
	if opt[0] == '-t':
		trackerIDs = opt[1].split(',')
		trackers = []
		for item in trackerIDs:
			trackers.append(int(item))
		trackers.sort()
	elif opt[0] == '-f':
		forceFlag = True
	elif opt[0] == '-m':
		mode = int(opt[1])

print "Running trackers: ",
for item in trackers:
	print "%d "%(item),
print ""
if forceFlag:
	print "Force overwrite of old results: True"
else:
	print "Force overwrite of old results: False"
print "Running with feature management mode %d"%(mode)
print ""

#Get list of test videos (absolute paths of directories in Test Videos folder)
testVidPaths = glob.glob("Test Videos" + os.sep + "*")
i = 0
while i < len(testVidPaths):
	testVidPaths[i] = os.path.abspath(testVidPaths[i])
	if not(os.path.isdir(testVidPaths[i])):
		testVidPaths.pop(i)
	else:
		i += 1

for testVid in testVidPaths:
	testVidName = os.path.basename(testVid)
	for trackerID in trackers:
		print "Running tracker (ID = %d) on "%(trackerID) + testVidName
		if mode == 1:
			runTracker_mode1(trackerID, testVid, forceFlag)
		elif mode == 2:
			runTracker_mode2(trackerID, testVid, forceFlag)
		else:
			print "Unsupported feature management mode."
			sys.exit(-1)
