#!/usr/bin/python
import os
import glob
import numpy as np
import cv
import cv2
import sys
import subprocess
import getopt

#This program allows the user to view the results of different trackers overlayed on top of the source video
#You can specify the result set with the command line switch -m x. If x = 1, view results of running trackers
#in feature management mode 1. If x=2, view results of running with feature management mode 2.

#In this program, a trajectory is represented by a trajectory object. A collection of trajectories is
#stored as a list of trajectory objects

#Specify default options
mode = 1

resultSets = list()
currentFrame = 0
currentFrameImage = cv.CreateMat(1, 1, cv.CV_8UC3)

#Initialize Font for overlaying text on images
font = cv.InitFont(cv.CV_FONT_HERSHEY_SIMPLEX, 0.7, 0.7, shear=0, thickness=2, lineType=8)

class trajectory:
	def __init__(self):
		self.startFrame = 0
		self.endFrame = 0
		self.trajectory = list()
		
	def getLocation(self, frameNumber):
		if (frameNumber < self.startFrame) or (frameNumber > self.endFrame):
			return(None)
		for cell in self.trajectory:
			if cell[0] == frameNumber:
				return((cell[1],cell[2]))
		return(None)

class resultSet:
	def __init__(self, trackerID):
		self.features = list()
		self.trackerID = trackerID
		self.active = False
		
		self.circleSize = 9
		self.keyPosition = (0,0)
		
		if trackerID == '0':
			self.color = (255,0,0)
		elif trackerID == '1':
			self.color = (255,0,0)
		elif trackerID == '2':
			self.color = (0,0,255)
		elif trackerID == '3':
			self.color = (128,128,0)
		elif trackerID == '4':
			self.color = (0,128,128)
		elif trackerID == '5':
			self.color = (128,0,128)
		elif trackerID == '6':
			self.color = (255,255,255)
		elif trackerID == '7':
			self.color = (128,128,0)
		elif trackerID == '8':
			self.color = (0,128,128)
		elif trackerID == '9':
			self.color = (128,0,128)
		elif trackerID == 'a':
			self.color = (255,255,255)
		elif trackerID == 'g':
			self.color = (0,255,0)
		else:
			self.color = (0,0,255)

def readTrajectories(filePath):
	trajectories = list()
	dataFile = open(filePath,'r')
	for line in dataFile:
		line = line.lstrip(' ')
		line = line.rstrip('\n\r')
		if (len(line) > 0) and (line[0] != '#'):
			cells = line.split(':')
			newTrajectory = trajectory()
			for cell in cells:
				cell = cell.lstrip('(')
				cell = cell.rstrip(')')
				numbers = cell.split(',')
				if len(numbers) != 3:
					print "Error parsing input file. Invalid cell. Aborting."
					return
				frameNumber = int(numbers[0])
				x = float(numbers[1])
				y = float(numbers[2])
				newTrajectory.trajectory.append([frameNumber,x,y])
			newTrajectory.startFrame = (newTrajectory.trajectory[0])[0]
			newTrajectory.endFrame = (newTrajectory.trajectory[len(newTrajectory.trajectory) - 1])[0]
			trajectories.append(newTrajectory)
	dataFile.close()
	return(trajectories)

def toggleStatus(identifier):
	for results in resultSets:
		if results.trackerID == identifier:
			if results.active == True:
				results.active = False
			else:
				results.active = True

def grabFrame(folderPath, frameNumber):
	#Make sure folderPath ends in a file separator
	if folderPath[len(folderPath)-1] != os.sep:
		folderPath = folderPath + os.sep
	filename = "Frame_" + str(frameNumber) + ".png"
	filePath = folderPath + filename
	frame = cv.LoadImageM(filePath)
	return(frame)

def getDisplayImage():
	#Render current frame with current features. If activeFeature is not None, it will be
	#rendered in an alternate color.
	vis = cv.CreateMat(currentFrameImage.rows, currentFrameImage.cols, cv.CV_8UC3)
	cv.Copy(currentFrameImage, vis)
	for results in resultSets:
		if results.active:
			cv.PutText(vis,"%s"%(results.trackerID),results.keyPosition,font,results.color)
			features = results.features
			for feature in features:
				currentLocation = feature.getLocation(currentFrame)
				if currentLocation != None:
					try:
						x = int(currentLocation[1])
						y = int(currentLocation[0])
						cv.Circle(vis, (x,y), results.circleSize, results.color, thickness=3, lineType=8, shift=0)
					except:
						pass
	return(vis)

def updateDisplay():
	vis = getDisplayImage()
	cv.ShowImage("Display",vis)

#Parse command-line options
try:
	opts, args = getopt.getopt(sys.argv[1:],"m:")
except getopt.GetoptError:
	print sys.argv[0] + ' -m m'
	sys.exit(2)

for opt in opts:
	if opt[0] == '-m':
		mode = int(opt[1])

#Select the recording to view
recordingPaths = glob.glob("Test Videos" + os.sep + "*")
recordingPaths.sort()
i = 0
while i < len(recordingPaths):
	if not(os.path.isdir(recordingPaths[i])):
		recordingPaths.pop(i)
	else:
		i += 1
command = "zenity --list --title='Choose a Test Video' --column='Available Results Sets'"
for i in range(len(recordingPaths)):
	command += " '" + recordingPaths[i] + "'"
output = subprocess.check_output(command, shell=True)
output = output[0:(len(output)-1)]
folderPath = os.path.abspath(output)
if (folderPath[len(folderPath) - 1] != os.sep):
	folderPath += os.sep

#Load frame 0
currentFrameImage = grabFrame(folderPath, currentFrame)

#Enumerate Trajectory files in folder
trajectoryFiles = glob.glob(folderPath + "Tracker_*_OutputTrajectories_mode%d.txt"%(mode))
trajectoryFiles.sort()
xKeyPosition = currentFrameImage.cols - 160
yKeyPosition = 20
for trajFile in trajectoryFiles:
	filename = os.path.basename(trajFile)
	parts = filename.split('_')
	ID = parts[1]
	if ID == '10':
		#Re-map ID 10 to 'a'
		ID = 'a'
	if ID == '11':
		#Re-map ID 11 to 'b'
		ID = 'b'
	if ID == '12':
		#Re-map ID 12 to 'c'
		ID = 'c'
	if ID == '13':
		#Re-map ID 13 to 'd'
		ID = 'd'
	results = resultSet(ID)
	results.features = readTrajectories(trajFile)
	results.keyPosition = (xKeyPosition, yKeyPosition)
	xKeyPosition += 20
	resultSets.append(results)

#Load Ground-Truth
groundTruthFilePath = folderPath + "GroundTruth.txt"
groundTruthSet = resultSet('g')
groundTruthSet.features = readTrajectories(groundTruthFilePath)
groundTruthSet.active = True
groundTruthSet.circleSize = 5
groundTruthSet.keyPosition = (xKeyPosition, yKeyPosition)
resultSets.append(groundTruthSet)

#Count frames in data folder
frameFilenames = glob.glob(folderPath + "Frame_*.png")
numFrames = len(frameFilenames)

cv.NamedWindow("Display",cv.CV_WINDOW_AUTOSIZE)
updateDisplay()

print("")
print("Viewing results with feature management mode = %d"%(mode))
print("")
print("Use   'n' and 'p' or '<' and '>' to navigate frames.")
print("Press numbers 0-9 to (de)-activeate the corresponding trackers results.")
print("Press 'g' to toggle groundtruth.")
print("Press spacebar to exit.")

while True:
	key = cv.WaitKey()
	if key == ord('0'):
		toggleStatus('0')
	elif key == ord('1'):
		toggleStatus('1')
	elif key == ord('2'):
		toggleStatus('2')
	elif key == ord('3'):
		toggleStatus('3')
	elif key == ord('4'):
		toggleStatus('4')
	elif key == ord('5'):
		toggleStatus('5')
	elif key == ord('6'):
		toggleStatus('6')
	elif key == ord('7'):
		toggleStatus('7')
	elif key == ord('8'):
		toggleStatus('8')
	elif key == ord('9'):
		toggleStatus('9')
	elif key == ord('g'):
		toggleStatus('g')
	elif key == ord('a'):
		toggleStatus('a')
	elif key == ord('b'):
		toggleStatus('b')
	elif key == ord('c'):
		toggleStatus('c')
	elif key == ord('d'):
		toggleStatus('d')
	elif key == ord('s'):
		#Save Image
		ID = "None"
		for Set in resultSets:
			if Set.active:
				trackerIDString = ''
				if (Set.trackerID == 'c'):
					trackerIDString = "12"
				else:
					trackerIDString = Set.trackerID
				ID = "TrackerID_%s"%(trackerIDString)
				break
		outPath = ID + "_Frame_%d.png"%(currentFrame)
		vis = getDisplayImage()
		cv.SaveImage(outPath, vis)
		sys.stderr.write("\nCurrent image (frame %d) saved to: %s\n"%(currentFrame,outPath))
	elif (key == 110) or (key == 46): #'n' or '>'
		currentFrame = currentFrame + 1
		if currentFrame >= numFrames:
			currentFrame = numFrames - 1
		currentFrameImage = grabFrame(folderPath, currentFrame)
		sys.stderr.write("\rCurrent Frame: %d"%(currentFrame))
	elif (key == 112) or (key == 44): #'p' or '<'
		currentFrame = currentFrame - 1
		if currentFrame < 0:
			currentFrame = 0
		currentFrameImage = grabFrame(folderPath, currentFrame)
		sys.stderr.write("\rCurrent Frame: %d"%(currentFrame))
	elif key == 32:    #Space
		break
	updateDisplay()
sys.stderr.write("\n\r")
